ASSEMBLE

* place a string on the stack to be used as a receive buffer
* (this string can be generated quickly using my blankstr.s subroutine)
* then place a hex string on stack representing a timeout value
* (note that a timeout value of about 0FFFF gives ~3 seconds on an SX)
* this program will receive characters from the serial port at 9600 baud
* and store them in the string buffer until a #127d char is received
* or a time out
* the string buffer will be truncated to the size of the input
* it is ok to allocate a string buffer the size of remaining memory

           NIBASC  /HPHP48-E/

ENDCHAR    EQU     127

RBR        EQU     #114
RCS        EQU     #111
CRER       EQU     #113
TBR        EQU     #116
TCS        EQU     #112
IOC        EQU     #110
BAU        EQU     #10D
ANNCTRL2   EQU     #10C

POPHXS     EQU     #53F8D

RPL

::
 CK2NOLASTWD
 CK&DISPATCH1 # 3B             ( *2: character string, 1: hex string* )
 ::
  OVER LEN$
  CODE
           GOSBVL  =POP#       pop string-length bint off stack
           R1=A.F  A           save size in R1
           GOSBVL  POPHXS      also does a SAVPTR, returns with P<>0 !
           R2=A.F  A           save timeout value in R2
           P=      0

           D0=(5)  IOC
           A=0     B
           DAT0=A  1           de-activate port, clear SON

           D0=(5)  BAU
           LA(1)   6
           DAT0=A  1           set baud rate to 9600

           D0=(5)  IOC
           LA(1)   8
           DAT0=A  1           activate port, set SON

           D0=(5)  CRER
           DAT0=A  1           clear any receive error

           D0=(5)  TCS
           A=0     B
           DAT0=A  1           clear transmit loopback and break

           GOSBVL  =D1=DSKTOP  -->-->string object prolog
           A=DAT1  A
           D1=A                -->string object prolog
           D1=D1+  10          -->string body

           C=R1.F  A           length of buffer in chars
           D=C     A
           B=0     A           number of bytes received
:MAINLOOP  ?D=0    A
           GOYES   :EXIT
           GOSUB   GET1
           GOC     :EXIT
           DAT1=C  B           store char in buffer
           D1=D1+  2
           B=B+1   A
           LA(2)   ENDCHAR
           ?A=C    B
           GOYES   :EXIT
           D=D-1   A
           GOTO    :MAINLOOP

:EXIT      D0=(5)  IOC
           A=0     B
           DAT0=A  1           de-activate port, clear SON

           A=B     A
           R0=A.F  A
           GOSBVL  =PUSH#      does GETPTR, push read length in chars as bint
           GOSBVL  =SAVPTR
           GOVLNG  =GETPTRLOOP


* ----------------------------------
* serial send subroutine
* send 2 :  bytes from R0, low first
* send 1 :  byte from C(B)
* uses A and C, and D0
*
SEND2      C=R0.F  A
           GOSUB   SEND1
           CSR     A
           CSR     A
SEND1      D0=(5)  TCS         send C(B)
           A=DAT0  1           wait until Transmit Buffer not full
           ?ABIT=1 0           test TBF, 1=full
           GOYES   SEND1
           D0=(2)  TBR
           DAT0=C  B           write byte to buffer
:SEND1L    D0=(2)  TCS         wait for byte to be sent and a little more
           A=DAT0  1
           ?ABIT=1 0
           GOYES   :SEND1L
           LA(2)   10
:TDL       A=A-1   B           5 cycles, do a small delay
           GONC    :TDL        10 cycles
           RTN


* --------------------------------------------------------------------
* serial receive subroutine
* get two bytes to C(A) from UART
* reads low order byte first
* GET1 gets only 1 byte to C(B)
* carry set if receive error or timeout
* timeout value must be in R2(A),
*   note that a timeout value of about 0FFFF gives ~3 seconds on an SX
* uses A and C, and D0
*
GET2       C=0     A
           GOSUB   GET1
           RTNC
           CSRC
           CSRC
           GOSUB   GET1
           CSLC
           CSLC
           RTN
GET1       A=0     W
           A=R2.F  A           put timeout counter value in A(M)
           ASL     W
           ASL     W
           ASL     W
           D0=(5)  RCS         wait until Receive Buffer full
:GET1L     A=A-1   M
           GOC     :BADGET     timeout?
           A=DAT0  1
           ?ABIT=0 0           test RBF, 1=full
           GOYES   :GET1L
           D0=(2)  RBR
           C=DAT0  B           read byte from buffer
           D0=(2)  RCS
           A=DAT0  1
           ?ABIT=1 2           test RER, 1=error
           GOYES   :BADGET
           RTNCC
:BADGET    RTNSC

  ENDCODE

  DUP#0= casedrop DROPNULL$      ( *nothing received, return a null string* )
  ONE SWAP SUB$                  ( *chop the buffer* )
 ;
;
