ASSEMBLE

* place a string on the stack to be used as a receive buffer
* (this string can be generated quickly using my blankstr.s subroutine)
* then place a hex string on stack representing a timeout value
* (note that a timeout value of about 0FFFF gives ~3 seconds on an SX)
* this program will receive characters from the serial port at 9600 baud
* and store them in the string buffer until it is full or a time out
* it is ok to allocate a string buffer the size of remaining memory

           NIBASC  /HPHP48-E/

DISPIO     EQU     #100
RBR        EQU     #114
RCS        EQU     #111
CRER       EQU     #113
TBR        EQU     #116
TCS        EQU     #112
IOC        EQU     #110
BAU        EQU     #10D

RPL

::
 CK1NOLASTWD
 CK&DISPATCH1 THREE            ( *1: character string* )
 ::
  DUPLEN$
  CODE
           GOSBVL  =POP#       pop string-length bint off stack
           R3=A.F  A           save size in R3
           GOSBVL  =SAVPTR
           D0=(5)  DISPIO      turn off display
           LA(1)   0
           DAT0=A  1
           LA(5)   #0FFFF      hard coded timeout value, between keychecks
           R2=A.F  A                  about 3 seconds on an SX
           A=0     A           clear serial receive error counter
           R1=A.F  A

           D0=(5)  IOC
           A=0     B
           DAT0=A  1           de-activate port, clear SON

           D0=(5)  BAU
           LA(1)   0
           DAT0=A  1           set baud rate to 1200

           D0=(5)  IOC
           LA(1)   8
           DAT0=A  1           activate port, set SON

           D0=(5)  CRER
           DAT0=A  1           clear any receive error

           D0=(5)  TCS
           A=0     B
           DAT0=A  1           clear transmit loopback and break

           GOSBVL  =D1=DSKTOP  -->-->string object prolog
           A=DAT1  A
           D1=A                -->string object prolog
           D1=D1+  10          -->string body

           C=R3.F  A           length of buffer in chars
           D=C     A
           B=0     A           number of bytes received
MAINLOOP   ?D=0    A
           GOYES   EXIT        buffer full
           GOSUB   GET1        read 1 char from uart to C(B)
           GOC     TESTKEYS    receive error or timeout, exit if key pressed
           DAT1=C  B           store char in buffer
           D1=D1+  2           inc buffer pointer
           B=B+1   A           inc number of chars in buffer
NEXT       D=D-1   A           dec buffer size remaining
           GOTO    MAINLOOP

TESTKEYS   GOSUB   SAVED1BD
           GOSBVL  =POPKEY
           GONC    KEYDN
           GOSUB   RECALLD1BD
           GOTO    MAINLOOP
KEYDN      GOSUB   RECALLD1BD

EXIT       D0=(5)  DISPIO
           LA(1)   8
           DAT0=A  1           turn on display
           D0=(5)  IOC
           A=0     B
           DAT0=A  1           de-activate port, clear SON

           A=B     A           number of chars read
           R0=A.F  A
           GOSBVL  =PUSH2#     does GETPTR, push read length and error counter
           GOSBVL  =SAVPTR               in R0(A) and R1(A)
           GOVLNG  =GETPTRLOOP



SAVED1BD   CD1EX
           GOSBVL  =CSLW5
           C=B     A
           GOSBVL  =CSLW5
           C=D     A
           R4=C.F  W
           RTN

RECALLD1BD C=R4.F  W
           D=C     A
           GOSBVL  =CSRW5
           B=C     A
           GOSBVL  =CSRW5
           D1=C
           RTN



* ----------------------------------
* serial send subroutine
* send 2 :  bytes from R0, low first
* send 1 :  byte from C(B)
* uses A and C, and D0
*
SEND2      C=R0.F  A
           GOSUB   SEND1
           CSR     A
           CSR     A
SEND1      D0=(5)  TCS         send C(B)
           A=DAT0  1           wait until Transmit Buffer not full
           ?ABIT=1 0           test TBF, 1=full
           GOYES   SEND1
           D0=(2)  TBR
           DAT0=C  B           write byte to buffer
SEND1L     D0=(2)  TCS         wait for byte to be sent and a little more
           A=DAT0  1
           ?ABIT=1 0
           GOYES   SEND1L
           LA(2)   10
TDL        A=A-1   B           5 cycles, do a small delay
           GONC    TDL         10 cycles
           RTN


* --------------------------------------------------------------------
* serial receive subroutine
* get two bytes to C(A) from UART
* reads low order byte first
* GET1 gets only 1 byte to C(B)
* carry set if receive error or timeout
* timeout value must be in R2(A),
* R1(A) is counter for serial receive errors, not including timeouts
*   note that a timeout value of about 0FFFF gives ~3 seconds on an SX
* uses A and C, and D0
*
GET2       C=0     A
           GOSUB   GET1
           RTNC
           CSRC
           CSRC
           GOSUB   GET1
           CSLC
           CSLC
           RTN
GET1       A=0     W
           A=R2.F  A           put timeout counter value in A(M)
           ASL     W
           ASL     W
           ASL     W
           D0=(5)  RCS         wait until Receive Buffer full
GET1L      A=A-1   M
           GOC     BADGET      exit if timeout w/o incr error counter
           A=DAT0  1
           ?ABIT=0 0           test RBF, 1=full
           GOYES   GET1L
           D0=(2)  RBR         uart has received one byte
           C=DAT0  B           read byte from uart buffer
           D0=(2)  RCS         receive control register
           A=DAT0  1
           ?ABIT=0 2           test RER, 0=ok, 1=error (0 stop bit or overrun)
           GOYES   OKGET
           D0=(2)  CRER        writing to this addr clears receive error
           DAT0=A  1           clear receive error
           A=R1.F  A           inc error counter in R1(A)
           A=A+1   A
           GOC     BADGET      don't wrap around
           R1=A.F  A
BADGET     RTNSC
OKGET      RTNCC

  ENDCODE
  UNROT
  DUP#0= ITE
  :: DROP DROPNULL$ ;     ( *nothing received, return a null string* )
  :: ONE SWAP SUB$ ;      ( *chop the buffer* )
  SWAP UNCOERCE
  RECLAIMDISP
 ;
;
