ASSEMBLE

* comment out the following line to compile with JAZZ
           NIBASC  /HPHP48-E/

* REC_GPS.S  V1.1
* by Randy Ding   Jan 6, 1996
* compiler used: rplcomp and sasm from HPTOOLS.

* version history:
*   1.0 initial
*   1.1 fixed potential empty buffer problem, not serious, did code cleanup

* This program opens the uart, waits for LEADCHAR to be received,
* then puts incomming into a string buffer until ^M^J pair is received.
* This program takes no arguments from the stack
* and always returns a string, "" if uart errors or timeout occurs.
* Eight retries are attempted.
* On success, the returned string begins with LEADCHAR and ends with ^M^J.
* If buffer fills before termination sequence (^M^J) is read, the returned
* string will not end with the termination sequence.  So if ^M^J does not
* appear as the last two characters, you know a buffer overflow occured.

* Timeout can be adjusted, as well as the max buffer length and baud rate,
* in the first fiew lines of the rpl section, and recompile.
* These values were placed there so that with minimal changes to the source
* the user could have these as parameters taken from the stack
* instead of being hard coded.
* See comments there for length of time elapsed from start to timeout.

* define ascii characters surrounding expected message string
LEADCHAR   EQU     36          '$' use 00 to disable this feature
ENDCHAR1   EQU     13          ^M
ENDCHAR2   EQU     10          ^J

* memory mapped i/o locations for uart control and misc stuff
RBR        EQU     #114
RCS        EQU     #111
CRER       EQU     #113
TBR        EQU     #116
TCS        EQU     #112
IOC        EQU     #110
BAU        EQU     #10D
ANNCTRL2   EQU     #10C

* some stable but un-supported(?) entries
POPHXS     EQU     #53F8D      NOTE: this returns with P <> 0 !

* misc constants
END1SEEN   EQU     0           ST register flag bit for ENDCHAR1 received
MXRETRY    EQU     8           total number of attempts, >= 1

RPL

::
 CK0NOLASTWD

 ( * place a hex string on stack for receive timeout value per retry      * )
 ( * there are 8 retries from timeout or serial receive before giving up  * )
 ( * NOTE: place HXS hex nibs in reverse order, see RPLCOMP.DOC           * )
 ( * timeout value of about 0FFFF gives ~2.5 seconds per retry on a 48SX  * )
 ( * I don't have a GX but expect their timeouts to be twice as fast      * )
 ( * So with this example, an SX times out in about 2.5 * 8 = 20 seconds. * )
 HXS 5 FFFF0

 ( * for BAUDRATE...       * )
 ( *   use 0 for 1200 baud * )
 ( *   use 2 for 2400 baud * )
 ( *   use 4 for 4800 buad * )
 ( *   use 6 for 9600 baud * )
 4

 ( * This code object creates a string of # spaces, stack: # --> $        * )
 ( * The length of this string is also the max chars to skip over looking * )
 ( * for the LEADCHAR, if not seen in that number of chars then "" is rtn * )
 ( * If the LEADCHAR is not seen in buflen chars, no retries are attemp.  * )
 ( * Buf len should be larger than expected string, it gets truncated.    * )
 128
 CODE
           GOSBVL  =SAVPTR
           A=DAT1  A           D1 -->DSKTOP
           D1=A                -->bint prolog on stack top
           A=DAT1  A           verify bint prolog
           LC(5)   =DOBINT
           ?A#C    A
           GOYES   CLEARDONE
           D1=D1+  5           skip bint prolog
           C=DAT1  A           read value of bint on dsktop
           C=C+C   A           * 2 for nibs
           GOSBVL  =MAKE$N     get memory
           A=R0.F  A           -->new string in tempob
           DAT1=A  A           overwrite pointer on top of stack
           D1=A                -->$ prolog
           A=DAT1  A           verify string prolog
           LC(5)   =DOCSTR
           ?A#C    A
           GOYES   CLEARDONE
           D1=D1+  5           skip prolog, -->length field
           C=DAT1  A           read created length in nibs
           C=C-CON A,5         -5 nibs for length field
           GOC     CLEARDONE   something wrong with length field
           SB=0
           CSRB.F  A           div nibs by 2 for # chars
           ?SB=0               test bit shifted off
           GOYES   CLEARLUP0
           GOTO    CLEARDONE   was 1 so something wrong
CLEARLUP0  C=C-1   A           count thru 0
           GOC     CLEARDONE   exit if 0 length
           D1=D1+  5           skip length field, -->body
           LA(2)   32          space
CLEARLUP   DAT1=A  B
           D1=D1+  2
           C=C-1   A
           GONC    CLEARLUP
CLEARDONE  GOVLNG  =GETPTRLOOP
 ENDCODE
 3UNROLL

 ( *   main receive code object, STACK:  $ HXS # --> $ #      * )
 ( *   on entry:                                              * )
 ( *     $ is receive buffer string                           * )
 ( *     HXS is timeout value, hex string 5 nibs max          * )
 ( *     # is baud rate code, see comments above              * )
 ( *   on exit:                                               * )
 ( *     $ is receive buffer, same size as on entry           * )
 ( *     # is number of chars actually received               * )
 CODE
           GOSBVL  =POP#       pop baud rate code bint to A(A)
           R4=A.F  B
           GOSBVL  POPHXS      to A(A), does SAVPTR, returns with P<>0 !
           R2=A.F  A           save timeout HXS in R2(A)
           P=      0           don't forget this after a POPHXS!

           D0=(5)  IOC
           A=0     B
           DAT0=A  1           de-activate port, clear SON

           D0=(5)  BAU
           A=R4.F  B           baudrate code
           ABIT=0  3           code & 7
           DAT0=A  1           set baud rate

           D0=(5)  IOC
           LA(1)   8
           DAT0=A  1           activate port, set SON

           D0=(5)  TCS
           A=0     B
           DAT0=A  1           clear transmit loopback and break

           A=0     A           pre-initialize receive buffer size
           R1=A.F  A
           LA(2)   MXRETRY     max retries on uart errors or timeouts
           R3=A.F  B           keep retries left in R3

RETRY      A=R3.F  B           number of attempts left
           A=A-1   B
           R3=A.F  B
           GOC     BUFRNOTOK   ran out of retries, exit now
           D0=(5)  CRER
           DAT0=A  1           clear any receive error

           GOSBVL  =D1=DSKTOP  -->-->string object prolog
           A=DAT1  A
           D1=A                -->string object prolog
           A=DAT1  A           verify it is a string object prolog
           LC(5)   =DOCSTR
           ?A#C    A
           GOYES   BUFRNOTOK
           D1=D1+  5           -->string length field
           A=DAT1  A           ob nib size, including length field
           A=A-CON A,5         sub nib size of just length field
           GOC     BUFRNOTOK   something wrong with object, exit now
           ?A=0    A
           GOYES   BUFRNOTOK   why is buffer empty? exit now
           SB=0
           ASRB.F  A           div by 2 for number of chars
           ?SB=0               error if buffer is odd nibsize
           GOYES   BUFROK
BUFRNOTOK  GOTO    BADEXIT     something wrong with buffer object, exit now
BUFROK     R1=A.F  A           save string buffer size in R1
           D1=D1+  5           -->string body, the start of actual buffer

           C=R1.F  A           length of buffer in chars
           D=C     A
           D=D-1   A           count thru 0
           GOC     BUFRNOTOK   underflow, why is buffer empty?
           B=0     A           number of bytes received
           ST=0    END1SEEN    flag bit for ENDCHAR1 received

           LA(2)   LEADCHAR    0 means don't wait for any leadchar
           ?A=0    B
           GOYES   MAINLOOP

LEADLOOP   C=R1.F  A           use buffer length as max chars to discard
           ?B>=C   A           here B is chars discarded
           GOYES   LEADLOOP2   LEADCHAR never seen, exit now
           GOSUB   GET1        read 1 char from uart, to C(B)
           GOC     LEADLOOP3   receive error, do a retry
           LA(2)   LEADCHAR    looking for LEADCHAR to begin recording
           ?A#C    B
           GOYES   LEADLOOP1
           B=0     A           found it, reset number bytes received
           GOTO    SAVECHAR
LEADLOOP1  B=B+1   A           isn't LEADCHAR, incr bytes discarded
           GONC    LEADLOOP
LEADLOOP2  GOTO    BADEXIT     just in case of overflow
LEADLOOP3  GOTO    RETRY

MAINLOOP   GOSUB   GET1        read 1 char from uart to C(B)
           GOC     LEADLOOP3   receive error, do a retry
SAVECHAR   DAT1=C  B           store char in buffer
           D1=D1+  2           2 nibs per char
           B=B+1   A           num chars received
           ?ST=0   END1SEEN    has ENDCHAR1 been received yet?
           GOYES   TEST1       no, go check if this is ENDCHAR1?
           LA(2)   ENDCHAR2    ENDCHAR1 was received, is this ENDCHAR2?
           ?A=C    B
           GOYES   EXIT        yes, both chars received so exit now
           ST=0    END1SEEN    no, not ENDCHAR2 so clear flag and start over
TEST1      LA(2)   ENDCHAR1
           ?A#C    B           is this ENDCHAR1?
           GOYES   NEXT
           ST=1    END1SEEN    yes, set flag for ENDCHAR1 received
NEXT       D=D-1   A           decr space remaining in buffer
           GONC    MAINLOOP    count thru 0
           GOC     EXIT        buffer full

BADEXIT    B=0     A           clear number of chars received to return ""

EXIT       D0=(5)  IOC
           A=0     B
           DAT0=A  1           de-activate port, clear SON

           A=B     A           chars received
           C=R1.F  A           size of receive string buffer
           ?A<=C   A           is this larger than our buffer?
           GOYES   EXIT1
           A=C     A           allow no larger than buffer size
EXIT1      GOVLNG  =PUSH#ALOOP   does GETPTR, push A(A) as #, Loop


* ----------------------------------
* serial send subroutine
* send 2 :  bytes from R0, low first
* send 1 :  byte from C(B)
* uses A and C, and D0
*
*SEND2      C=R0.F  A
*           GOSUB   SEND1
*           CSR     A
*           CSR     A
*SEND1      D0=(5)  TCS         send C(B)
*           A=DAT0  1           wait until Transmit Buffer not full
*           ?ABIT=1 0           test TBF, 1=full
*           GOYES   SEND1
*           D0=(2)  TBR
*           DAT0=C  B           write byte to buffer
*SEND1L     D0=(2)  TCS         wait for byte to be sent and a little more
*           A=DAT0  1
*           ?ABIT=1 0
*           GOYES   SEND1L
*           LA(2)   10
*TDL        A=A-1   B           5 cycles, do a small delay
*           GONC    TDL         10 cycles
*           RTN


* --------------------------------------------------------------------
* serial receive subroutine
* get two bytes to C(A) from UART
* reads low order byte first
* GET1 gets only 1 byte to C(B)
* carry set if receive error or timeout
* timeout value must be in R2(A),
*   note that a timeout value of about 0FFFF gives ~2.5 seconds on an SX
* uses A and C, and D0
*
*GET2       C=0     A
*           GOSUB   GET1
*           RTNC
*           CSRC
*           CSRC
*           GOSUB   GET1
*           CSLC
*           CSLC
*           RTN
GET1       A=0     W
           A=R2.F  A           put timeout counter value in A(M)
           ASL     W
           ASL     W
           ASL     W
           D0=(5)  RCS         wait until Receive Buffer full
GET1L      A=A-1   M
           GOC     BADGET      timeout?
           A=DAT0  1
           ?ABIT=0 0           test RBF, 1=full
           GOYES   GET1L
           D0=(2)  RBR
           C=DAT0  B           read byte from buffer
           D0=(2)  RCS
           A=DAT0  1
           ?ABIT=1 2           test RER, 1=error
           GOYES   BADGET
           RTNCC
BADGET     RTNSC
 ENDCODE

 DUP#0= casedrop DROPNULL$     ( *nothing received, return a null string* )
 ONESWAP SUB$                  ( *chop the buffer* )
;
