
* I imagined I had a working GCD algorithm using distributed
* representation so I collected all the code handling distributed
* polynomials into this file. The GCD algorithm fails for many
* multivariate polynomials when an equal leading monomial gets
* created in the calculation - thus causing an overflow in time
* or possibly even an eternal loop. PTHYI.

ASSEMBLE

**********************************************************************
* Name:		DISTDIV?
* Desc:		Divides two multivariate polynomials, returning the
*		quotient and remainders.
* Stack:	( P1 P2 --> quo rem TRUE / P1 P2 FALSE )
* Notes:	Handles both P1=0 and P2=0 properly.
*		Both inputs must be polynomials.
**********************************************************************

DISTDIV?	ST=0	sGARB		No GC yet

DDretry		GOSBVL	=SAVPTR		Re-entry point after GC

* Before trying to allocate any work memory we first check that
* all the coefficients are indeed short. This way we can avoid an
* unnecessary data allocation.

		GOSUBL	QSHORT2?	CS if both short
		GOC	+

* Overflow occurred - return original polynomials and FALSE flag

DDoverflow	P=	0
		GOVLNG	=GPPushFLoop	( --> P1 P2 FALSE )


* Allocate working space. We require 1000 bytes since my instinct is that
* avoiding an imminent GC won't win as much as avoiding running out of
* memory during the division subroutine - possibly once we're almost
* finished. My guess is that going below 500 bytes would often prolong
* running time for the above reason.


+		GOSBVL	=ROOM
		A=C	A		A[A] = room
		LC(5)	2000		2000 nibbles required
		?A>C	A
		GOYES	+
DDmemerr	GOSBVL	=DOGARBAGE	Handles sGARB properly
		GOTO	DDretry

+		CSR	A		2000/16 = 125 (left for SOL)
		C=A-C	A
		R4=C	A		R4[A] = free work memory
		GOSBVL	=MAKE$N		Use $ for the benefit of DDmemerr

* Count the number of variables to be used in the distributed representations

		GOSUBL	QVARS2
		R3=A	A		R3[A]=vars

* Now:
*	R0[A] = ->$	= ->denom-10
*	R3[A] = vars
*	R4[A] = free


* Now we convert the polynomials into distributed form.
* The format used is that required by the division subroutine,
* namely denominator first, then one slot free from which
* the quotient grows, then nominator.

		A=DAT1	A
		D0=A			->P2
		C=R0	A
		D1=C
		D1=D1+	10		->output for denominator
		C=R3	A
		B=C	A		B[A] = vars
		GOSUBL	Distribute	D1 = ->output tail
		GOC	DDmemerr	CS: Insufficient memory

* Allocate the extra slot

		GOSUBL	EXTRASLOT
		GOC	DDmemerr	CS: Insufficient memory
		R1=A	A		R1[A] = ->nominator

* And convert the nominator

		GOSBVL	=D0=DSKTOP
		D0=D0+	5
		A=DAT0	A		->P1
		D0=A
		GOSUBL	Distribute	D1 = ->output tail
		GOC	DDmemerr	CS: Insufficient memory
		CD1EX
		R2=C	A		R2[A] = ->nominator tail

* Call the division subroutine

		GOSUB	DDIVIDE
		GONC	+		CC: Success

* Handle error in division

		?P=	0
		GOYES	DDmemerr	P==0 ==> insufficient memory
		GOTO	DDoverflow	P<>0 ==> overflow

* Division done - convert results into recursive representation

* First move the quotient and remainder to end of work memory

+		A=R0	A		A[A] = ->$
		A=A+CON	A,5
		D0=A
		C=DAT0	A
		A=A+C	A
		D1=A			D1 = ->$tail
		D0=D0+	5		D0 = ->denominator

--		A=DAT0	W		A[W] = denominator coeff
		D0=D0+	16
		?A=0	W
		GOYES	+		Denominator skipped
		C=R3	A		C[A] = vars
-		D0=D0+	5		Skip power in slot
		C=C-1	A
		?C#0	A
		GOYES	-		Loop until all powers skipped
		GONC	--		Loop until end of denominator
+
		AD0EX			D0 = ->quotient
		C=R2	A		C[A] = ->nominator tail
		D0=C
		C=C-A	A		C[A] = nibbles in quo & nom
		GOSBVL	=MOVEUP		D1 = ->quotient (new location)

* Now do the conversions - output quotient to start of $

		A=R0	A		A[A] = ->$
		A=A+CON	A,10		A[A] = ->Pquo (output)
		CD1EX
		D0=C			D0 = ->input (quotient)
		D1=A			D1 = ->output (quotient)
		C=C-A	A
		R4=C	A		R4[A] = free memory (new)

		C=R3	A
		B=C	A		B[A] = vars
		GOSUBL	Recurse		D1 = ->Pquo tail
		GOC	DDrecmemerr	CS: Insufficient memory
		CD1EX
		R2=C	A		R2[A] = ->Prem
		CD1EX
		AD0EX			Skip the "powers" in the terminator
		A=A+B	A		slot of the quotient.
		A=A+B	A
		A=A+B	A
		A=A+B	A
		A=A+B	A
		AD0EX
		GOSUBL	Recurse		D1 = ->Prem tail
		GOC	DDrecmemerr	CS: Insufficient memory

* Shrink work memory

		AD1EX
		D0=A
		GOSBVL	=Shrink$	Uses R1[A] - thus R2[A] used above

* And return the results

		GOSBVL	=GETPTR
		D1=D1+	5
		A=R0	A
		A=A+CON	A,10		A[A] = ->Pquo
		DAT1=A	A		Overwrite P1
		D1=D1-	5
		A=R2	A		A[A] = ->Prem
		DAT1=A	A		Overwrite P2
		GOVLNG	=PushTLoop	( --> Pquo Prem TRUE )

* Insufficient memory in conversion to recursive form. Could optimize
* this to avoid a full rerun and just try the conversion again - but
* it does not seem to be worth the effort.

DDrecmemerr	GOTO	DDmemerr	Just try full restart


**********************************************************************
* Name:		DISTGCD?
* Desc:		GCD of two multivariate polynomials.
* Stack:	( P1 P2 --> gcd TRUE / P1 P2 FALSE )
* Notes:	Both inputs must be polynomials and integer content
*		has been extracted.
**********************************************************************

sLEX	EQU 8		Status flag to indicate lex order detected
sLEX1	EQU 9		Status flag to indicate the order

DISTGCD?	ST=0	sGARB		No GCD yet

DGretry		GOSBVL	=SAVPTR		Re-entry point after GC

* Before trying to allocate any work memory we first check that
* all the coefficients are short as required by the implementation.
* This way we can avoid an unnecessary data allocation.

		GOSUBL	QSHORT2?	CC if both not short
		GOC	+		CS: Continue

* Overflow occurred - return original polynomials and FALSE flag

DGoverflow	P=	0
		GOVLNG	=GPPushFLoop	( --> P1 P2 FALSE )

* Allocate working space. We require 1000 bytes since my guess is that
* avoiding an imminent GC won't win as much as avoiding running out of
* memory during the GCD subroutine - possibly once we're almost finished.

+		GOSBVL	=ROOM
		A=C	A		A[A] = room
		LC(5)	2000		2000 nibbles required
		?A>C	A
		GOYES	+

DGmemerr	GOSBVL	=DOGARBAGE	Handles sGARB properly
		GOTO	DGretry

+		CSR	A		2000/16 = 125 (left for SOL)
		C=A-C	A
		R4=C	A		R4[A] = free work memory
		GOSBVL	=MAKE$N		Use $ for the benefit of DGmemerr

		A=R0	A
		GOSBVL	=ASLW5
		A=R0.F	A
		A=A+CON	A,10		->slot1
		R0=A


* Count the number of variables to be used in the distributed representations

		GOSUB	QVARS2
		R3=A	A		R3[A] = vars

* Now:
*	R0[A] = ->slot1		R0[A1] = ->$	
*	R3[A] = vars
*	R4[A] = free

* Now we convert the polynomials into distributed form.
* The format used is that required by the gcd subroutine,
* namely each polynomial will be preceded by an extra slot
* to hold the multiplication coefficients in the S-Poly
* calculation.

* Allocate extra slot 1

		A=R0	A
		D1=A			->slot1
		GOSUB	EXTRASLOT
DGmemerr2	GOC	DGmemerr	CS: Insufficient memory

* Convert polynomial 1

		GOSBVL	=D0=DSKTOP
		D0=D0+	5
		A=DAT0	A
		D0=A			->P1
		GOSUBL	Distribute	D1 = ->output tail
		GOC	DGmemerr	CS: Insufficient memory
		CD1EX
		R1=C	A		R1[A] = ->slot2
		CD1EX

* Allocate extra slot 2

		GOSUB	EXTRASLOT
		GOC	DGmemerr	CS: Insufficient memory

* Convert polynomial 2

		GOSBVL	=D0=DSKTOP
		A=DAT0	A
		D0=A			->P2
		GOSUBL	Distribute	D1 = ->output tail
		GOC	DGmemerr2	CS: Insufficient memory
		CD1EX
		R2=C	A		R2[A] = ->slot3 (output)

* The actual GCD loop consists of repeated S-poly calculations
* until S-poly is zero - in which case the lexicographically
* smaller polynomial is the GCD. We always remove the integer
* content of the S-poly to avoid overflows.

distgcdloop	GOSUB	S-POLY
		GOC	DGmemerr2	CS: Out of memory
		CR2EX.F	A		R2[A]=->slot3 tail
		D0=C			D0 = ->slot3
		GOSUB	DISTPPZ		D0 = ->slot3
		GOC	distgcdok	CS: s-poly is zero - we're done
		CD0EX
		CR2EX.F	A		R2[A]=->slot3	C[A]=->slot3 tail

* Now if poly1 was lexicographically greater then we move down
* poly2 and s-poly to replace the old polynomials. We don't bother
* to check if they actually had equal powers in the leading terms
* - doesn't make any difference since we can then pick either one to be
* replaced.

		?ST=0	sLEX1
		GOYES	+
		RSTK=C			RSTK = ->slot3tail
		A=R0	A		->slot1
		D1=A			is the target
		A=R1	A		->slot2
		D0=A			is the source
		C=R2	A		->slot3
		C=C-A	A		->slot2 - ->slot3 moves down
		GOSBVL	=MOVEDOWN	to overwrite old slot1
		CD1EX
		R1=C	A		->slot2 (new target)
		C=RSTK			C[A] = ->slot3tail

* We move down s-poly separately - easier this way since slot3 has
* no preceding empty slot.

+		A=C	A		A[A] = ->slot3tail
		GOSUBL	SLOTSIZE	B[A] = slotsize
		C=R1	A		->slot2
		D1=C			is the target
		C=R2	A		->slot3
		C=C-B	A		Skip back to imaginary empty slot
		D0=C			which is the source
		C=A-C	A		size(slot3) + empty slot
		GOSBVL	=MOVEDOWN
		CD1EX
		R2=C	A		->slot3 (new one)

* Recalculate free memory to reflect new polynomials in use.
* The free memory is the area between ->slot3 and the end of the
* work string.

		A=R0	W
		GOSBVL	=ASRW5
		A=A+CON	A,5		->$len
		D0=A
		C=DAT0	A
		A=A+C	A		->$tail
		C=R2	A		->slot3
		A=A-C	A
		R4=A	A		free memory = ->$tail - ->slot3
		GOTO	distgcdloop


* S-poly is now zero - the lexicographically smaller polynomial is
* the gcd to be returned.

* For the conversion we move the gcd to the end of the work memory
* and then convert it to the start of the work memory

distgcdok	A=R0	W
		GOSBVL	=ASRW5		A[A] = ->$
		A=A+CON	A,5
		D0=A
		C=DAT0	A
		A=A+C	A
		D1=A			->$tail
		GOSUB	SLOTSIZE	B[A] = slotsize
		A=R0	A
		A=A+B	A		->poly1
		C=R1	A		->poly1tail
		?ST#1	sLEX1
		GOYES	+		poly1 wasn't greater - ok
		A=C	A
		A=A+B	A		->poly2
		C=R2	A		->poly2tail
+		D0=C			->source
		C=C-A	A
		GOSBVL	=MOVEUP		D1 = ->gcd
		CD1EX
		D0=C
		GOSUBL	DISTPPZ		Remove integer content
		GOSUBL	DISTABS		Force leading coefficient positive
		A=R0	A		->slot1 = ->$ + 10
		D1=A			D1 = ->output
		CD0EX
		D0=C			D0 = ->gcd
		C=C-A	A
		R4=C	A		R4[A] = free memory
		C=R3	A
		B=C	A		B[A]=vars
		GOSUBL	Recurse		D1 = ->Pgcd tail
		GONC	+
		GOTO	DGmemerr	CS: Insufficient memory

* Shrink work memory

+		AD1EX
		D0=A			->Pgcd tail
		C=R0	A
		GOSBVL	=CSRW5
		R0=C			->$
		GOSBVL	=Shrink$

* And return the result

		GOSBVL	=GETPTR
		D1=D1+	5
		A=R0	A
		A=A+CON	A,10		A[A] = ->Pgcd
		DAT1=A	A
		D1=D1-	5
		GOVLNG	=OverWrTLoop	( --> Pgcd TRUE )

*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
* Name:		DDIVIDE
* Desc:		Divide two polynomials in recursive representation.
*
* Entry:	R0[A] = ->$	= ->denominator-10
*		extra slot between these two
*		R1[A] = ->nominator
*		R2[A] = ->nominator tail
*		R3[A] = vars
*		R4[A] = free memory
*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

DDIVIDE		A=R1	A
		D0=A			D0 = ->nominator
		C=R0	A
		D1=C
		D1=D1+	10		D1 = ->denominator
		A=DAT1	W		A[W] = lcoeff(denom)
		?A=0	W
		GOYES	ddiverror	Division by zero!
		D0=D0+	16		->nompower1
		D1=D1+	16		->denpower1

* First divide the powers. Note that the quotient slot is right before
* the nominator slot being operated with.

		GOSUB	SLOTSIZE	B[A]=slotsize	C[A]=vars
		D=C	A		D[A] = vars

--		A=DAT0	A		A[A] = nompower
		C=DAT1	A		C[A] = denpower
		A=A-C	A		A[A] = quopower
		GOC	ddivfinish	Overflow - division at end!
		CD0EX
		C=C-B	A		->quopower
		D0=C
		DAT0=A	A		Store power in quotient
		C=C+B	A		->nompower
		D0=C		
		D0=D0+	5		->nompower++
		D1=D1+	5		->denpower++
		D=D-1	A		vars--
		?D#0	A
		GOYES	--		Until all powers divided
		GONC	+		Continue to handle the coefficient

* Overflow in power division means we're done

ddivfinish	A=R1	A		A[A] = ->nominator
		A=A-B	A
		D0=A
		A=0	W		Terminate the quotient
ddivdone	DAT0=A	W
		RTNCC			CC: Successful division

* Division by zero or inexact coefficient division is an error

ddiverror	LC(5)	#304
		GOVLNG	=GPErrjmpC	"Undefined Result"

* Powers divided - handle the coefficient

+		A=R1	A
		D0=A			D0 = ->nominator
		C=R0	A
		D1=C
		D1=D1+	10		D1 = ->denominator
		A=DAT0	W		A[W] = nomcoeff
		C=DAT1	W		C[W] = dencoeff
		AD0EX
		A=A-B	A		->quocoeff
		AD0EX
		?A=0	W		If nominator is zero then we're done!
		GOYES	ddivdone
		GOSUBL	ZSDIV		A[W]=quotient	B[W]=remainder
		?B#0	W
		GOYES	ddiverror	remainder<>0 ==> error
		DAT0=A	W		Store quocoeff

* Now calculate the new "nominator" = nom-quocoeff*den. The easy way to
* handle the memory limitations is to move the nominator to the end
* of the work space and output the new one into it's proper location.
* This way we also free memory along the way as the coefficients of
* the old nominator become unneeded.

		A=R0	A		A[A] = ->$
		A=A+CON	A,5
		D1=A
		C=DAT1	A
		A=A+C	A
		D1=A			D1 = ->$tail
		A=R2	A
		D0=A			D0 = ->nominator tail
		C=R1	A		C[A] = ->nominator
		C=A-C	A		C[A] = nibbles in nominator
		GOSBVL	=MOVEUP		D1 = ->nominator (new location)

		GOSUB	SLOTSIZE	B[A]=slotsize	C[A]=vars

* New nominator moves one slot to higher memory, but since we already
* know we'll just skip the first coefficient in the old nominator
* we also free up one slot - thus no need to update free memory in here.

		CD0EX			Skip first slot in nominator (new)
		C=C+B	A		   to make room for the next quotient
		D0=C			   coefficient
		R1=C	A		R1[A] = ->nominator (new)
		C=C-B	A
		C=C-B	A
		D=C	A		D[A] = ->quotient coefficient
		A=R0	A
		A=A+CON	A,10		A[A] = ->denominator
		A=A+B	A		Skip first slot in denominator
		AD1EX
		A=A+B	A		Skip first slot in nominator
		AD1EX


* Now:	D0   = ->nominator' (output)		(advances in loop)
*	D1   = ->nominator  (input)		(advances in loop)
*	D[A] = ->quotient coefficient
*	A[A] = ->denominator (input)		(advances in loop)

ddremloop	P=	0		Reset P from previous pass

		AD1EX
		C=DAT1	W		C[W] = nomcoeff
		AD1EX
		?C#0	W
		GOYES	+
		P=	1		P=1 : denominator traversed
+		C=DAT1	W		C[W] = dencoeff
		?C#0	W
		GOYES	++		nominator not traversed - continue
		?P#	1
		GOYES	+		P=2 : nominator traversed

* Both nominator and denominator traversed - the sum is done

		P=	0
		DAT0=C	W		Terminate the new nominator
		D0=D0+	16
		CD0EX
		R2=C	A		R2[A] = ->nominator tail
		GOTO	DDIVIDE

+		P=	2		P=2: nominator traversed

* Now output one new nominator term at a time

++		A=A+CON	A,16		->denpower
		D=D+CON	A,16		->quopower
		D0=D0+	16		->outpower
		D1=D1+	16		->nompower
		C=R3	A
ddnompowloop	RSTK=C			RSTK = vars counter
		AD0EX
		C=DAT0	A
		B=C	A		B[A] = denpower
		AD0EX
		CDEX	A		Swap C[A] and D1
		CD0EX
		CDEX	A
		C=DAT0	A		C[A] = quopower
		CDEX	A
		CD0EX			Swap C[A] and D1 back
		CDEX	A
		B=B+C	A		B[A] = denpower + quopower
		C=DAT1	A		C[A] = nompower
		
* Handle the exceptions

		?P=	1		If P=1 then denominator traversed
		GOYES	+		Then just use nompower
		CBEX	A		C[A] = denpower + quopower
		?P=	2		If P=2 then nominator traversed
		GOYES	+		Then just use denpower + quopower
		?B=C	A		Else if same power then
		GOYES	+		it certainly is valid for output
		P=	2		Else assume nominator has higher power
		?C>B	A		
		GOYES	+		It has
		P=	1		It hasn't - output quo*den
		CBEX	A		Fix the power to be output
+		DAT0=C	A		Output the power
		A=A+CON	A,5		->denpower++
		D=D+CON	A,5		->quopower++
		D0=D0+	5		->outpower++
		D1=D1+	5		->nompower++

		C=RSTK
		C=C-1	A		C[A] = vars--
		?C#0	A
		GOYES	ddnompowloop	Loop until all powers handled

* Now skip back to the start of the slot

		GOSUB	SLOTSIZE	B[A]=slotsize	C[A]=vars
		A=A-B	A		A[A] = ->denominator
		C=D	A
		C=C-B	A
		D=C	A		D[A] = ->quotient
		CD0EX
		C=C-B	A
		CD0EX			D0 = ->nominator (output)
		CD1EX
		C=C-B	A
		CD1EX			D1 = ->nominator (input)
		
* And calculate nom-quo*den ignoring the proper parts if P<>0

		?P=	1		If denominator traversed
		GOYES	+		then ignore -quo*den
		C=A	A
		RSTK=C			RSTK = ->denominator
		CD0EX
		A=DAT0	W		A[W] = denominator
		CDEX	A
		D0=C
		C=DAT0	W		C[W] = quotient
		GOSUBL	ZSMUL		A[W] = quo*den
		A=-A-1	S
		B=A	W		B[W] = -quo*den
		C=D	A
		CD0EX			D0 = ->nominator (output)
		D=C	A		D[A] = ->quotient
		C=RSTK
		A=C	A		A[A] = ->denominator

+		ABEX	W		A[W] = -quo*den (if calculated)
		?P=	2		If nominator traversed
		GOYES	+		then ignore +nom
		C=DAT1	W
		ACEX	W		A[W]=nom	C[W]=-quo*den
		?P#	0
		GOYES	+		Ignore addition if only nom
		GOSUBL	ZSADD		A[W] = nom-quo*den
+		C=A	W		C[W] = nom-quo*den
		A=B	A		A[A] = ->denominator
		DAT0=C	W		Output the coefficient

		GOSUB	SLOTSIZE	B[A]=slotsize	C[A]=vars

* Advance the pointers for next loop

		C=DAT0	W		Don't advance nominator (output)
		?C#0	W		if the result was zero!
		GOYES	+
		CR4EX
		C=C+B	A		We won some free memory
		CR4EX		
		GONC	++
+		CD0EX
		C=C+B	A		->nominator++ (output)
		CD0EX
++		?P=	2		Don't advance nominator (input)
		GOYES	+		if nominator already traversed
		CD1EX
		C=C+B	A		->nominator++ (input)
		CD1EX
+		?P=	1		Don't advance denominator
		GOYES	+		if denominator already traversed
		A=A+B	A		->denominator++
+		GOTO	ddremloop

*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
* Name:		S-POLY
* Desc:		Calculate S-polynomial of two polynomials
*		in distributed representation.
*
* Entry:	R0[A] = ->slot1
*		R1[A] = ->slot2
*		R2[A] = ->slot3
*		R3[A] = vars
*		R4[A] = free memory
*
* Exit:		CS:	insufficient memory
*		CC:	slot3 contains the S-polynomial
*			C[A] = ->slot3 tail
*			sLEX is set if the leading term of either
*			polynomial is lexocographically greater
*			than the other one. If so then sLEX1 indicates
*			if slot1 was the greater one.
*
* Alters:	A[W] B[W] C[W] D[W] sLEX sLEX1
*
* Stack:	4 (SHORTCANNO call)
*
* Notes:
*		Both slot1 and slot2 contain the input polynomials
*		preceded by an empty coefficient slot. The coefficient
*		slots will be used to hold the coefficients used
*		in the substraction of the two polynomials.
*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

S-POLY

* First calculate the integer coefficients to be used
* The actual formula uses LCM of the leading coefficients
* divided by the leading coefficients, in our case using
* rational simplification and then swapping the resulting
* coefficients will give the same result with simpler code.

		GOSUB	SLOTSIZE	B[A]=slotsize	C[A]=vars
		A=R0	A
		A=A+B	A
		D0=A			->poly1
		A=R1	A
		A=A+B	A
		D1=A			->poly2
		A=DAT0	W		A[W] = coeff1
		C=DAT1	W		C[W] = coeff2
		GOSUBL	ZSCANNO		A[W] = c1	C[W] = c2
		A=-A-1	S		Negate the coefficient for poly2
		B=C	W		B[W] = c2
		C=R1	A		->slot2
		CD0EX
		DAT0=A	W		Store -c1 as coefficient for poly2
		D0=C
		A=B	W		A[W] = c2
		C=R0	A
		CD0EX
		DAT0=A	W		Store c2 as coefficient for poly1
		D0=C

* Calculate the powers of the variables in the coefficients
* Here the LCM calculation becomes simpler, LCM is simply the
* greater power and the division becomes a substraction. Most
* notably the coefficient with the larger power will always
* get a zero power as the multiplication coefficient.

* We also detect which polynomial is greater in lexicographic ordering
* here for the benefit of the calling code. For example the GCD
* calculation can use it to determine which of the polynomials to
* replace with the S-poly in the GCD loop.

	
		GOSUB	SLOTSIZE	B[A]=slotsize	C[A]=vars
		D=C	A		D[A]=vars
		D0=D0+	16		->powers1
		D1=D1+	16		->powers2
		ST=0	sLEX		No order detected yet
spolycpowlp	A=DAT0	A		A[A] = pow1
		C=DAT1	A		C[A] = pow2

		?ST=1	sLEX		If order already known then
		GOYES	+		just calculate the powers
		?A=C	A		Else if equal powers then
		GOYES	+		we still don't know the order
		ST=1	sLEX		Else we do
		ST=1	sLEX1		Assume pow1 is greater
		?A>C	A
		GOYES	+		It was
		ST=0	sLEX1		pow2 is greater
+
		?A>=C	A		Calculate the power coefficients
		GOYES	+
		C=C-A	A		C[A] = pow2-pow1
		A=0	A		A[A] = 0
		GONC	++
+		A=A-C	A		A[A] = pow1-pow2
		C=0	A		C[A] = 0
++
		AD0EX			Skip back to the coefficient slot
		A=A-B	A
		AD1EX
		A=A-B	A
		AD0EX
		DAT0=A	A		Store the power coefficients
		DAT1=C	A
		AD0EX			And skip back to the leading terms
		A=A+B	A
		AD1EX
		A=A+B	A
		AD0EX
		D0=D0+	5		Next variables
		D1=D1+	5
		D=D-1	A		variables--
		?D#0	A
		GOYES	spolycpowlp	Calculate for all variables

* Now we add the polynomials together out to slot 3. Each polynomial
* is multiplied with the coefficient slot preceding it in the addition.
* Note that the first coefficient from both polynomials has now been
* skipped - they're not needed as they are known to cancel out.

		C=R2	A
		D=C	A		D[A] = ->output

spolyaddlp

		P=	0		Assume both still non-empty

* Allocate slot for the next output coefficient

		C=R3	A		C[A] = vars
		CBEX	A		B[A] = vars	C[A] = slotsize
		A=R4	A
		A=A-C	A
		RTNC			CS: Insufficient memory
		R4=A	A

* Check if one of the input polynomials has terminated already

		A=DAT0	W		A[W] = coeff1
		C=DAT1	W		C[W] = coeff2
		D0=D0+	16
		D1=D1+	16
		?A#0	W
		GOYES	+		coeff1<>0 - still work to do
		P=	1		Flag poly1 exhausted
		?C#0	W
		GOYES	++		Continue of coeff2<>0 
		P=	0		Else we're done
		C=D	A
		CD1EX
		DAT1=A	W		Terminate slot3
		D1=D1+	16
		CD1EX			C[A] = ->slot3 tail
		A=R4	A		Fix free memory back
		A=A+B	A
		B=B+B	A
		B=B+B	A
		A=A+B	A
		R4=A	A
		RTNCC			CC: Successful calculation

+		?C#0	W
		GOYES	++		Both polynomials nonzero
		P=	2		Flag poly2 exhausted
++

* Calculate the powers of the output coefficient first so we can
* detect if the powers from the two polynomials are not equal and
* thus no addition will actually be done.

		D=D+CON	A,16		->output powers
		A=R0	A
		A=A+CON	A,16		->slot1 powers
		R0=A	A
		A=R1	A
		A=A+CON	A,16		->slot2 powers
		R1=A	A

spolyaddpows	?P=	1		If poly1 is exhausted then
		GOYES	+		Ignore the powers from it
		C=R0	A		->slot1
		CD0EX
		A=DAT0	A		pow1
		D0=C
		C=DAT0	A		coeffpow1
		C=C+A	A		C[A] = productpow1

+		?P=	2		If poly2 is exhausted then
		GOYES	+		Ignore the powers from it
		RSTK=C
		C=R1	A
		CD1EX
		A=DAT1	A		pow2
		D1=C
		C=DAT1	A		coeffpo2
		A=A+C	A		A[A] = productpow2
		C=RSTK

+		?P#	0		If either one is ignored then
		GOYES	+		Ignore equality test
		?A=C	A
		GOYES	++		Equal - continue as before
		P=	1		Assume productpow2 is greater
		?A>C	A
		GOYES	++		It was - continue
		P=	2		Wasn't - flag it
+		?P=	1
		GOYES	++
		A=C	A		Get greater power to A[A]
++		C=D	A
		CD1EX
		DAT1=A	A		And output it
		D1=C
		D0=D0+	5		Next variable
		D1=D1+	5
		D=D+CON	A,5
		A=R0	A
		A=A+CON	A,5
		R0=A
		A=R1	A
		A=A+CON	A,5
		R1=A
		B=B-1	A		variables--
		?B#0	A
		GOYES	spolyaddpows	Loop until powers added

* Now calculate the output coefficient as indicated by the flag in P

		GOSUB	SLOTSIZE	B[A]=slotsize	C[A]=vars
		C=D	A		Back to start of output slot
		C=C-B	A
		D=C	A		
		CD0EX			Back to start of poly1 slot
		C=C-B	A
		CD0EX
		CD1EX			Back to start of poly2 slot
		C=C-B	A
		CD1EX
		C=R0	A		Back to ->slot1 start
		C=C-B	A
		R0=C	A
		C=R1	A		Back to ->slot2 start
		C=C-B	A
		R1=C	A

		A=0	W		Assume zero coefficient from poly1
		?P=	1		Zero is right when poly1 is exhausted
		GOYES	+		Yep - skip the calculation for poly1
		C=R0	A		->slot1
		CD0EX
		A=DAT0	W		coeff1
		D0=C
		C=DAT0	W		polycoeff1
		GOSUBL	ZSMUL		A[W] = coefficient from poly1
+
		C=D	A		->output coefficient
		CD1EX
		DAT1=A	W		Output the intermediate result
		D1=C
		
		A=0	W		Assume zero coefficient from poly2
		?P=	2		Zero is right when poly2 is exhausted
		GOYES	+		Yep - skip the calculation for poly2
		C=R1	A		->slot2
		CD1EX
		A=DAT1	W		coeff2
		D1=C
		C=DAT1	W		polycoeff2
		GOSUBL	ZSMUL		A[W] = coefficient from poly2

+		C=D	A		->output coefficient
		CD1EX
		D=C	A
		C=DAT1	W		C[W] = coefficient from poly1
		GOSUBL	ZSADD		A[W] = final coefficient
		DAT1=A	W		Store it
		C=D	A
		CD1EX
		D=C	A		->output coefficient

* Now prepare the calculation for the next term properly

		GOSUB	SLOTSIZE	B[A]=slotsize	C[A]=vars

		?A#0	W
		GOYES	+
		A=R4	A		The terms canceled - fix memory
		A=A+B	A		back
		R4=A	A
		GONC	++
+		C=D	A		Skip to next output slot
		C=C+B	A
		D=C	A
++
		?P=	1		If poly1 exhausted then
		GOYES	+		don't skip poly1 terminator
		AD0EX
		A=A+B	A
		AD0EX
+
		?P=	2		If poly2 exhausted then
		GOYES	+		don't skip poly2 terminator
		AD1EX
		A=A+B	A
		AD1EX
+
		GOTO	spolyaddlp	Loop until both polys exhausted

*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
* Name:		DISTPPZ
* Entry:	D0	->poly
*		R3[A]	vars
* Exit:		D0	->poly'
*		A[W]	gcd of all the coefficients
*		CS if polynomial is zero!
* Alters:	A[W] B[A] C[W] D[A]
* Stack:	2
*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

DISTPPZ		CD0EX
		B=C	A		B[A] = ->poly
		CD0EX
		A=0	W		Init gcd = 0
--		C=0	S
		C=DAT0	15		C[W] = |coeff|
		?C=0	W
		GOYES	++
		GOSUBL	SHORTGCD	A[W] = gcd'
		A=A-1	A		Use [A] for speed & size
		?A=0	W
		GOYES	+		Quick exit if gcd=1
		A=A+1	A
		GOSUB	NEXTSLOT1	D0 = ->next poly slot
		GONC	--		Continue until gcd=1 or terminator

* Special exit when early detection of gcd=1

+		A=A+1	A		gcd=1
		RTNCC			CC: nonzero polynomial though

* gcd not 1 - handle it

++		C=B	A
		D0=C
		?A=0	W		If gcd=0 then polynomial was zero!
		RTNYES			CS: zero polynomial
		D=C	A		D[A] = ->poly
		C=A	W		C[W] = gcd
--		A=0	S		Divide the coefficient by the gcd
		A=DAT0	15		The sign remains the same..
		?A=0	W
		GOYES	+
		GOSUBL	BDIV
		DAT0=A	15
		GOSUB	NEXTSLOT1	D0 = ->next poly slot
		GONC	--
+		A=C	W		A[W] = gcd
		C=D	A
		D0=C			->poly
		RTNCC			CC: nonzero polynomial

* Help subroutine to skip a slot in D0 when we don't have slotsize handy

NEXTSLOT1	RSTK=C			Save C[A] so it won't be modified
		C=R3.F	A		C[A]=vars
		AD0EX
		A=A+C	A
		C=C+C	A
		C=C+C	A
		A=A+C	A		skipped 5*vars
		AD0EX
		C=RSTK			Restore C[A]
		D0=D0+	16		skip extra 16 for the coefficient
		RTNCC

*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
* Name:		DISTABS
* Desc:		Force leading coefficient of polynomial to be positive
* Entry:	D0	->poly
*		R3[A]	vars
* Exit:		D0	->poly		CS if already was
*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
DISTABS		A=DAT0	W		Check first coefficient
		?A=0	S
		RTNYES
		CD0EX
		D=C	A		D[A] = ->poly
		CD0EX
		GOSUB	SLOTSIZE	B[A] = slotsize
-		A=DAT0	W
		?A=0	W
		GOYES	+		Exit when terminator found
		A=-A-1	S
		DAT0=A	W
		AD0EX
		A=A+B	A
		AD0EX
		GONC	-
+		C=D	A
		D0=C			D0 = ->poly
		RTNCC			CC: Was negated

*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
* Name:		SLOTSIZE
* Desc:		Calculate slot size information for distributed polys
* Entry:	R3[A] = vars
* Exit:		B[A] = slotsize		CC
*		C[A] = vars
*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
SLOTSIZE	C=R3.F	A		C[A]=vars
		B=C	A
		B=B+B	A
		B=B+B	A
		B=B+C	A
		B=B+CON	A,16		B[A]=5*vars+16 = slotsize
		RTNCC

*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
* Name:		EXTRASLOT
* Desc:		Allocate an extra slot for distributed polynomials
* Entry:	D1    = ->memory
*		R3[A] = vars
*		R4[A] = free memory
* Exit:		CS: Insufficient memory
*		CC: free memory updated, one slot skipped in D1
*		B[A] = vars
*		D1 = ->slot++
*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
EXTRASLOT	GOSUB	SLOTSIZE	B[A] = slotsize
		CBEX	A		B[A] = vars	C[A]=slotsize
		A=R4	A
		A=A-C	A		free memory - slotsize
		R4=A	A
		RTNC			CS: Insufficient memory
		AD1EX
		A=A+C	A		Skip the new slot
		D1=A
		RTNCC			CC: Sufficient memory

*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
* Name:		Distribute
* Desc:		Convert polynomial to distributed form
* Entry:	D0	->polynomial
*		D1	->output
*		B[A]	variables
*		R4[A]	free memory
* Output:	D0	->polynomial tail	CS: Insufficient memory
*		D1	->output tail		CC: Success
*		B[A]	variables
* Alters:	A[W] B[W] C[W] D[W]
* Stack:	2
*
* Algorithm:	Initialize slot to zero.
*		Initialize depth to zero.
*		If list then initialize degree(depth) = length(list)
*		degree(depth) = degree(depth)-1
*		If semi then depth = depth-1 and loop
*		If dohstr then coeff(slot)=z and duplicate slot, advance
*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

Distribute	CD1EX			* Initialize first slot
		D=C	A		D[A] = ->output (save)
		D1=C
		C=B	A		C[A] = vars
		C=C+C	A
		C=C+C	A
		C=C+B	A
		C=C+CON	A,16		C[A] = slotsize
		AR4EX
		A=A-C	A		Allocate the slot
		AR4EX
		RTNC			CS: Insufficient memory
		GOSBVL	=WIPEOUT
		C=D	A
		D1=C			D1 = ->output
		D=0	A		D[A] = depth = 0
		GOTO	DISTentry	Enter at proper place

* Distribution loop

DISTloop	AD1EX			A[A] = ->output
		D1=A
		C=D	A		C[A] = depth
		A=A+C	A
		C=C+C	A
		C=C+C	A
		A=A+C	A
		AD1EX
		D1=D1+	16-5		D1 = ->degree(depth)
		C=DAT1	A
		C=C-1	A		degree(depth)--
		GONC	DISTcontinue	No overflow - no SEMI either then
		D1=A			D1 = ->output
		D0=D0+	5		Skip SEMI
		D=D-1	A		D[A] = depth--
DISTnext	?D#0	A
		GOYES	DISTloop	Loop until depth=0
		A=0	W		Set the terminator value
		DAT1=A	W
		D1=D1+	16		Skip past the terminator
		RTNCC

DISTcontinue	DAT1=C	A		Store the decreased degree
		D1=A			D1 = ->output
		A=DAT0	A
		LC(5)	=DOHSTR
		?A=C	A
		GOYES	DISThstr	Handle coefficient

DISTentry	P=	16-5
-		BSL	W		Save B[A] and D[A]
		DSL	W
		P=P+1
		GONC	-
		GOSUBL	QLEN		A[A] = length of list
		P=	16-5
-		BSR	W		Restore B[A] and D[A]
		DSR	W
		P=P+1
		GONC	-
		CD1EX
		D1=C			D1 = ->output
		C=C+D	A
		C=C+D	A
		C=C+D	A
		C=C+D	A
		C=C+D	A
		CD1EX
		D1=D1+	16		->degree(depth)
		DAT1=A	A		Store the degree
		D1=C
		D=D+1	A		depth++
		D0=D0+	5		Skip the DOLIST
		GOTO	DISTloop	And continue

DISThstr	D0=D0+	5		Skip DOHSTR
		C=DAT0	A
		C=C-CON	A,7		Zlen-7
		GONC	+		No overflow if nonzero
		D0=D0+	6		Skip remaining part of Z0
		GOTO	DISTloop	And continue
+		D0=D0+	5		Skip Zlen
		P=C	0
		A=0	W
		A=DAT0	WP		A[W] = mantissa
		CD0EX
		C+P+1			Skip mantissa
		CD0EX
		P=	0
		A=DAT0	S		A[S] = sign
		D0=D0+	1		Skip sign
		DAT1=A	W		Store the coefficient

* Now make new slot with the very same values

		CD0EX
		RSTK=C			Save ->input
		AD1EX
		D0=A			D0 = ->slot
		C=B	A
		C=C+C	A
		C=C+C	A
		C=C+B	A
		C=C+CON	A,16		C[A] = slotsize
		AR4EX
		A=A-C	A		Allocate slot
		AR4EX
		GOC	DISTmemerr	Insufficient memory
		A=A+C	A
		D1=A			D1 = ->slot (next)
		GOSBVL	=MOVEDOWN	Copy
		C=RSTK
		CD0EX			D0 = ->input
		D1=C			D1 = ->output (the new slot)
*		A=0	W		Initialization of terminator
*		DAT1=A	W		not necessary - at exit only
		GOTO	DISTnext	And loop until list traversed

DISTmemerr	C=RSTK			Pop the saved D0
		RTNSC			CS: Insufficient memory

*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
* Name:		Recurse
* Desc:		Convert distributed polynomial into the standard
*		recursive representation.
* Entry:	D0	->polynomial (distributed)
*		D1	->output
*		B[A]	vars
*		R4[A]	free memory
* Output:	D0	->polynomial tail	CS: Insufficient memory
*		D1	->output tail		CC: Success
*		B[A]	vars
* Alters:	A[W] B[W] C[W] D[W]
* Stack:	1
* Variables:
*		vars	= total number of variables
*		vars1	= variables in previous term
*		vars2	= variables in current term
*		same	= number of equal leading powers
*		diff	= the change in the differing power (decrease)
*
* Algorithm:
*	
*	The number of variables in a term is the number of trailing
*	zero powers substracted from the total number of variables.
*
*	Assume we have processed the previous term. Then
*
*	Number of lists to terminate = vars1-same-1
*	  o Terminating a list means traversing the powers of the
*	    previous term backwards - skipping over the trailing
*	    zeros first - and outputting as many zeros as the power
*	    indicates - followed by a SEMI.
*	Number of zeros to output = diff-1
*	  o This is for the change in the degree
*	Number of lists to start = vars2-same-1 (if>0)
*	  o Simply output DOLIST's
*	Output the coefficient
*	  o Always nonzero
*	Number of lists to end = -(vars2-same-1) (if>0)
*	  o This terminates any list that had a nonzero coefficient
*	    besides the ones having zero powers.
*	    For example:  { { 1 0 0 0 1 } 1 1 }
*	                                ^--------- output 1 SEMI
*
*	Above seems to be the simplest possible algorithm (uh!).
*
*	Special cases:
*
*	Leading nonzero coefficient is handled properly if we start the loop
*	with variables vars1=0, same=-1, diff=1.
*
*	The terminator is handled properly if we execute the loop with
*	vars2=0, same=-1, diff=1 and then check for a zero coefficient
*	when outputting it is due.
*
*	Above fails for a zero polynomial so we must check for it first.
*
*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

Recurse		B=0	S		B[S] = 0  indicating no output yet
		D=0	A		D[A] = vars1 = 0
		A=DAT0	W		A[W] = coeff1
		?A#0	W
		GOYES	RECURloop	If nonzero then start the loop

		D0=D0+	16		Skip coeff
		CR4EX
		C=C-CON	A,11		Allocate Z0
		CR4EX
		RTNC			CS: Insufficient memory
		LCZ0			Output Z0
		DAT1=C	11
		D1=D1+	11
		RTNCC			CC: Success

* The main loop

RECURloop	A=DAT0	W		A[W] = coefficient
		?A#0	W
		GOYES	+		Continue if not terminator
		C=0	A		C[A] = vars2 = 0
		A=0	W		A[A1] = zeros = 0
		A=A-1	A		A[A] = same = -1
		GOTO	RECURnow	Handle the variables

+		CD0EX			Count the number of variables
		RSTK=C			in the new term
		C=C+B	A
		C=C+B	A
		C=C+B	A
		C=C+B	A
		C=C+B	A
		D0=C
		D0=D0+	16		D0 = ->slot tail
		A=B	A		A[A] = vars
-		D0=D0-	5		D0 = ->power--
		C=DAT0	A		C[A] = power
		?C#0	A
		GOYES	+		Done if found nonzero
		A=A-1	A		A[A] = vars--
		?A#0	A
		GOYES	-		Loop until all checked
+		C=RSTK
		D0=C			D0 = ->slot
		C=A	A		C[A] = vars2

		?B#0	S		Special variables if first one
		GOYES	+		Not first one - continue
		B=B+1	S		B[A]<>0 - got first coefficient
		A=0	W		A[A1] = zeros = 0
		A=A-1	A		A[A] = same = -1
		GOTO	RECURnow	Handle the variables

* Now compute the number of equal leading powers (variable 'same')

+		GOSBVL	=CSLW5		C[A1] = vars2
		CD0EX
		D0=C			D0 = ->slot
		D0=D0+	16		
		C=C-B	A
		C=C-B	A
		C=C-B	A
		C=C-B	A
		C=C-B	A
		CD1EX			D1 = ->previous slot
		P=	16-5
-		CSL	W		Save D1 in C[A1] (vars2 in C[A2])
		DSL	W		Save vars1 in D[A1] (also D[A]=0)
		P=P+1
		GONC	-	
-		A=DAT1	A		A[A] = oldpower
		C=DAT0	A		C[A] = newpower
		D0=D0+	5		->oldpower++
		D1=D1+	5		->newpower++
		D=D+1	A		D[A] = same = same+1
		?A=C	A		Continue until change found
		GOYES	-		Guaranteed to terminate!
		D=D-1	A		D[A] = same (fix back by one)
		A=A-C	A		A[A] = diff
		A=A-1	A		A[A] = diff-1 = zeros
		C=D	A		C[A] = same
		P=	16-5
-		ASL	W		A[A1] = zeros
		DSR	W		D[A] = vars1
		P=P+1
		GONC	-		
		A=C	A		A[A] = same
		GOSBVL	=CSRW5		C[A] = ->output	C[A1] = vars2
		D1=C
		GOSBVL	=CSRW5		C[A] = vars2

		CD0EX			Skip back to start of slot
		C=C-A	A		-5*same nibbles
		C=C-A	A
		C=C-A	A
		C=C-A	A
		C=C-A	A		
		CD0EX
		D0=D0-	16		-16 for the coefficient
		D0=D0-	5		-5 for the extra D0=D0+ 5

* Now handle the variables as instructed in the heading
*
*	B[A]	= vars		A[A]	= same
*	D[A]	= vars1		A[A1]	= zeros (diff-1)
*	C[A]	= vars2

RECURnow

* Terminate vars1-same-1 lists

		CDEX	A		D[A] = vars2 (to be new vars1)
		C=C-A	A
		C=C-1	A		C[A] = vars1-same-1
		?C#0	A
		GOYES	+		If nonzero then do terminate
		GOTO	RECURzeros	Else skip to outputting the zeros

+		B=B-C	A		vars-(vars1-same-1)
		B=B-A	A		vars-(vars1-same-1)-same
		B=B-1	A		vars-vars1
		CD0EX
		RSTK=C			RSTK = ->input
		C=C-B	A		Skip back over the trailing
		C=C-B	A		zero powers in previous slot
		C=C-B	A
		C=C-B	A
		C=C-B	A
		CD0EX
		B=B+1	A		Calculate vars back from
		B=B+A	A		vars-vars1
		B=B+C	A
		P=	16-5
-		ASL	W		A[A1] = same	A[A2] = zeros
		DSL	W		D[A1] = vars2
		P=P+1
		GONC	-
		D=C	A		D[A] =vars1-same-1

		LCZ0			Output [power] zeros
--		D0=D0-	5
		A=DAT0	A		A[A] = power
-		A=A-1	A		power--
		GOC	+
		CR4EX
		C=C-CON	A,11		Allocate Z0
		CR4EX
		GOC	RECURmem1	CS: Insufficient memory
		LC(5)	=DOHSTR		Output Z0
		DAT1=C	11
		D1=D1+	11
		GONC	-
+		CR4EX
		C=C-CON	A,5		Allocate SEMI
		CR4EX
		GOC	RECURmem1	CS: Insufficient memory
		LC(5)	=SEMI		Output SEMI
		DAT1=C	A
		D1=D1+	5
		D=D-1	A		D[A] = (vars1-same-1)--
		?D#0	A
		GOYES	--		Loop until terminated all
		P=	16-5
-		ASR	W		A[A] = same	A[A1] = zeros
		DSR	W		D[A] = vars2
		P=P+1
		GONC	-
		C=RSTK
		D0=C			D0 = ->input

* Output diff-1 zeros

RECURzeros	C=A	A		C[A] = same
		GOSBVL	=ASRW5		A[A] = zeros = diff-1
		RSTK=C			RSTK = same
		LCZ0
-		A=A-1	A		zeros--
		GOC	+
		CR4EX
		C=C-CON	A,11		Allocate Z0
		CR4EX
RECURmem1	GOC	RECURmemerr	CS: Insufficient memory
		DAT1=C	11
		D1=D1+	11
		GONC	-	
+		C=RSTK			C[A] = same
	
* Start vars2-same-1 new lists

		A=C	A		A[A] = same
		A=A+1	A		A[A] = same+1
		C=D	A		C[A] = vars2
		C=C-A	A		C[A] = vars2-same-1
		GOC	++		If overflow then none!
		A=C	A		A[A] = vars2-same-1
		RSTK=C			RSTK = vars2-same-1
		LC(5)	=DOLIST
-		A=A-1	A		(vars2-same-1)--
		GOC	+
		CR4EX
		C=C-CON	A,5		Allocate DOLIST
		CR4EX
		GOC	RECURmemerr	CS: Insufficient memory
		DAT1=C	A
		D1=D1+	5
		GONC	-
RECURmemerr	C=RSTK			Pop saved pointer
		RTNSC			CS: Insufficient memory
RECURdone	RTNCC
+		C=RSTK

* Output coefficient

++		A=DAT0	W		A[W] = coefficient
		D0=D0+	16		Skip it
		?A=0	W
		GOYES	RECURdone	If zero then done!
		RSTK=C			C[A] = vars2-same-1
		CR4EX
		C=C-CON	A,5		Allocate DOHSTR
		CR4EX
		GOC	RECURmemerr	CS: Insufficient memory
		LC(5)	=DOHSTR		Output DOHSTR
		DAT1=C	A
		D1=D1+	5
		P=	15
-		P=P-1			Trim the coefficient
		?A=0	P
		GOYES	-
		C=0	A
		C+P+1
		C=C+CON	A,6		C[A] = Zlen
		AR4EX
		A=A-C	A		Allocate Zlen and body
		AR4EX
RECURmem2	GOC	RECURmemerr	CS: Insufficient memory
		DAT1=C	A		Output Zlen
		D1=D1+	5
		DAT1=A	WP		Output mantissa
		CD1EX
		C+P+1
		CD1EX
		DAT1=A	S		Output sign
		D1=D1+	1
		P=	0
		C=RSTK
* Output -(vars2-same-1) lists
		A=C	A		A[A] = vars2-same-1
		C=C+C	A		CS if negative
		GONC	+		If not then slot done
		LC(5)	=SEMI		Else output SEMI's
-		CR4EX
		C=C-CON	A,5		Allocate SEMI
		CR4EX
		GOC	RECURmem2	CS: Insufficient memory
		DAT1=C	A
		D1=D1+	5
		A=A+1	A		-(vars2-same-1)--
		GONC	-		Loop until done

* Slot done - skip to next one and loop

+		CD0EX			->slot
		C=C+B	A
		C=C+B	A
		C=C+B	A
		C=C+B	A
		C=C+B	A
		CD0EX			->slot++
		GOTO	RECURloop
**********************************************************************
RPL		

