*	a) Bugged
*	b) No speed increase

**********************************************************************
* Desc:		Modular multiplication of normalized long integers.
* Stack:	( Za Zb Zn --> Z )
**********************************************************************

CODE
		GOSBVL	=SAVPTR
		A=DAT1	A		->Zn
		D0=A
		D0=D0+	5
		A=DAT0	A		Nsize
		ASR	A
		A=A+1	A		>= Nwords
		ASL	A		>= Nnibbles
		A=A+A	A
		A=A+A	A		*4 (*3 is minimum)
		LC(5)	100		Chicken
		C=C+A	A
		GOSBVL	=GETTEMP
		AD0EX
		R0=A	A		->output

		GOSBVL	=D1=DSKTOP
		A=DAT1	A
		B=A	A		->Zn
		D1=D1+	5
		A=DAT1	A
		D0=A			->Zb
		D1=D1+	5
		A=DAT1	A
		D1=A			->Za

		GOSUB	MULMOD		R0[A] = ->Z
		A=R0	A
		A=A+CON	A,5		->Zsize
		D0=A
		C=DAT0	A		Zsize
		A=A+C	A
		D0=A			->Ztail
		C=C+CON	A,5+6		+prolog + link + GCnib
		DAT0=C	A
		D0=D0+	5
		GOSBVL	=recover

		GOSBVL	=GETPTR
		D1=D1+	10		Drop Zb Zn
		D=D+1	A
		D=D+1	A
		GOVLNG	=OverWrR0Lp	Overwrite Za with Z

**********************************************************************
* Desc:		Modular multiplication of normalized long integers.
* Input:	D0	->Z1
*		D1	->Z2
*		R0[A]	->output with sufficient memory
*		B[A]	->Zn
* Output:	R0[A]	->Z (result)
*
* Uses:		A[W] B[A] C[W] D[A] R1[A]
*
* Stack:	2 (RSTK save + Normalize call)
*
* Algorithm:
*
*	0<=Z1<Zn  and 0<=Z2<Zn is assumed
*
*	Take smaller number as the multiplier X
*	Take higher number as the multiplicand Y
*	Init result to zero
*	Start from high bit of X
*		If bit is set, then add Y to result
*		If overflow then substract N
*		Double the reslt, modulo N
*	Until all bits done
*
*	Note that this only requires:
*		- adding to the result modulo N
*		- substracting from the result (to normalize)
*		- doubling the result modulo N
*
*	In particular, once the original input is in suitable form we do
*	not need to change them in any way. This makes the algorithm more
*	straight forward to implement than the right-to-left algorithm.
*
*	Multiplication followed by division would be the clearest
*	implementation of course, but is too slow for factoring
*	algorithms.
*
*	Needed variables:
*		Current X-bit pointer
*		- Offset from the start of X in bits
*			- Easy to fetch current bit
*			- Easy to terminate at the lower bound
*		- Slot size
*			- Modular overflow checks
*			- Addition loop counter
*			- Substraction loop counter
*		- Slot for Y (to get easy register addition)
*		- Slot for N (to get easy register substraction & comparison)
*		- Slot for R (the result)
**********************************************************************


MULMOD

* 1) Get ~smaller of the numbers into D0

		D0=D0+	5
		D1=D1+	5
		A=DAT0	A
		C=DAT0	A
		?A<=C	A
		GOYES	+		D0 is already it
		AD0EX
		AD1EX
		AD0EX
+

* 2) Calculate slot size from Zn size

		C=B	A		->Zn
		CD0EX
		D=C	A		->Xsize

		D0=D0+	5
		C=DAT0	A
*		C=C-CON	A,6		Zn mantissa size
*		C=C+1	A		+ 1 nibble to handle overflows!
		C=C-CON	A,5
		CSR	A
		C=C+1	A		Words in slot
		R1=C.F	A		Save it for further use

* 3) We use the following data areas:
*	R0[A]			->Y		(copy for easy addition!)
*	R0[A] + slotsize	->Zn		(copy for easy testing)
*	R0[A] + 2*slotsize	->result
*    And the data variable:
*	R1[A]			words in slot
* X of course can stay where it is, as we only test bits from it
*

* 4) Initialize result and do the padding at the same time

		A=R0	A		->Yslot
		AD1EX
		D0=A			->Zy size

		C=R1	A		Words in slot
		A=C	A
		C=C+C	A
		C=C+A	A		3 slots
		CSL	A		as nibbles
		GOSBVL	=WIPEOUT

* 5) Copy Zy mantissa to its slot

		A=R0	A		->Yslot
		D1=A
		C=DAT0	A		Zysize
		D0=D0+	5
		C=C-CON	A,6		Zy mantissa size
		GOSBVL	=MOVEDOWN

* 6) Copy Zn mantissa to its slot

		A=R0	A		->Yslot
		C=R1	A		Words in slot
		CSL	A		Nibbles in slot
		A=A+C	A		->Nslot
		D1=A
		A=B	A
		D0=A			->Zn
		D0=D0+	5
		C=DAT0	A
		D0=D0+	5
		C=C-CON	A,6
		GOSBVL	=MOVEDOWN

* 7) Result slot is already initialized to zeros at step 4


* 8) Initialize the bit counter

		C=D	A		->Xsize
		D0=C
		C=DAT0	A
		D0=D0+	5
		C=C-CON	A,7		Xmantissa size-1
		AD0EX
		A=A+C	A
		D0=A			->Xhighnibble
		C=C+C	A
		C=C+C	A		4* to get bits instead of nibbles

		C=C+CON	A,3		Assume highest bit is set
		A=DAT0	B		A[0] = high nibble
-		A=A+A	P
		GOC	+		Found highest bit - exit
		C=C-1	A		Less bits in X
		GONC	-

* If C[A] underflows it means X is 0 and thus result is 0

		A=R0	A		->output
		D0=A
		LC(N)	11
		CON(5)	=DOHSTR
		CON(5)	6
		CON(1)	0
		DAT0=C	11
		RTNSC			CS: Zero result (useful?)

+		RSTK=C			RSTK = bit counter

* Now start multiplying to output

MULMODLOOP	C=RSTK			C[A] = bit counter
		RSTK=C
		A=C	A		
		CSRB.F	A
		CSRB.F	A		offset to proper nibble
		C=C+D	A
		C=C+CON	A,5		->nibble

		D0=C
		LC(1)	3		Mask low bits
		C=C&A	P
		P=C	0		[0 1 2 3]
		LCHEX	1248		[1 2 4 8]
		P=C	3
		C=P	0		The mask
		P=	0
		A=DAT0	B
		A=A&C	P		A[0] is now <> 0 if bit is set

		?A#0	P
		GOYES	+
		GOTO	mulmodnext	Bit is clear - no +Y to the result

* Bit is set, add Y to the result

+		A=R0	A
		D0=A			->Yslot

		C=R1	A		Words in slot
		B=C	A		counter

		CSL	A
		A=A+C	A		->Nslot
		A=A+C	A		->Rslot
		D1=A

mulmod+loop	A=DAT1	W
		C=DAT0	W
		A=A+C	W
-		DAT1=A	W
		GOC	mulmod+cs	CS for next word
		D0=D0+	16		No carry from previous addition
		D1=D1+	16		Thus we might reach end
		B=B-1	A
		?B#0	A
		GOYES	mulmod+loop
		GONC	++		Addition done

* Here we have carry from the addition. Based on 2) there will be 
* a next addition for sure.

mulmod+cs	B=B-1	A		This will not reach zero
		D0=D0+	16
		D1=D1+	16

		A=DAT1	W
		C=DAT0	W
		A=A+1	W		Add carry
		GOC	+		CS again
		A=A+C	W
		GONC	-		Back into regular non CC loop
		A=A-C	W
+		A=A+C	W		CS or not, we do have one now for sure
		DAT1=A	W
		GOTO	mulmod+cs

++		

* Addition now performed. Check result is normalized by comparing
* R and N from the highest nibbles downwards. If R>=N then R -= N

		GOSUB	Normalize	Take modulo if needed

* Now Y has been added to the result, and the result is normalized
* If there are more bits to go then we should double the result

mulmodnext	C=RSTK
		C=C-1	A
		GONC	+		More to go - double result
		GOTO	mulmodfinish	No more - return result as Z
+		RSTK=C

* Double the result the fast way - shift the bits upwards

		A=R0	A		->Y
		C=R1	A		Words in slot
		B=C	A
		CSL	A		nibbles in slot
		A=A+C	A
		A=A+C	A		->R
		D0=A

mulmod*2loop	A=DAT0	W
		A=A+A	W
-		DAT0=A	W
		GOC	mulmod*2cs	Handle carry
		D0=D0+	16
		B=B-1	A
		?B#0	A
		GOYES	mulmod*2loop
		GONC	++		Exit *2 loop

mulmod*2cs	D0=D0+	16
		B=B-1	A
		A=DAT0	W
		A=A+A	W
		GOC	+		Carry again
		A=A+1	W
		GONC	-		Back into the CS loop
		A=A-1	W
+		A=A+1	W
		DAT0=A	W
		GOTO	mulmod*2cs
++

* Now we normalize again, and continue the multiplication loop

		GOSUB	Normalize
		GOTO	MULMODLOOP


* Modular multiplication done - return the result in normalized Z form

mulmodfinish	A=R0	A		->Y
		D1=A			->result-to-be
		LC(5)	=DOHSTR
		DAT1=C	A
		D1=D1+	10		->mantissa-to-be

		C=R1	A		Words in slot
		CSL	A		nibbles in slot
		A=A+C	A		->N
		A=A+C	A		->R
		D0=A
		GOSBVL	=MOVEDOWN	mantissa moved ok

-		D1=D1-	1		Trimming loop
		A=DAT1	B
		?A=0	P
		GOYES	-
		D1=D1+	1		->result-to-be tail

		C=R0	A		->result-to-be
		C=C+CON	A,5
		D0=C			->result-to-be size field
		AD1EX
		A=A-C	A		size-to-be ?
		DAT0=A	A
		LC(5)	6		Must be atleast 6
		?A>=C	A
		GOYES	+
		DAT0=C	A
+		RTN

***
*** Normalize the current result
***
Normalize	A=R0	A		->Y
		C=R1	A		Words in slot
		B=C	A
		CSL	A		nibbles in slot
		A=A+C	A		->N
		A=A+C	A		->R = ->Ntail
		D0=A
		A=A+C	A		->Rtail
		D1=A

-		D0=D0-	16
		D1=D1-	16
		A=DAT0	W		Nnibbles
		C=DAT1	W		Rnibbles
		?C<A	W
		RTNYES			R<N - is normalized already
		?C>A	W
		GOYES	+		R>N - must normalize
		B=B-1	A
		?B#0	A
		GOYES	-		Loop until last nibbles

* Here we have R>=N - must do R -= N

+		A=R0	A		->Y
		C=R1	A		Words in slot
		B=C	A
		CSL	A		nibbles in slot
		A=A+C	A		->N
		D0=A
		A=A+C	A		->R
		D1=A

mulmod-loop	A=DAT1	W		Rnibbles
		C=DAT0	W		Nnibbles
		A=A-C	W
-		DAT1=A	W
		GOC	mulmod-cs	If CS then we must -1 from next
		D0=D0+	16
		D1=D1+	16
		B=B-1	A
		?B#0	A
		GOYES	mulmod-loop
		RTN

mulmod-cs	D0=D0+	16
		D1=D1+	16
		B=B-1	A		Cannot terminate yet
		A=DAT1	W
		C=DAT0	W
		A=A-1	W		Substract carry
		GOC	+
		A=A-C	W
		GONC	-		Back into CC loop
		A=A+C	W
+		A=A-C	W		CS in any case
		DAT1=C	W
		GOTO	mulmod-cs	Back into the CS loop

**********************************************************************

ENDCODE
