**********************************************************************
* Desc:		Trial division for polynomials
* Stack:	( P --> Lf P' )
**********************************************************************
NULLNAME PTrialDiv
::
   NULL{} SWAP				( L P )

   SUnivar? NOT?SEMI			( Not sunivariate - fail )

   ZERO ELEVEN				( L P #start #stop )
   BEGIN
	3PICK LENCOMP			( L P #start #stop d+1 )
	THREE #<			( L P #start #stop flag )
	ITE 2DROPFALSE		

* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * 
* Desc:		Find simple roots for polynomial.
* Stack:	( P #start #stop --> P FALSE | P #root #stop TRUE )
* Algorithm:	The polynomial is evaluated with trial values starting 
*		at start+1, until value is zero (root found) or stop 
*		value has been reached.
*
*		If given #0 as start value, search proceeds as follows:
*		-1, 1, -2, 2, -3, 4 ... #stop-1
*
* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * 
CODE
sXPOS	EQU 5			* Flag to mark current trial root >= 0

		GOSBVL	=POP2#
		R0=A			start
		R1=C			end
		GOSBVL	=SAVPTR
		A=DAT1	A

* Scan for ->Z1 and ->Zn

		A=A+CON	A,5		->Z1
		R3=A	A		Save ->Z1 for Horner evaluation
		D0=A
-		LC(5)	=DOHSTR
		A=DAT0	A
		D0=D0+	5
		?A#C	A
		GOYES	trynxt		Found SEMI - coefficients ok
		A=DAT0	A		Zlen
		CD0EX
		R4=C	A		->Zlastlen
		C=C+A	A
		D0=C
		GONC	-		Loop until last coefficient found

trynxt		A=R0	A		previous X value
		C=R1	A		stop
		ST=1	sXPOS		setup new X sign
		B=A	A
		B=B+B	A
		GOC	tryxpos		X was neg -> try pos value -X
		ST=0	sXPOS		X was pos -> try neg value -(X+1)
		A=A+1	A
		?A=C	A
		GOYES	noroot		Further trials not desired
tryxpos		A=-A	A
		R0=A	A		new X value
		GOSUB	TryRoot
		GOC	trynxt		Failed due to overflow, try next
		?B#0	W
		GOYES	trynxt		Value not 0 - try next
* got a root
		GOSBVL	=PUSH2#
		GOVLNG	=PushTLoop

noroot		GOVLNG	=GPPushFLoop


* Evaluate polynomial at value X in R0[A] ==> value in B[W], sign in D[S]
* First coefficient address in R3[A], last coefficient length address in R4[A]

TryRoot		A=R3	A
		D0=A			->Z1
		D0=D0+	5		->Z1len
		C=R4	A
		D=C	A		->Z2len
		B=0	W		rmant
		D=0	S		rsign

* Horner method evaluation

hornloop	C=DAT0	A		Zlen
		D0=D0+	5
		C=C-CON	A,7
		GOC	pevnxt		No need to add Z0
		P=C	0
		A=0	W
		A=DAT0	WP
		CD0EX
		C+P+1			->Zsign
		CD0EX
		P=	0
		C=DAT0	S		Zsign
		?C#0	S
		GOYES	pevaddN
		?D=0	S
		GOYES	pevaddPP
pevaddPN
pevaddNP	B=B-A	W		r = r - x
		GONC	pevnxt
		B=-B	W		r changes sign
		D=-D-1	S
		GONC	pevnxt
pevaddN		?D=0	S
		GOYES	pevaddNP
pevaddPP	B=B+A	W		r = r + x
		RTNC			CS: Overflow
pevnxt		D0=D0+	1+5		Skip sign & next DOHSTR
		CD0EX
		?C>D	A
		GOYES	pevdone		Added last Z - done
		D0=C

		A=B	W		r
		B=0	W
		C=R0.F	A		x
		?ST=1	sXPOS
		GOYES	pevmul
		C=-C	A
		D=-D-1	S		r changes sign
pevmul		SB=0
pevmullp	CSRB.F	A
		?SB=0
		GOYES	pevmul+
		SB=0
		B=B+A	W
		RTNC			CS: Overflow
pevmul+		A=A+A	W
		RTNC			CS: Overflow
		?C#0	A
		GOYES	pevmullp
		GOTO	hornloop
pevdone		RTNCC		

ENDCODE

   WHILE				( L P #root #stop )
	   2SWAP HXS 2 10 		( #root #stop L P Z1 )
	   5PICK DUPDUP #2*		( #root #stop L P Z1 #r #r #2r )
	   #> ITE
	     ::				( negative root )
		ZERO SWAP#-		( #root #stop L P Z1 #-r )
		#>Z			( #root #stop L P Z1 -Zroot )
 	     ;
	     ::				( positive root )
	 	#>Z			( #root #stop L P Z1 Zroot )
		QNeg			( #root #stop L P Z1 -Zroot )
	     ;
	   TWO{}N			( #root #stop L P F )
	   ROTOVER >TCOMP		( #root #stop P F L' )
	   5UNROLL QDiv			( L' #root #stop P' )

           ( Check P' is still sunivariate )
	   SUnivar? NOTcase

	   :: UNROT2DROP 2RDROP ;	( L' P' )
	   UNROT			( L' P' #root #stop )
   REPEAT
;

**********************************************************************
