EXTERNAL xRESULTANT	( symb1 symb2 var --> symb )
EXTERNAL xSUBPRS	( symb1 symb2 var --> Lsymb )

EXTERNAL SymResultant	( symb1 symb2 var --> symb )
EXTERNAL SymSubPRS	( symb1 symb2 var --> Lsymb )
EXTERNAL Resultant	( Q1 Q2 --> Q )
EXTERNAL SubPRS		( Q1 Q2 --> q1 .. qn #n )
EXTERNAL AA2RR{}	( symb1 symb2 var/list --> L Q1 R1 Q2 R2 )

**********************************************************************
*		Programs
**********************************************************************

**********************************************************************
* Name:		xRESULTANT
* Desc:		Calculate resultant of two symbolics 
* Stack:	( symb1 symb2 var --> symb' )
**********************************************************************
ASSEMBLE
	CON(1)	8
RPL
xNAME RESULTANT
::
  CK3&Dispatch
    #AAA	:: FastLib SymResultant ;
    #1AA	:: FastLib SymResultant ;
    #A1A	:: FastLib SymResultant ;
    #11A	:: FastLib SymResultant ;
;

**********************************************************************
* Name:		xSUBPRS
* Desc:		Calculate subresultant p.r.s of two symbolics
* Stack:	( symb1 symb2 var --> Meta )
**********************************************************************
ASSEMBLE
	CON(1)	8
RPL
xNAME SUBPRS
::
  CK3&Dispatch
    #AAA	:: FastLib SymSubPRS ;
    #1AA	:: FastLib SymSubPRS ;
    #A1A	:: FastLib SymSubPRS ;
    #11A	:: FastLib SymSubPRS ;
;

*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
* Name:		SymResultant
* Desc:		Calculate resultant of two symbolics
* Stack:	( symb1 symb2 var --> symb )
* Note:		Should handle rational coefficients..
*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
NULLNAME SymResultant
::
  AA2RR{}				( L Q1 R1 Q2 R2 )
  ZIsOne? NcaseTYPEERR			( L Q1 R1 Q2 )
  SWAP ZIsOne? NcaseTYPEERR		( L Q1 Q2 )
  Resultant				( L Q )
  HXS 2 10 ROT R2A			( symb )
;

*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
* Name:		SymSubPRS
* Desc:		Calculate subresultant PRS of two symbolics
* Stack:	( symb1 symb2 var --> L )
* Note:		Should handle rational coefficients..
*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
NULLNAME SymSubPRS
::
  AA2RR{}				( L Q1 R1 Q2 R2 )
  ZIsOne? NcaseTYPEERR			( L Q1 R1 Q2 )
  SWAP ZIsOne? NcaseTYPEERR		( L Q1 Q2 )
  SubPRS				( L q1 q2 .. qn #n )
  DUP ZERO_DO (DO)
     DUP #1+ROLL
     HXS 2 10 3PICK #3+PICK R2A
     SWAP
  LOOP					( L symb1 .. symbn #n )
  get1 DROP				( symb1 .. symbn #n )
  {}N
;

*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
* Name:		Resultant
* Desc:		Calculate resultant of two polynomials
* Stack:	( Q1 Q2 --> Q )
*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
NULLNAME Resultant
::
  SubPRS				( q1 .. qn #n )
  pull OVER#2+UNROLL NDROP		( qn )
;

*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
* Name:		SubPRS
* Desc:		Calculate subresultant P.R.S of two polynomials
* Stack:	( Q1 Q2 --> q1 .. qn #n )
* Note:		Values for special input are questionable (zeros, constants)
* Reference:	Maple v4.0 share library
*
* Main program is as follows
*
*   S = u
*   while dv > 0
*	d=du-dv
*	c=c*(-1)^(du*dv)
*	r=prem(u,v)
*	u=v; v=r; du=dv; dv=degree(r)
*	j=degree(v);
*	S=S, u*s^((ddu-j)*(ddv-j))
*	v=divide(v,g*h^d)
*	g=coeff(u,du)
*	if d=1 then
*	   h=g
*	else
*	   h=divide(g^d,h^(d-1))
*	end
*   repeat
*   tmp=c*v*s^(ddu*ddv)
*   if du=1 then
*	S=S,tmp
*   else
*	r=divide(v,h)
*	S=S,tmp*r^(du-1)
*   end
*   return(S)
*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
NULLNAME SubPRS
::
  OVER DUPTYPELIST? ITE LENCOMP DROPONE #1-	( Q1 Q2 n1 )
  OVER DUPTYPELIST? ITE LENCOMP DROPONE #1-	( Q1 Q2 n1 n2 )
* Swap polynomials if deg(Q1)<deg(Q2)
  2DUP#< ITE
    :: 4ROLLSWAP ROT2DUP #* ONE #AND #0= ;
    TRUE
  3PICK3PICK					( Q1 Q2 n1 n2 sign n1 n2 )
  TRUE HXS 2 10 DUPZERO

  {{ u v du dv s ddu ddv c g h d }}

* Resultant(0,v) = Resultant(u,0) = u

  u QIsZero? v QIsZero? ORcase
  :: u ONE ABND ;

* Resultant(const,const) = 1

  du #0= dv #0= ANDcase
  :: h ONE ABND ;

* Resultant(u,const) = u,v,v^du

  dv #0=case
  :: u v DUP du PPow# THREE ABND ;

* Main algorithm

  u ONE				( Initial sequence )
  BEGIN
	du dv #- d!

	c du v #* ONE #AND #1= IT NOT c!

	u v PseudoDiv DROPSWAPDROP
	v u! v!
	dv du!

	v DUPTYPELIST? ITE LENCOMP DROPONE #1- dv!

	u s ?SKIP
	:: ddu du #- ddv du #- #*
	   ONE #AND #1= IT QNeg
	;
	SWAP#1+

	v g h d PPow# QMul QDiv v!

	u DUPTYPELIST? IT CARCOMP
	DUPTYPELIST? IT ONE{}N g!

	d #1= ITE
	:: g h! ;
	:: g d PPow# h d #1- PPow# QDiv h! ;

	dv #0=
  UNTIL

  v c ?SKIP QNeg
  s ?SKIP
  :: ddu ddv #* ONE #AND #1= IT QNeg ;
  SWAP#1+

  du #1=case ABND

  SWAP
  v du #1- PPow# QMul		( Must multiply first! )
  h du #1- PPow# QDiv
  SWAP
  ABND
;

*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
* Name:		AA2RR{}
* Desc:		Convert symbolics to rationals with given ordering
* Stack:	( symb1 symb2 var/list --> L Q1 R1 Q2 R2 )
*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
NULLNAME AA2RR{}
::
  DUPTYPELIST? ?SKIP ONE{}N
  3PICK3PICK XAdd
  SWAP PVars{}
  ROTOVER A2R RCanno
  4ROLL 4PICK A2R RCanno
;

**********************************************************************
