	STITLE	ALG Division Algorithm
**********************************************************************
*		ALG Library:	Polynomial/Integer division routines
**********************************************************************

**********************************************************************
* EXPORT:
*
*   EuclidDiv	( A B --> q1 q2 r1 r2 )
*   PseudoDiv	( A B --> q r c )
*
*   QDiv?	( Q1 Q1 --> Q TRUE / FALSE )
*   QDiv	( Q1 Q2 --> Q1/Q2 )
*   QDivRem	( Q1 Q1 --> Q1/Q1 Q1remQ2 )
**********************************************************************

**********************************************************************
* Desc:		Euclidean division over F[x] 
*                 A/B  = q1/q2 + r1/r2 
*		  s.t. deg(r1,x) < deg(r2,x) <= deg(Q,x) & deg(q2,x)=0
* Stack:	( A B -> q1 q2 r1 r2 )
**********************************************************************
NULLNAME EuclidDiv
::
  SWAPOVER PseudoDiv		( B q r c )
  ROTOVER RCanno		( B r c q1 q2 )
  2SWAP 5ROLL QMul		( q1 q2 r c*B )
  RCanno			( q1 q2 r1 r2 )
;

**********************************************************************
* Desc:		Pseudo division over Z[X]
*                 c*A = B*q + r 
*		  s.t. deg(r,x) < deg(B,x) & deg(c,x)=0
* Stack:	( A B -> q r c )
**********************************************************************a
NULLNAME PseudoDiv
::
  DUPTYPELIST? ITE DUPLENCOMP ONE		( a b bn )
  DUP#1= casedrop				( deg[b] = 0 )
  :: HXS 1 0 SWAP ;				( --> q=a r=0 c=b )

  3PICK DUPTYPELIST? ITE LENCOMP DROPONE	( a b bn )
  SWAP 2DUP#< case2drop				( deg[a]<deg[b] )
  :: DROP HXS 1 0 SWAP HXS 2 10 ;		( --> q=0 r=a c=1 )

  ( deg[a] >= deg[b] > 0 )			( a b an bn )

  #-+1 OVER CARCOMP SWAP PPow#			( a b c' )
  DUPTYPELIST? IT ONE{}N			( a b c )

  UNROTSWAP 3PICK QMul 				( c b a*c )
  SWAP QDivRem 					( c q' r' )
  PTrim SWAP PTrim SWAPROT			( c q' r )
;


**********************************************************************
* Desc:		If Q1 is divisible by Q2, return quotient and TRUE
*		FALSE otherwise
* Stack:	( Q1 Q2 --> Q TRUE / FALSE )
**********************************************************************
NULLNAME QDiv?
::
  DEPTH 1LAMBIND
  ERRSET
  CODE
	GOTO	QDivRemNow		For Profiler!
  ENDCODE
* QDivRem
  ERRTRAP
    ::
	ERROR@ #304 ( Undef Res Error ) #<>case ERRJMP
	DEPTH #2+ 1GETABND #- NDROP
        FALSE RDROP
    ;
  ABND
  QIsZero? DUP ?SEMI
  SWAPDROP
;

**********************************************************************
* Desc:		Divide two polynomials and return the quotient
*		The result is trimmed
* Stack:	( Q1 Q2 --> Q1/Q2 )
**********************************************************************
NULLNAME QDiv
ASSEMBLE
		CON(5)	=DOCODE
		REL(5)	->QDivEnd

		GOSUBL	AttnAbort?
		GOSBVL	=SAVPTR

* First test for division by 1
		C=DAT1	A
		D0=C			->Q2
		A=DAT0	12
		C=A	W
		LCZ1
		?A=C	W
		GOYES	qdiv1
* Then test for division by -1
		P=	12-1
		C=-C-1	P
		P=	0
		?A=C	W
		GOYES	qdiv-1
* Then test for equality
		D1=D1+	5
		C=DAT1	A
		CD0EX			->Q1
		D1=C			->Q2
		GOSUB	EQUAL?		CC: Equal Q's
		GOC	qdivdispatch	Different Q's - dispatch

qdivequ		GOSBVL	=GETPTR		Q/Q = 1
		D1=D1+	10
		D=D+1	A		Drop Q2
		D=D+1	A		Drop Q1
		GOSBVL	=SAVPTR
		A=0	W
		A=A+1	A
		P=	1
		GOVLNG	=PUSHhxsLoop

qdiv1		GOSBVL	=GETPTR		Q1/1 = Q1
		GOVLNG	=DropLoop

qdiv-1		GOSBVL	=GETPTR		Q1/-1 = -Q1
		D1=D1+	5		Drop -Z1
		D=D+1	A
		GOLONG	DoQNeg		Negate Q1

* Here:	Q2 is not 1 or -1, and Q1<>Q2
*	D0 = ->Q1	D1 = ->Q2
* Dispatch down to individual subroutines from Q types

qdivdispatch	LC(5)	=DOHSTR
		A=DAT1	A
		?A=C	A
		GOYES	qdivz
		A=DAT0	A
		?A=C	A
		GOYES	zdivp
		GOTO	pdivp
qdivz		A=DAT0	A
		?A=C	A
		GOYES	zdivz
		GOTO	pdivz

**********************************************************************
* Calculate Z/P
* Special cases:	- Always 0
**********************************************************************

zdivp		A=0	W		Return Z0
		GOSBVL	=PUSHhxs
		A=DAT1	A
		D1=D1+	10
		D=D+1	A
		D=D+1	A
		GOVLNG	=OverWrALoop

**********************************************************************
* Calculate Z1/Z2
* Special cases:	- Z known not to be 1 or -1 in here
*			- Z1 <> Z2 known
*			- If |Z1| < |Z2| then Z1/Z2 = 0 ???
*			- If |Z1| = |Z2| then Z1/Z2 = 1 ???
**********************************************************************

zdivz		D0=D0+	5
		A=DAT1	A		Z1size
		A=A+A	A		2xZ1size
		LC(5)	10+6+20		2xPrologs + link + safety
		C=C+A	A
		GOSBVL	=GETTEMP	D0 ->tempobj

		GOSBVL	=D1=DSKTOP
		C=DAT1	A
		D1=D1+	5
		A=DAT1	A
		AD0EX			D0 = ->Z1
		R0=A	A		R0[A] = ->output
		D1=C			D1 = ->Z2

		GOSUB	ZQUO		Divide into output

		C=R0	A		->Quotient
		C=C+CON	A,5		Skip prolog
		D0=C
		A=DAT0	A		QUOsize
		C=C+A	A		C[A] = ->Quotient tail
		GOSUBL	ShrinkQ		Shrink R0-C
		D1=D1+	5		Drop Z2
		D=D+1	A
		GOVLNG	=OverWrALoop	And overwrite Z1 with quotient

**********************************************************************
* Calculate P/Z
* Special cases:	- Z known not to be 1 or -1 in here
*
* Algorithm:
*	Simply allocate sufficient memory and divide elements one by one
**********************************************************************

* IRAMBUFF storage for loop pointers/variables is
* needed because ZQUO uses so many registers

Z2PTR	EQU #100		Pointer into Z2 running pointer
CDEPTH	EQU #105		Composite depth
DIV$	EQU #10A		Start address of output object

pdivz		GOSUBL	QSIZE		C[A] = Qsize
		C=C+C	A		2 * Qsize
		LA(5)	10+6+100	2xPrologs + link + safety
		C=C+A	A
		GOSBVL	=GETTEMP	D0 ->tempobj

* Now init the loop variables

		AD0EX			->tempobj
		D0=(5)	(=IRAM@)-4
		C=DAT0	A
		LC(4)	DIV$		->storage
		D0=C
		DAT0=A	A		->output object

		R0=A	A		->output pointer
		B=0	A		cdepth = 0
		
		GOSBVL	=D1=DSKTOP
		A=DAT1	A		->Z
		D1=D1+	5
		C=DAT1	A		->P
		D0=C			->P
		D1=A			->Z		

pdivzloop	A=DAT0	A
		LC(5)	=DOHSTR
		?A=C	A
		GOYES	pdivznow	Execute Zi/Z2
		C=R0	A		->output ptr
		CD1EX
		DAT1=A	A		Copy DOLIST/SEMI
		D0=D0+	5
		D1=D1+	5
		CD1EX
		R0=C	A		->output ptr'
		LC(5)	=SEMI
		?A=C	A
		GOYES	pdivzsemi
		B=B+1	A		cdepth++
		GONC	pdivzloop
pdivzsemi	B=B-1	A		cdepth--
		?B#0	A
		GOYES	pdivzloop

* Done - shrink output & push result

		D0=(5)	(=IRAM@)-4
		C=DAT0	A
		LC(4)	DIV$
		D0=C
		A=DAT0	A		->output
		C=R0	A		->output tail
		R0=A	A		->output

		GOSUBL	ShrinkQ		A[A]=->output
		D1=D1+	5		Drop Z2
		D=D+1	A
		DAT1=A	A
		GOLONG	DoPTrim		!!! P/Z may need trimming !!!

*---------------------------------------------------------------------

pdivznow	AD0EX			->Zi
		D0=(5)	(=IRAM@)-4
		C=DAT0	A
		LC(4)	Z2PTR
		D0=C
		DAT0=A	A		Save ->Zi
		D0=(2)	CDEPTH
		C=B	A
		DAT0=C	A		Save cdepth
		D0=A
		GOSUB	ZQUO		Divide into output

		C=R0	A		Update ->output'
		C=C+CON	A,5		skip prolog
		D0=C
		A=DAT0	A		QUOsize
		C=C+A	A
		R0=C	A		->output'

		GOSBVL	=D1=DSKTOP	Restore ->Z
		C=DAT1	A
		D1=C			->Z

		D0=(5)	(=IRAM@)-4
		C=DAT0	A
		LC(4)	CDEPTH
		D0=C
		A=DAT0	A		Restore CDEPTH
		B=A	A
		D0=(2)	Z2PTR
		A=DAT0	A		Restore ->Zi
		A=A+CON	A,5		And skip it to continue
		D0=A
		C=DAT0	A		Zisize
		A=A+C	A
		D0=A			->Zi+1

		GOTO	pdivzloop

**********************************************************************
* Calculate P1/P2
* Special cases:	- P1 <> P2 known
*			- if len(P1) < len(P2) then return 0
**********************************************************************

pdivp		GOSUBL	QLEN2		A[A]=len1	C[A]=len2
		A=A-C	A		len1-len2
		GONC	+		CC: Must try division
		A=0	W
		GOLONG	PushZ2Loop	Drop 2 and push zero
+		GOSBVL	=GETPTR
		A=PC
		A=A+CON	A,10
		PC=(A)
RPL

:: PRemP DROP PTrim ;

ASSEMBLE
->QDivEnd
RPL

**********************************************************************
* Desc:		Divide two polynomials and return the quotient and
*		the remainder. The results are NOT trimmed.
* Stack:	( Q1 Q2 --> Q1/Q2 Q1remQ2 )
**********************************************************************
NULLNAME QDivRem
ASSEMBLE
		CON(5)	=DOCODE
		REL(5)	->QDivRemEnd

QDivRemNow	GOSUBL	AttnAbort?
		GOSBVL	=SAVPTR

* Fetch arguments

		A=DAT1	A
		D1=D1+	5
		C=DAT1	A
		D0=C			->Q1
		D1=A			->Q2

* First test for division by 1
		A=DAT1	12
		C=A	W
		LCZ1
		?A#C	W
		GOYES	qremdispatch

* NOTE: Code elsewhere assumes results are not trimmed.
*	Thus we must not check for the special case
*	of equal input

qrem1		GOSBVL	=GETPTR		Q1/1 = Q1, 0
		D1=D1+	5		Drop 1
		D=D+1	A
		GOSBVL	=SAVPTR
		A=0	W
		GOVLNG	=PUSHhxsLoop	Push Z0

* Here:	Q2 is not 1, and Q1<>Q2
*	D0 = ->Q1	D1 = ->Q2
* Dispatch down to individual subroutines from Q types

qremdispatch	LC(5)	=DOHSTR
		A=DAT1	A
		?A=C	A
		GOYES	qremz
		A=DAT0	A
		?A=C	A
		GOYES	zremp
		GOTO	premp
qremz		A=DAT0	A
		?A=C	A
		GOYES	zremz
		GOTO	premz

**********************************************************************
* Calculate Z/P with remainder
* Special cases:	- Z/P always 0, remainder is Z
**********************************************************************

zremp		A=0	W
		GOSBVL	=PUSHhxs	Push Z0
		A=DAT1	A		( Z P Z0 )
		D=D+1	A
		D1=D1+	10
		C=DAT1	A		->Z
		DAT1=A	A
		D1=D1-	5
		DAT1=C	A		( Z0 Z )
		GOVLNG	=Loop

**********************************************************************
* Calculate Z1/Z2 with remainder
* Special cases:	- Z known not to be 1
*			- Z1 <> Z2 known
**********************************************************************

zremz		D0=D0+	5
		A=DAT1	A		Z1size
		A=A+A	A		2xZ1size
		LC(5)	10+6+20		2xPrologs + link + safety
		C=C+A	A
		GOSBVL	=GETTEMP	D0 ->tempobj

		GOSBVL	=D1=DSKTOP
		C=DAT1	A
		D1=D1+	5
		A=DAT1	A
		AD0EX			D0 = ->Z1
		R0=A	A		R0[A] = ->output
		D1=C			D1 = ->Z2

		GOSUB	ZQUO		Divide into output

		C=R0	A		->Quotient
		C=C+CON	A,5		Skip prolog
		D0=C
		A=DAT0	A		QUOsize
		C=C+A	A
		D0=C			->Quotient tail
		A=A+CON	A,11		link = QUOzise+prolog+link+GCnib
		DAT0=A	A		write link
		D0=D0+	11		skip link+GCnib+remprolog

		C=C+CON	A,6
		R1=C			R1[A] = ->Remainder

		A=DAT0	A		REMsize
		CD0EX
		C=C+A	A		->Remainder tail
		CD0EX
		A=A+CON	A,11		link = QUOsize+prolog+link+GCnib
		DAT0=A	A		write link
		D0=D0+	5		skip link

		GOSBVL	=recover	shrink allocated memory

		GOSBVL	=GETPTR
		D1=D1+	5
		A=R0
		DAT1=A	A		stack level 2 = ->Quotient
		D1=D1-	5
		A=R1
		GOVLNG	=OverWrALoop	stack level 1 = ->Remainder

**********************************************************************
* Calculate P/Z with remainder
* Special cases:	- Z known not to be 1
**********************************************************************

premz		GOSBVL	=GETPTR
		A=PC
		A=A+CON	A,10
		PC=(A)
RPL

::
  2DUP QDiv			( P Z P/Z )
  DUP4UNROLL			( P/Z P Z P/Z )
  SWAP QMul			( P/Z P P' )
  QSub				( P/Z PremZ )
;

ASSEMBLE

**********************************************************************
* Calculate P1/P2 with remainder
* Special cases:	- P1 <> P2 known
*			- If len(P1) < len(P2) then return 0
**********************************************************************

premp		GOSUBL	QLEN2		A[A]=len1	C[A]=len2
		A=A-C	A		len1-len2
		GONC	+		CC: Must try division
		GOTO	zremp		Same result as for Z/P	
+		GOSBVL	=GETPTR
		A=PC
		A=A+CON	A,10
		PC=(A)
RPL

**********************************************************************
* Desc:		Divide two non-constant polynomials.
*		The results are NOT trimmed.
* Stack:	( P1 P2 --> P1/P2 P1remP2 )
**********************************************************************

NULLNAME PRemP
::

* First try fast division for polynomials with short coefficients

  CODE
	GOTO	DISTDIV?
  ENDCODE

* Stack: ( A B FALSE / quo rem TRUE )

  ?SEMI					( Fast division succeeded - done )

* Failed - use the regular slow algorithm

  DUPLENCOMP ROT			( B n A )
  DUPLENCOMP ROT			( B A m n )

* len(A) >= len(B) always - see code at pdivp and premp
*
*	2DUP#< case2drop
*	:: SWAPDROP HXS 1 0 SWAP ;	( zero A )

  #-+1					( B A m-n+1 )
  DUP4UNROLL				( m-n+1 B A m-n+1 )
  #1+_ONE_DO (DO)			( m-n+1 B C )
	DUP CARCOMP			( m-n+1 B C Ck )
	CODE
prempspeed	C=DAT1	A		* DupQIsZero?
		CD0EX
		A=DAT0	11
		D0=C
		C=A	W
		LCZ0
		?A=C	W
		GOYES	+
		GOVLNG	=Loop		* Return to RPL
+		A=PC			* Continue at CDRCOMP
		LC(5)	(prempcont)-(*)
		A=A+C	A
		D0=A
		GOVLNG	(=4UNROLL)+5
	ENDCODE

	3PICK CARCOMP			( m-n+1 B C Ck Bk )
	QDivRem QIsZero?		( m-n+1 B C Qk flag )
	NOTcase ErrUndefRes		( undefined result error )

	CODE
		GOTO	prempspeed	* Continue at CDRCOMP?
	ENDCODE

	4UNROLL CDRCOMP			( Qk m-n+1 B C' )
	4PICK 3PICK			( Qk m-n+1 B C' Qk B )
	FIVE NEXTCOMPOB 2DROP		( Qk m-n+1 B C' Qk B #off )
	BEGIN
		NEXTCOMPOB		( .. Qk B #off' ob TRUE )
	WHILE
		4PICK QMul
		4UNROLL			( .. Qk*Bi Qk B #off' )
	REPEAT
	SWAPDROP LENCOMP #1-		( Qk m-n+1 B C' Meta )
	HXS 1 0 ISTOP-INDEX #1- NDUPN	( Qk m-n+1 B C' Meta Meta0 )
	top& {}N QSub			( Qk m-n+1 B C'' )
	SKIP
LABEL prempcont
	CDRCOMP				( Qk m-n+1 B C' )
  LOOP					( Qm-n .. Q0 m-n+1 B R )
  SWAPDROP OVER#2+UNROLL		( R Qm-n .. Q0 m-n+1 )
  {}N SWAPDUP NULLCOMP?			( Q R flag )
  casedrop HXS 1 0			( Q R' )
;

ASSEMBLE
->QDivRemEnd

**********************************************************************
* Name:		ZQUO
*
* Desc:		Compute Z1/Z2 and Z1 mod Z2
*
* In:		D0	->Z1
*		D1	->Z2
*		R0[A]	->Output
*
* Out:		D0	->Z1tail
*		D1	->Z2tail
*		R0[A]	->Output
*		
*  		----------------------------------------
*               | Zquotient | xx 6nibs xx | Zremainder |
*  		----------------------------------------
*
*
* Requires:	Z2 must be trimmed (last nib of mantissa <> 0)
*		P=0	
*
* Alters:	A[W], B[W], C[W], D[W], R1, R2, R3, R4, SB
*
* Calls:	BDIV, SHORTOUT, BMULCRY, MOVEDOWN
*
* Stack Level:	3
*
**********************************************************************

ZQUO		D0=D0+	5
		D1=D1+	5
		A=DAT0	A		Z1size
		C=DAT1	A		Z2size
		?A<C	A
		GOYES	ZQUOis0		Z1len<Z2len -> Quo=0 Rem=Z1
		C=C-CON	A,7
		GONC	ZQUOnormal

* Quotient=0, Remainder=Z1

ZQUOis0		A=DAT1	A		Z2size
		C=R0
		CD1EX			D1 -> Output
		C=C+A	A		C[A] -> Z2tail
		RSTK=C			save it
		LCZ0
		DAT1=C	11		output 0
		D1=D1+	11
		D1=D1+	6		skip link space
		C=DAT0	A		Z2size
		C=C+CON	A,5
		D0=D0-	5
		GOSBVL	=MOVEDOWN	copy Z1 to output; D0 -> Z1tail
		C=RSTK			
		D1=C			D1 ->Z2tail
		RTNCC

* Not a special case

ZQUOnormal	

* First check whether calc can be done in register

		A=DAT0	A		
		A=A-CON	A,7		A = Z1 mantissa len
		A=0	P
		?A=0	A		
		GOYES	ZQUOB		no longer than 16 nibs -> fast div
		GOTO	ZQUOX		

* Z1 (and thus Z2 since Z1len >= Z2len) is no longer than 16 nibs 
* Do fast division in register

ZQUOB

* Pop the integers

		GOSUBL	shortpopd1
		B=A	W
		D=C	S		D[S] = sign2
		GOSUBL	shortpop	A[W] = Z1
		CBEX	W		C[W] = Z2	B[S] = sign1

* Divide A[W]B[S] / C[W]D[S] --> A[W]=quo B[W]=rem D[S]=sign

		GOSUBL	SHORT/
		C=D	S		C[S] = sign
		
* output quotient and remainder

		C=R0.F	A
		CD1EX			D1 ->Output
		RSTK=C			save ->Z2tail
		GOSUBL 	SHORTOUT	output quotient 
		D1=D1+	6		skip link space
		A=B	W
		D0=D0-	1
		C=DAT0	S		sign(rem) = sign(Z1)
		D0=D0+	1
		GOSUBL	SHORTOUT	output remainder
		C=RSTK
		D1=C			D1 ->Z2tail
		RTNCC

* Must do div in memory - check first if the divisor is short

ZQUOX		A=DAT1	A		Z2size
		A=A-CON	A,15		Z2 mantissa <= 8 nibs ?
		GOC	ZQUOS		yes -> short division
		GOTO	ZQUOL		no  -> long division

* Divisor is no longer than 8nibs 
* use short division, with 8nibs blocks (b=2^32)

ZQUOS

* Get divisor (Z2) in B[W] 

		GOSUBL	shortpopd1
		B=A	W		B[W] = Z2 mantissa
		D=C	S		D[S] = Z2sign

* Special entry to do division with a short divisor
* In:		D0	->Z1size
*		B[W]	divisor
*		D[S]	divisor sign
*		R0[A]	->Output
* Out:		D0	->Z1tail
*		B[W]	remainder
*		D[S]	divisor sign
*		R0[A]	->Output
* Alters:	

ZQUOint		C=R0
		CD1EX			D1 ->Output
		RSTK=C			save ->Z2tail

* Copy Z1 (including prolog but excluding sign) to output

		C=DAT0	A
		D=C	A		D[A]=Z2size
		C=C+CON	A,4		size incl. prolog but not sign
		D0=D0-	5
		GOSBVL	=MOVEDOWN	copy Z1 to output; D0 ->Z1sign
		C=0	W
		DAT1=C	W		pad with 0's to fill last block

* Compute size and # of blocks (n) :
* 
*	m = #nibs of Z2 mantissa = Z2size-6
*	n = #blocks = m+7 DIV 8
*	d = #zero nibs to fill last block = n*8-m = 7-(m+7 MOD 8)
*
		D=D+1	A		D[A]=m+7
		C=D	A
		LA(5)	7
		C=C&A	A		C[A]=m+7 MOD 8 
		A=A-C	A		A[A]=d
		CD1EX
		C=C+A	A
		CD1EX			D1 ->past last block of result
		DSRB.F	A
		DSRB.F	A
		DSRB.F	A		
		D=D-1	A		D[A]=n-1

* Init division

		C=B	W		C[W]=divisor
		B=0	W		initial r=0

* Compute quotient block by block

ZQUOSlp1	A=B	W		A[W]=r
		ASL	W
		ASL	W
		ASL	W
		ASL	W
		ASL	W
		ASL	W
		ASL	W
		ASL	W		A[W]=b*r
		D1=D1-	8		->next block
		A=DAT1	8		A[W]=b*r+Ui
		GOSUBL	BDIV		Division A[W]/C[W] -> A[W]=Qi B[W]=r
		DAT1=A	8		Write quotient block
		D=D-1	A		More blocks ?
		GONC	ZQUOSlp1	yes -> repeat

* Now D1 ->Quotient and B[W]=final remainder
* Trim quotient
		
		D1=D1-	5
		A=DAT1	A		A[A]=Z1size (incl. sign)
		CD1EX
		RSTK=C			save ->QUOsize
		C=C+A	A
		C=C-1	A
		CD1EX			D1 ->QUO(untrimmed)-tail
ZQUOStlp	D1=D1-	1
		C=DAT1	S
		?C#0	S
		GOYES	ZQUOStdn
		A=A-1	A
		GONC	ZQUOStlp
ZQUOStdn	C=RSTK			recover ->QUOsize
		CD1EX
		DAT1=A	A		write adjusted QUOsize
		CD1EX
		D1=D1+	1		D1 ->QUOsign

* Compute and write quotient sign

		C=DAT0	S
		CDEX	S		D[S]=sign(Z1)=sign(rem)  C[S]=sign(Z2)
		D0=D0+	1		D0 ->Z1tail
		C=C+D	S
		GONC	ZQUOSs
		C=0	S
ZQUOSs		DAT1=C	S		write quotient sign
		D1=D1+	1+6		skip sign + link space

* Output remainder

		A=B	W		A[W]=remainder
		C=D	S		C[S]=remainder sign
		GOSUBL	SHORTOUT	trim and write remainder
		C=RSTK	
		D1=C			D1 ->Z2tail
		RTNCC

* Dividend has more than 16nibs and divisor has more than 8nibs
* Do long division. 
*
* Use Knuth's algo with 8nibs blocks (b=2^32):
*
*   Z1 mantissa = Um Um-1 .. U3 U2 U1 | U0
*   Z2 mantissa = Vn Vn-1 .. V2 V1
*   Q  mantissa = Qm-n .. Q1 Q0
*
*   compute Qj-guess = min( b-1 , Uj'Uj+1' DIV V1' )     
*
*   where V1'        = normalised V1 so that b>V1'>=b/2
*		     = leading block of 2^(4e+f)(V1 V2) 
*         Ui'Ui'-1   = leading two blocks of 2^(4e+f)(Uj Uj+1 Uj+2)
*
*   Always have:  Qj-guess - 2 =< Qj-true =< Qj-guess 
*
* Scratch registers:
*
*   R0[A]  ->Output = base quotient - 10
*   R1[A]  = length (nibs) of Z2 mantissa
*   R1[S]  = f (0 =< f =< 3)
*   R2[A]  = #nibs left to calc in quotient; start with 8(m-n).
*   R2[S]  = sign Z1
*   R3[W]  = current Qi
*   R4[W]  = V1' 

ZQUOL		

* Get sizes and prepare quotient in output

		A=DAT1	A		Z2size
		D1=D1+	5		->Z2mantissa
		B=A	A
		B=B+1	A		B[A]=#nibs in Z2 mantissa + 7
		A=A-CON	A,6
		R1=A			R1[A]=length (nibs) of Z2 mantissa
		C=R0
		CD1EX			D1 ->Output
		RSTK=C			save ->Z2mantissa
		LC(5)	=DOHSTR
		DAT1=C	A		write quotient prolog
		D1=D1+	5
		C=DAT0	A		Z1size
		C=C+1	A		C[A]=#nibs in Z1 mantissa + 7
		LAHEX	FFFF8		A[A]=mask 3 LS bits
		C=C&A	A		C[A]=8*m
		A=A&B	A		A[A]=8*n
		C=C-A	A
		R2=C			R2[A]=8(m-n)
		C=C+CON	A,14		length quo = 8(m-n+1)+sign+5
		DAT1=C	A		write length quotient
		AD1EX
		A=A+C	A		skip quotient space
		A=A+CON	A,6		skip link space
		B=A	A		B[A] ->OutputRem
		AD1EX			D1   ->OutputRem

* Copy Z1 (incl. prolog but not sign) to output

		C=DAT0	A		Z1size
		C=C+CON	A,4		size incl. prolog but not sign
		D0=D0-	5
		GOSBVL	=MOVEDOWN	copy Z1 to output; D0 ->Z1sign
		C=0	W
		DAT1=C	W		pad with zeros

* Get Z1sign and slider start position

		A=R2			A[A]=8(n-m)
		A=DAT0	S		Z1sign
		D0=D0+	1		D0 ->Z1tail
		R2=A			R2[S]=sign Z1
		B=B+CON	A,10		B[A] ->Um
		A=A+B	A		A[A] ->(Um+(m-n))=->Un
		AD0EX			D0 ->Un = slider start pos
		C=RSTK
		CAEX	A		A[A] ->Z2mantissa
		RSTK=C			save ->Z1tail
		
* Compute V1'
		C=R1			
		A=A+C	A
		D1=A			D1 ->Z2sign
		D1=D1-	9
		C=DAT1	S		C[S]=4 MSBits of 2nd block 2^4e(V1V2)
		D1=D1+	1
		A=DAT1	8		A[1..7] = 1st block of 2^4e(V1V2)
		D1=D1+	8		D1 ->Z2sign
		D=0	S		D[S]=f counter
		P=	7
ZQUOLln1	B=A	WP		B[1..7] = 1st block of 2^(4e+f)(V1V2)
		A=A+A	WP		is B[1..7]>=b/2 ?
		GOC	ZQUOLn1dn	yes -> normalization done
		D=D+1	S		no  -> increment f
		C=C+C	S		       and shift V1V2 left 1bit	
		GONC	ZQUOLln1
		A=A+1	WP
		GONC	ZQUOLln1
ZQUOLn1dn	A=0	W
		A=B	WP
		P=	0
		R4=A			R4[W]=V1'
		C=D	S
		R1=C.F	S		R1[S]=f

* Loop on quotient block Qj
*
* At this point:  D0 ->slider = ->Ui	(i=j+n)
*		  D1 ->Z2sign 

ZQUOLljbeg	

* First compute Uj'Uj+1 = leading two blocks of 2^(4e+f)(Uj Uj+1 Uj+2)

		A=R1			A[A]=length Z2 mantissa
		C=A	S		
		D=C	S		D[S]=f
		CD0EX
		RSTK=C			save ->slider
		C=C+A	A
		D0=C
		D0=D0-	9
		C=DAT0	S		4 MSBits of 3rd blk of 2^4e(Uj..Uj+2)
		D0=D0+	1
		A=DAT0	W		First 2blks of 2^4e(Uj..Uj+2)
		C=RSTK
		D0=C			D0 ->slider

* Shift 2^4e(Uj..Uj+2) left f bits to get Uj'Uj+1'

ZQUOLln2	D=D-1	S		
		GOC 	ZQUOLn2dn
		A=A+A	W
		C=C+C	S
		GONC 	ZQUOLln2
		A=A+1	W
		GONC 	ZQUOLln2

* Now A[W]=Uj'Uj+1' - compute Qj-guess = min( b-1 , Uj'Uj+1' DIV V1' )

ZQUOLn2dn	C=R4			C[W]=V1'
		GOSUBL	BDIV		A[W]=Uj'Uj+1' DIV V1'
		C=0	W
		LCHEX	FFFFFFFF	C[W]=b-1
		?A<C	W
		GOYES	ZQUOLguess
		A=C	W
ZQUOLguess	R3=A			R3[W]=Qj-guess

* save ->slider = ->Uj+n

		CD0EX
		RSTK=C			save ->slider 
		CD0EX			D0 ->Uj+n

* if Qj=0 skip rest of current iteration

		?A#0	W
		GOYES   ZQUOLlm0
		GOTO	ZQUOLljnxt

* Compute [Uj .. Uj+n] - Qj-guess * [0 V1 .. Vn]

ZQUOLlm0	C=R1			
		D=C	A		D[A]=#nibs Z2 mantissa
		AD1EX
		A=A-C	A
		AD1EX			D1 ->Z2mantissa
		D=0	S		D[S]    = borrow from substraction
		B=0	W		B[7-15] = carry from multiplication
		P=	7
ZQUOLlm1	D=D-CON	A,8		Try using 8nibs from Z2
		GONC	ZQUOLlm2	
		C=D	B		Not that much left
		C=C+CON B,7		Use what's left
		P=C	0		
		D=0	A
ZQUOLlm2	A=0	W
		A=DAT1	WP		A=Vk (block k of Z2)
		CD1EX
		C+P+1
		CD1EX			D1 -> next block of Z2
		C=R3
		BCEX	W		B[W]=Qj  C[8-15]=carry
		GOSUBL	BMULCRY		A[W]=Vk*Qj + carry
		B=A	W		B[8-15]=new carry
		?D#0	A		last block of Z2 ?
		GOYES	ZQUOLlm3	no  -> do single block Uj+k
		P=	15		yes -> do last two blocks Uj Uj+1
ZQUOLlm3	C=DAT0	WP		get block(s) of U 
		?D=0	S		borrow from previous block ?
		GOYES	ZQUOLlm4
		C=C-1	WP		yes -> substract it
		GOC	ZQUOLlm4	new borrow ?
		D=0	S		no -> clear borrow
ZQUOLlm4	C=C-A	WP		substract Vk*Qj (mod b)
		GONC	ZQUOLlm5	borrow ?
		D=D+1	S		yes -> set borrow
ZQUOLlm5	DAT0=C	WP		write block(s) of U
		D0=D0+	8		D0 ->next block
		?D#0	A		more blocks in Z2
		GOYES	ZQUOLlm1	yes -> repeat
		P=	0
		?D#0	S		is there a final borrow ?
		GOYES	ZQUOLadj	yes -> Qj-guess was too big, adjust
		GOTO	ZQUOLljnxt	no -> Qj-guess was good, continue
		
* Current Qj is too big. Adjust it and add back

ZQUOLadj	A=R3			Qj <- Qj-1
		A=A-1	W
		R3=A			
		
* Add [0 V1 .. Vn] back to [Uj .. Uj+n]

		C=RSTK
		RSTK=C
		D0=C			D0 ->slider
		C=R1			
		D=C	A		D[A]=#nibs Z2 mantissa
		AD1EX
		A=A-C	A
		AD1EX			D1 ->Z2mantissa
		D=0	S		D[S] = carry
		P=	7
ZQUOLla1	D=D-CON	A,8		Try using 8nibs from Z2
		GONC	ZQUOLla2	
		C=D	B		Not that much left
		C=C+CON B,7		Use what's left
		P=C	0		
		D=0	A
ZQUOLla2	A=0	W
		A=DAT1	WP		A=Vk (block k of Z2)
		CD1EX
		C+P+1
		CD1EX			D1 -> next block of Z2
		?D#0	A		last block of Z2 ?
		GOYES	ZQUOLla3	no  -> do single block Uj+k
		P=	15		yes -> do last two blocks Uj Uj+1
ZQUOLla3	C=DAT0	WP		get block(s) of U 
		?D=0	S		carry from previous block ?
		GOYES	ZQUOLla4
		C=C+1	WP		yes -> add it 
		GOC	ZQUOLla4	new carry ?
		D=0	S		no -> clear carry
ZQUOLla4	C=C+A	WP		add Vk
		GONC	ZQUOLla5	carry ?
		D=D+1	S		yes -> set carry
ZQUOLla5	DAT0=C	WP		write block(s) of U
		D0=D0+	8		D0 ->next block
		?D#0	A		more blocks in Z2
		GOYES	ZQUOLla1	yes -> repeat
		P=	0
		?D=0	S		is there a final carry ?
		GOYES	ZQUOLadj	no -> Qj still 1 too big, repeat	

* Current Qj is ok and so is Uj..Uj+n and D1 ->Z2sign
* Write quotient block Qj - increment j - and repeat if j<=m-n

ZQUOLljnxt	C=R0
		C=C+CON	A,10		C[A] ->Quotient start (->Qm-n)
		A=R2			A[A] = #nibs left to calc in quotient
		C=C+A	A
		D0=C			D0 ->Qj location
		C=R3
		DAT0=C	8		write Qj
		C=RSTK
		D0=C			D0 ->old slider position (->Uj+n)
		A=A-CON	A,8		more quotient blocks ?
		GOC	ZQUOLljdone	no -> done with main loop
		R2=A			R2[A] = remaining #nibs in quotient
		D0=D0-	8		D0 ->new slider start
		GOTO	ZQUOLljbeg	repeat

* Main loop done. At this point:
*
*	D0 	  ->Remainder mantissa
*	D1        ->Z2sign
*	top[RTSK] ->Z1tail  
*	A[S]      = Z1sign

ZQUOLljdone

* Compute signs

		B=A	S		B[S] = sign Z1 = sign remainder
		C=DAT1	S		Z2sign
		D1=D1+	1		D1 ->Z2tail
		C=C+B	S
		GONC	ZQUOLs
		C=0	S
ZQUOLs		D=C	S		D[S] = sign quotient
		
* Trim remainder "in place" and set its sign

		A=R1			A[A] = Z2len >= length remainder
		CD0EX
		D0=C
		D0=D0-	5		D0 -> remainder size
		C=C+A	A
		CD1EX			D1 -> remainder tail
		RSTK=C			save ->Z2tail
ZQUOLtrlp	A=A-1	A		
		GOC	ZQUOLtrdn	remainder=0 -> no sign to set
		D1=D1-	1
		C=DAT1	S
		?C=0	S
		GOYES	ZQUOLtrlp
		D1=D1+	1
		C=B	S
		DAT1=C	S		write remainder sign
ZQUOLtrdn	A=A+CON	A,7		remainder size
		DAT0=A	A		write remainder size
		B=A	A		B[A] = remainder size

* Trim quotient and set its sign

		C=R0
		C=C+CON	A,5
		D=C	A		D[A] -> quotient size
		D1=C
		C=DAT1	A		quotient size
		A=C	A
		A=A-CON	A,6		A[A] = length quotient mantissa
		C=C+D	A		-> quotient tail (after sign)
		C=C-1	A		
		D1=C			D1 ->quotient mantissa tail
ZQUOLtqlp	A=A-1	A		
		GOC	ZQUOLtqdn	quotient=0 -> no sign to set
		D1=D1-	1
		C=DAT1	S
		?C=0	S
		GOYES	ZQUOLtqlp
		D1=D1+	1
		C=D	S
		DAT1=C	S		write quotient sign
ZQUOLtqdn	A=A+CON	A,7		quotient size
		C=D	A
		CD1EX
		DAT1=A	A		write quotient size
		CD1EX			D1 ->quotient sign
		
* Move remainder down, after trimmed quotient

		D1=D1+	7		skip quo sign + link space
		D0=D0-	5		D0 ->remainder prolog
		C=B	A
		C=C+CON	A,5		C[A]=rem size (incl. prolog)
		GOSBVL	=MOVEDOWN	move remainder down

* restore D1 and D0

		C=RSTK
		D1=C			D1 ->Z2tail
		C=RSTK
		D0=C			D0 ->Z1tail
		RTNCC

**********************************************************************
* Name:		DISTDIV?
* Desc:		Divides two multivariate polynomials, returning the
*		quotient and remainders.
* Stack:	( P1 P2 --> quo rem )
* Notes:	Handles both P1=0 and P2=0 properly.
*		Both inputs must be polynomials.
**********************************************************************

* Equate in global.s
* sDUMMYCOEFF	EQU 8			* Flag dummy coefficients
* sMODULAR	EQU 9			* Flag modular division

DISTDIV?	ST=0	sMODULAR	Not modular division
DDentry		ST=0	sGARB		No GC yet
		ST=1	sTRASH		Trash top-tempob if ZS-overflow

DDretry		GOSBVL	=SAVPTR		Re-entry point after GC

* Before trying to allocate any work memory we first check that
* all the coefficients are indeed short. This way we can avoid an
* unnecessary data allocation.

		?ST=1	sMODULAR	No need to test if we have a
		GOYES	+		modular division

		GOSUBL	QSHORT2?	CS if both short
		GOC	+

* Overflow occurred - return original polynomials and FALSE flag

DDoverflow	GOVLNG	=GPPushFLoop	( --> P1 P2 FALSE )


* Allocate working space. We require 1000 bytes since my instinct is that
* avoiding an imminent GC won't win as much as avoiding running out of
* memory during the division subroutine - possibly once we're almost
* finished. My guess is that going below 500 bytes would often prolong
* running time for the above reason.


+		GOSBVL	=ROOM
		A=C	A		A[A] = room
		LC(5)	2000		2000 nibbles required
		?A>C	A
		GOYES	+
DDmemerr	GOSBVL	=DOGARBAGE	Handles sGARB properly
		GOTO	DDretry

+		CSR	A		2000/16 = 125 (left for SOL)
		C=A-C	A
		R4=C	A		R4[A] = free work memory
		GOSBVL	=MAKE$N		Use $ for the benefit of DDmemerr

* Count the number of variables to be used in the distributed representations

		GOSUBL	QVARS2		A[A] = max(vars(P1),vars(P2))
		R3=A	A		R3[A] = vars

* Now:
*	R0[A] = ->$	= ->denom-10
*	R3[A] = vars
*	R4[A] = free


* Now we convert the polynomials into distributed form.
* The format used is that required by the division subroutine,
* namely denominator first, then one slot free from which
* the quotient grows, then nominator.

		A=DAT1	A
		D0=A			->P2
		C=R0	A
		D1=C
		D1=D1+	10		->output for denominator
		C=R3	A
		B=C	A		B[A] = vars
		GOSUB	Distribute	D1 = ->output tail (obeys sMODULAR)
		GOC	DDmemerr	CS: Insufficient memory

* Allocate the extra slot

		A=B	A
		A=A+A	A
		A=A+A	A
		A=A+B	A		A[A] = 5*vars
		A=A+CON	A,16		A[A] = slotsize
		C=R4	A
		C=C-A	A
		R4=C	A
		GOC	DDmemerr2	CS: Insufficient memory
		CD1EX
		C=C+A	A		Skip the slot
		D1=C
		R1=C	A		R1[A] = ->nominator

* And convert the nominator

		GOSBVL	=D0=DSKTOP
		D0=D0+	5
		A=DAT0	A		->P1
		D0=A
		GOSUB	Distribute	D1 = ->output tail
		GOC	DDmemerr2	CS: Insufficient memory
		CD1EX
		R2=C	A		R2[A] = ->nominator tail

* Call the division subroutine. If we are using modular math
* the divisor is always monic so we don't have to prepare anything here

		GOSUB	DDIVIDE
		GONC	+		CC: Success
DDmemerr2	GOTO	DDmemerr	CS: Insufficient memory

* Division done - convert results into recursive representation

* First move the quotient and remainder to end of work memory

+		A=R0	A		A[A] = ->$
		A=A+CON	A,5
		D0=A
		C=DAT0	A
		A=A+C	A
		D1=A			D1 = ->$tail
		D0=D0+	5		D0 = ->denominator

--		A=DAT0	W		A[W] = denominator coeff
		D0=D0+	16
		?A=0	W
		GOYES	+		Denominator skipped
		C=R3	A		C[A] = vars
-		D0=D0+	5		Skip power in slot
		C=C-1	A
		?C#0	A
		GOYES	-		Loop until all powers skipped
		GONC	--		Loop until end of denominator
+
		AD0EX			D0 = ->quotient
		C=R2	A		C[A] = ->nominator tail
		D0=C
		C=C-A	A		C[A] = nibbles in quo & nom
		GOSBVL	=MOVEUP		D1 = ->quotient (new location)

* Now do the conversions - output quotient to start of $

		A=R0	A		A[A] = ->$
		A=A+CON	A,10		A[A] = ->Pquo (output)
		CD1EX
		D0=C			D0 = ->input (quotient)
		D1=A			D1 = ->output (quotient)
		C=C-A	A
		R4=C	A		R4[A] = free memory (new)

		C=R3	A
		B=C	A		B[A] = vars
		GOSUB	Recurse		D1 = ->Pquo tail
		GOC	DDrecmemerr	CS: Insufficient memory
		CD1EX
		R2=C	A		R2[A] = ->Prem
		CD1EX
		AD0EX			Skip the "powers" in the terminator
		A=A+B	A		slot of the quotient.
		A=A+B	A
		A=A+B	A
		A=A+B	A
		A=A+B	A
		AD0EX
		GOSUB	Recurse		D1 = ->Prem tail
		GOC	DDrecmemerr	CS: Insufficient memory

* Shrink work memory

		AD1EX
		D0=A
		GOSBVL	=Shrink$	Uses R1[A] - thus R2[A] used above

* And return the results

		GOSBVL	=GETPTR
		D1=D1+	5
		A=R0	A
		A=A+CON	A,10		A[A] = ->Pquo
		DAT1=A	A		Overwrite P1
		D1=D1-	5
		A=R2	A		A[A] = ->Prem
		DAT1=A	A		Overwrite P2
		GOVLNG	=PushTLoop	( --> Pquo Prem TRUE )

* Insufficient memory in conversion to recursive form. Could optimize
* this to avoid a full rerun and just try the conversion again - but
* it does not seem to be worth the effort.

DDrecmemerr	GOTO	DDmemerr	Just try full restart


*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
* Name:		DDIVIDE
* Desc:		Divide two polynomials in recursive representation.
*
* Entry:	R0[A] = ->$	= ->denominator-10
*		extra slot between these two
*		R1[A] = ->nominator
*		R2[A] = ->nominator tail
*		R3[A] = vars
*		R4[A] = free memory
*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

DDIVIDE		A=R1	A
		D0=A			D0 = ->nominator
		C=R0	A
		D1=C
		D1=D1+	10		D1 = ->denominator
		A=DAT1	W		A[W] = lcoeff(denom)
		?A=0	W
		GOYES	ddiverror	Division by zero!
		D0=D0+	16		->nompower1
		D1=D1+	16		->denpower1

* First divide the powers. Note that the quotient slot is right before
* the nominator slot being operated with.

		C=R3	A
		D=C	A		D[A] = vars
		B=C	A
		B=B+B	A
		B=B+B	A
		B=B+C	A
		B=B+CON	A,16		B[A] = slotsize

--		A=DAT0	A		A[A] = nompower
		C=DAT1	A		C[A] = denpower
		A=A-C	A		A[A] = quopower
		GOC	ddivfinish	Overflow - division at end!
		CD0EX
		C=C-B	A		->quopower
		D0=C
		DAT0=A	A		Store power in quotient
		C=C+B	A		->nompower
		D0=C		
		D0=D0+	5		->nompower++
		D1=D1+	5		->denpower++
		D=D-1	A		vars--
		?D#0	A
		GOYES	--		Until all powers divided
		GONC	+		Continue to handle the coefficient

* Overflow in power division means we're done

ddivfinish	A=R1	A		A[A] = ->nominator
		A=A-B	A
		D0=A
		A=0	W		Terminate the quotient
ddivdone	DAT0=A	W
		RTNCC			CC: Successful division

* Division by zero or inexact coefficient division is an error

ddiverror	LC(5)	#304
		GOVLNG	=GPErrjmpC	"Undefined Result"

* Powers divided - handle the coefficient

+		A=R1	A
		D0=A			D0 = ->nominator
		C=R0	A
		D1=C
		D1=D1+	10		D1 = ->denominator
		A=DAT0	W		A[W] = nomcoeff
		C=DAT1	W		C[W] = dencoeff
		AD0EX
		A=A-B	A		->quocoeff
		AD0EX
		?A=0	W		If nominator is zero then we're done!
		GOYES	ddivdone
		?ST=1	sMODULAR	No division needed in modular
		GOYES	+		math, lcoeff is always 1
		GOSUBL	ZSDIV		A[W]=quotient	B[W]=remainder
		?B#0	W
		GOYES	ddiverror	remainder<>0 ==> error
+		DAT0=A	W		Store quocoeff

* Now calculate the new "nominator" = nom-quocoeff*den. The easy way to
* handle the memory limitations is to move the nominator to the end
* of the work space and output the new one into it's proper location.
* This way we also free memory along the way as the coefficients of
* the old nominator become unneeded.

		A=R0	A		A[A] = ->$
		A=A+CON	A,5
		D1=A
		C=DAT1	A
		A=A+C	A
		D1=A			D1 = ->$tail
		A=R2	A
		D0=A			D0 = ->nominator tail
		C=R1	A		C[A] = ->nominator
		C=A-C	A		C[A] = nibbles in nominator
		GOSBVL	=MOVEUP		D1 = ->nominator (new location)

		C=R3	A		C[A] = vars
		B=C	A
		B=B+B	A
		B=B+B	A
		B=B+C	A
		B=B+CON	A,16		B[A] = slotsize

* New nominator moves one slot to higher memory, but since we already
* know we'll just skip the first coefficient in the old nominator
* we also free up one slot - thus no need to update free memory in here.

		CD0EX			Skip first slot in nominator (new)
		C=C+B	A		   to make room for the next quotient
		D0=C			   coefficient
		R1=C	A		R1[A] = ->nominator (new)
		C=C-B	A
		C=C-B	A
		D=C	A		D[A] = ->quotient coefficient
		A=R0	A
		A=A+CON	A,10		A[A] = ->denominator
		A=A+B	A		Skip first slot in denominator
		AD1EX
		A=A+B	A		Skip first slot in nominator
		AD1EX


* Now:	D0   = ->nominator' (output)		(advances in loop)
*	D1   = ->nominator  (input)		(advances in loop)
*	D[A] = ->quotient coefficient
*	A[A] = ->denominator (input)		(advances in loop)

ddremloop	P=	0		Reset P from previous pass

		AD1EX
		C=DAT1	W		C[W] = nomcoeff
		AD1EX
		?C#0	W
		GOYES	+
		P=	1		P=1 : denominator traversed
+		C=DAT1	W		C[W] = dencoeff
		?C#0	W
		GOYES	++		nominator not traversed - continue
		?P#	1
		GOYES	+		P=2 : nominator traversed

* Both nominator and denominator traversed - the sum is done

		P=	0
		DAT0=C	W		Terminate the new nominator
		D0=D0+	16
		CD0EX
		R2=C	A		R2[A] = ->nominator tail
		GOTO	DDIVIDE

+		P=	2		P=2: nominator traversed

* Now output one new nominator term at a time

++		A=A+CON	A,16		->denpower
		D=D+CON	A,16		->quopower
		D0=D0+	16		->outpower
		D1=D1+	16		->nompower
		C=R3	A
ddnompowloop	RSTK=C			RSTK = vars counter
		AD0EX
		C=DAT0	A
		B=C	A		B[A] = denpower
		AD0EX
		CDEX	A		Swap C[A] and D1
		CD0EX
		CDEX	A
		C=DAT0	A		C[A] = quopower
		CDEX	A
		CD0EX			Swap C[A] and D1 back
		CDEX	A
		B=B+C	A		B[A] = denpower + quopower
		C=DAT1	A		C[A] = nompower
		
* Handle the exceptions

		?P=	1		If P=1 then denominator traversed
		GOYES	+		Then just use nompower
		CBEX	A		C[A] = denpower + quopower
		?P=	2		If P=2 then nominator traversed
		GOYES	+		Then just use denpower + quopower
		?B=C	A		Else if same power then
		GOYES	+		it certainly is valid for output
		P=	2		Else assume nominator has higher power
		?C>B	A		
		GOYES	+		It has
		P=	1		It hasn't - output quo*den
		CBEX	A		Fix the power to be output
+		DAT0=C	A		Output the power
		A=A+CON	A,5		->denpower++
		D=D+CON	A,5		->quopower++
		D0=D0+	5		->outpower++
		D1=D1+	5		->nompower++

		C=RSTK
		C=C-1	A		C[A] = vars--
		?C#0	A
		GOYES	ddnompowloop	Loop until all powers handled

* Now skip back to the start of the slot

		C=R3	A		C[A] = vars
		B=C	A
		B=B+B	A
		B=B+B	A
		B=B+C	A
		B=B+CON	A,16		B[A] = slotsize
		A=A-B	A		A[A] = ->denominator
		C=D	A
		C=C-B	A
		D=C	A		D[A] = ->quotient
		CD0EX
		C=C-B	A
		CD0EX			D0 = ->nominator (output)
		CD1EX
		C=C-B	A
		CD1EX			D1 = ->nominator (input)
		
* And calculate nom-quo*den ignoring the proper parts if P<>0


		?P=	1
		GOYES	ddignpart1
		C=A	A
		RSTK=C			RSTK = ->denominator
		CD0EX
		RSTK=C			RSTK = ->nominator (output)
		A=DAT0	W		A[W] = denominator
		?ST=0	sMODULAR
		GOYES	+		Establish modulo if needed
		D0=(5)	(=IRAM@)-4
		C=DAT0	A
		D0=C
		D0=(4)	#100
		C=DAT0	W
		B=C	W		B[W] = N
+		C=D	A
		D0=C			D0 = ->quotient
		C=DAT0	W		C[W] = quotient
		A=A+C	S		multiply the signs
		GONC	+
		A=0	S
+		R3=A.F	S		save sign
		A=0	S
		C=0	S
		?ST=0	sMODULAR
		GOYES	+		Call proper multiplication subroutine
		GOSUBL	SHORTMULMOD	A[W]=|quo*den| mod N	(always CC)
		A=R3.F	S		A[W]=quo*den mod N
		GONC	++
+		GOSUBL	ZSMULentry	A[W]=quo*den
++		A=-A-1	S		negate
		B=A	W		B[W] = -quo*den
		C=RSTK			C[A] = ->nominator (output)
		CD0EX
		D=C	A		D[A] = ->quotient
		C=RSTK
		A=C	A		A[A] = ->denominator

ddignpart1	ABEX	W		A[W] = -quo*dem (if calculated)
		?P=	2		If nominator traversed
		GOYES	++		then ignore +nom
		C=DAT1	W
		ACEX	W		A[W] = nom	C[W] = -quo*den
		?P#	0		Ignore addition if only nom
		GOYES	++
		?ST=1	sMODULAR
		GOYES	+
		GOSUBL	ZSADD		A[W] = nom-quo*dem
++		GOTO	ddignpart2	And continue

+		ACEX	W		A[W] = -quo*den
		C=B	A
		RSTK=C			RSTK = ->denominator
		CD1EX
		B=C	A		B[A] = ->nominator (input)
		D1=(5)	(=IRAM@)-4
		C=DAT1	A
		D1=C
		D1=(4)	#100
		C=DAT1	W		C[W] = N
		CBEX	W		B[W] = N
		D1=C			D1 = ->nominator (input)
		C=DAT1	W		C[W] = nom
		R3=A.F	S		Assume z1 sign will be correct
		?A#C	S
		GOYES	ddpart2modsub	Handle substraction
		A=0	S		Handle addition
		C=0	S
		A=A+C	W		Always CC since signs are clear
		?A<B	W
		GOYES	++
		A=A-B	W		Fix to modular value
		GONC	++
ddpart2modsub	A=0	S		Handle substraction
		C=0	S		Could comment out these two clears?
		?A>=C	W		Check which mantissa is bigger
		GOYES	+
		AR3EX.F	S
		A=-A-1	S		Negate answer
		AR3EX.F	S
		ACEX	W		Swap max to A[W]
+		A=A-C	W
++		?A=0	W
		GOYES	+		Answer is fine if it is zero
		A=R3.F	S		A[S] = sign
+		C=RSTK
		B=C	A		B[A] = ->denominator

ddignpart2	?ST=0	sMODULAR	Make symmetric if modulo active
		GOYES	++
		CD0EX
		RSTK=C
		D0=(5)	(=IRAM@)-4
		C=DAT0	A
		D0=C
		D0=(4)	#100
		C=DAT0	W		C[W] = N
		?A=0	S
		GOYES	+		coeff>=0 - ok
		A=0	S
		A=A-C	W		else make it positive mod p
		A=-A	W
+		C=C-A	W		then convert to mods p
		?A<=C	W
		GOYES	+
		A=C	W
		A=-A-1	S
+		C=RSTK
		D0=C
++		DAT0=A	W		Output the coefficient

		A=B	A		A[A] = ->denominator
		C=R3	A		C[A] = vars
		B=C	A
		B=B+B	A
		B=B+B	A
		B=B+C	A
		B=B+CON	A,16		B[A] = slotsize

* Advance the pointers for next loop

		C=DAT0	W		Don't advance nominator (output)
		?C#0	W		if the result was zero!
		GOYES	+
		CR4EX
		C=C+B	A		We won some free memory
		CR4EX		
		GONC	++
+		CD0EX
		C=C+B	A		->nominator++ (output)
		CD0EX
++		?P=	2		Don't advance nominator (input)
		GOYES	+		if nominator already traversed
		CD1EX
		C=C+B	A		->nominator++ (input)
		CD1EX
+		?P=	1		Don't advance denominator
		GOYES	+		if denominator already traversed
		A=A+B	A		->denominator++
+		GOTO	ddremloop

*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
* Name:		Distribute
* Desc:		Convert polynomial to distributed form
* Entry:	D0	->polynomial
*		D1	->output
*		B[A]	variables
*		R4[A]	free memory
*		IRAMBUFF contains modulo if sMODULAR is set
*		sDUMMYCOEFF indicates set coeffs=1
* Output:	D0	->polynomial tail	CS: Insufficient memory
*		D1	->output tail		CC: Success
*		B[A]	variables
* Alters:	A[W] B[W] C[W] D[W]
* Stack:	2
*
* Algorithm:	Initialize slot to zero.
*		Initialize depth to zero.
*		If list then initialize degree(depth) = length(list)
*		degree(depth) = degree(depth)-1
*		If semi then depth = depth-1 and loop
*		If dohstr then coeff(slot)=z and duplicate slot, advance
*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

Distribute	ST=0	sDUMMYCOEFF
Distribute1	CD1EX			* Initialize first slot
		D=C	A		D[A] = ->output (save)
		D1=C
		C=B	A		C[A] = vars
		C=C+C	A
		C=C+C	A
		C=C+B	A
		C=C+CON	A,16		C[A] = slotsize
		AR4EX
		A=A-C	A		Allocate the slot
		AR4EX
		RTNC			CS: Insufficient memory
		GOSBVL	=WIPEOUT
		C=D	A
		D1=C			D1 = ->output
		D=0	A		D[A] = depth = 0
		GOTO	DISTentry	Enter at proper place

* Distribution loop

DISTloop	AD1EX			A[A] = ->output
		D1=A
		C=D	A		C[A] = depth
		A=A+C	A
		C=C+C	A
		C=C+C	A
		A=A+C	A
		AD1EX
		D1=D1+	16-5		D1 = ->degree(depth)
		C=DAT1	A
		C=C-1	A		degree(depth)--
		GONC	DISTcontinue	No overflow - no SEMI either then
		D1=A			D1 = ->output
		D0=D0+	5		Skip SEMI
		D=D-1	A		D[A] = depth--
DISTnext	?D#0	A
		GOYES	DISTloop	Loop until depth=0
		A=0	W		Set the terminator value
		DAT1=A	W
		D1=D1+	16		Skip past the terminator
		RTNCC

DISTcontinue	DAT1=C	A		Store the decreased degree
		D1=A			D1 = ->output
		A=DAT0	A
		LC(5)	=DOHSTR
		?A=C	A
		GOYES	DISThstr	Handle coefficient

DISTentry	P=	16-5
-		BSL	W		Save B[A] and D[A]
		DSL	W
		P=P+1
		GONC	-
		GOSUBL	QLEN		A[A] = length of list
		P=	16-5
-		BSR	W		Restore B[A] and D[A]
		DSR	W
		P=P+1
		GONC	-
		CD1EX
		D1=C			D1 = ->output
		C=C+D	A
		C=C+D	A
		C=C+D	A
		C=C+D	A
		C=C+D	A
		CD1EX
		D1=D1+	16		->degree(depth)
		DAT1=A	A		Store the degree
		D1=C
		D=D+1	A		depth++
		D0=D0+	5		Skip the DOLIST
		GOTO	DISTloop	And continue

* Get the coefficient itself

DISThstr	D0=D0+	5		Skip DOHSTR
		C=DAT0	A
		C=C-CON	A,7		Zlen-7
		GONC	+		No overflow if nonzero
		D0=D0+	6		Skip remaining part of Z0
		GOTO	DISTloop	And continue
+		?ST=1	sDUMMYCOEFF
		GOYES	+
		?ST=1	sMODULAR
		GOYES	DISThstrmod
+		D0=D0+	5		Skip Zlen
		P=C	0
		A=0	W
		A=DAT0	WP		A[W] = mantissa
		P=	0
		AD0EX
		A=A+C	A		Skip mantissa (don't use P!)
		AD0EX
		D0=D0+	1
		A=DAT0	S		A[S] = sign
		D0=D0+	1		Skip sign
		?ST=0	sDUMMYCOEFF
		GOYES	+
		A=0	W		Set coeff=1 if sDUMMYCOEFF
		A=A+1	A
+		DAT1=A	W		Store the coefficient
		GOTO	DISTslotcp

DISThstrmod	D0=D0-	5		->coefficient
		GOSUB	MODSAVE		Save all but A[W] and D0
		C=DAT1	W		C[W] = N
		GOSUBL	ZMODSHORT	A[W] = |coeff| mod N	D[S]=sign
		C=D	S
		A=C	S		A[W] = coeff mod N
		GOSUB	MODRESTORE
		?A#0	W
		GOYES	+
		GOTO	DISTloop	Don't store a zero!
+		DAT1=A	W		Store the coefficient
		
* Now make new slot with the very same values

DISTslotcp	CD0EX
		RSTK=C			Save ->input
		AD1EX
		D0=A			D0 = ->slot
		C=B	A
		C=C+C	A
		C=C+C	A
		C=C+B	A
		C=C+CON	A,16		C[A] = slotsize
		AR4EX
		A=A-C	A		Allocate slot
		AR4EX
		GOC	DISTmemerr	Insufficient memory
		A=A+C	A
		D1=A			D1 = ->slot (next)
		GOSBVL	=MOVEDOWN	Copy
		C=RSTK
		CD0EX			D0 = ->input
		D1=C			D1 = ->output (the new slot)
*		A=0	W		Initialization of terminator
*		DAT1=A	W		not necessary - at exit only
		GOTO	DISTnext	And loop until list traversed

DISTmemerr	C=RSTK			Pop the saved D0
		RTNSC			CS: Insufficient memory

* Modular arithmetic utility
* Save D1, C[W], B[W], D[W], RSTK1, RSTK2, RSTK3, RSTK4 in IRAMBUFF
* A[W] is free for use

MODSAVE		AD1EX			A[A] = "D1"
		D1=(5)	=IRAM@
		A=DAT1	S		A[S] = base
		D1=(5)	#70110		SX address
		A=A+A	S
		GONC	+
		D1=(5)	#80110		GX address
+		DAT1=A	A		Save "D1"
		D1=D1+	5
		DAT1=C	W		Save C[W]
		D1=D1+	16
		C=B	W
		DAT1=C	W		Save B[W]
		D1=D1+	16
		C=D	W
		DAT1=C	W		Save D[W]
		D1=D1+	16
		C=RSTK
		A=C	A		A[A] = return address
		C=RSTK
		DAT1=C	A		Save RSTK1
		D1=D1+	5
		C=RSTK
		DAT1=C	A		Save RSTK2
		D1=D1+	5
		C=RSTK
		DAT1=C	A		Save RSTK3	
		D1=D1+	5
		C=RSTK
		DAT1=C	A		Save RSTK4
		D1=D1+	5
		D1=(2)	00		Return modulo address in D1
		PC=A			And return

* Modular arithmetic utility
* Restore D1, C[W], B[W], D[W], RSTK1, RSTK2, RSTK3, RSTK4 from IRAMBUFF
* A[W] is not changed

MODRESTORE	C=RSTK
		B=C	A		D[A] = return address
		D1=(5)	(=IRAM@)-4
		C=DAT1	A
		D1=C
		D1=(4)	#154		->RSTK4 save
		C=DAT1	A
		RSTK=C			Restore RSTK4
		D1=D1-	5
		C=DAT1	A
		RSTK=C			Restore RSTK3
		D1=D1-	5
		C=DAT1	A
		RSTK=C			Restore RSTK2
		D1=D1-	5
		C=DAT1	A
		RSTK=C			Restore RSTK1
		D1=D1-	16
		C=DAT1	W
		D=C	W		Restore D[W]
		D1=D1-	16
		C=DAT1	W
		CBEX	W		Restore B[W]
		RSTK=C			RSTK = return address
		D1=D1-	16
		C=DAT1	W		Restore C[W]
		D1=D1-	5
		RSTK=C			Save C[A]
		C=DAT1	A
		D1=C			Restore D1
		C=RSTK			Restore C[A]
		RTN

*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
* Name:		Recurse
* Desc:		Convert distributed polynomial into the standard
*		recursive representation.
* Entry:	D0	->polynomial (distributed)
*		D1	->output
*		B[A]	vars
*		R4[A]	free memory
*		sMODULAR is ignored since we require all math was
*		done modularly anyway.
* Output:	D0	->polynomial tail	CS: Insufficient memory
*		D1	->output tail		CC: Success
*		B[A]	vars
* Alters:	A[W] B[W] C[W] D[W]
* Stack:	1
* Variables:
*		vars	= total number of variables
*		vars1	= variables in previous term
*		vars2	= variables in current term
*		same	= number of equal leading powers
*		diff	= the change in the differing power (decrease)
*
* Algorithm:
*	
*	The number of variables in a term is the number of trailing
*	zero powers substracted from the total number of variables.
*
*	Assume we have processed the previous term. Then
*
*	Number of lists to terminate = vars1-same-1
*	  o Terminating a list means traversing the powers of the
*	    previous term backwards - skipping over the trailing
*	    zeros first - and outputting as many zeros as the power
*	    indicates - followed by a SEMI.
*	Number of zeros to output = diff-1
*	  o This is for the change in the degree
*	Number of lists to start = vars2-same-1 (if>0)
*	  o Simply output DOLIST's
*	Output the coefficient
*	  o Always nonzero
*	Number of lists to end = -(vars2-same-1) (if>0)
*	  o This terminates any list that had a nonzero coefficient
*	    besides the ones having zero powers.
*	    For example:  { { 1 0 0 0 1 } 1 1 }
*	                                ^--------- output 1 SEMI
*
*	Above seems to be the simplest possible algorithm (uh!).
*
*	Special cases:
*
*	Leading nonzero coefficient is handled properly if we start the loop
*	with variables vars1=0, same=-1, diff=1.
*
*	The terminator is handled properly if we execute the loop with
*	vars2=0, same=-1, diff=1 and then check for a zero coefficient
*	when outputting it is due.
*
*	Above fails for a zero polynomial so we must check for it first.
*
*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

Recurse		B=0	S		B[S] = 0  indicating no output yet
		D=0	A		D[A] = vars1 = 0
		A=DAT0	W		A[W] = coeff1
		?A#0	W
		GOYES	RECURloop	If nonzero then start the loop

		D0=D0+	16		Skip coeff
		CR4EX
		C=C-CON	A,11		Allocate Z0
		CR4EX
		RTNC			CS: Insufficient memory
		LCZ0			Output Z0
		DAT1=C	11
		D1=D1+	11
		RTNCC			CC: Success

* The main loop

RECURloop	A=DAT0	W		A[W] = coefficient
		?A#0	W
		GOYES	+		Continue if not terminator
		C=0	A		C[A] = vars2 = 0
		A=0	W		A[A1] = zeros = 0
		A=A-1	A		A[A] = same = -1
		GOTO	RECURnow	Handle the variables

+		CD0EX			Count the number of variables
		RSTK=C			in the new term
		C=C+B	A
		C=C+B	A
		C=C+B	A
		C=C+B	A
		C=C+B	A
		D0=C
		D0=D0+	16		D0 = ->slot tail
		A=B	A		A[A] = vars
-		D0=D0-	5		D0 = ->power--
		C=DAT0	A		C[A] = power
		?C#0	A
		GOYES	+		Done if found nonzero
		A=A-1	A		A[A] = vars--
		?A#0	A
		GOYES	-		Loop until all checked
+		C=RSTK
		D0=C			D0 = ->slot
		C=A	A		C[A] = vars2

		?B#0	S		Special variables if first one
		GOYES	+		Not first one - continue
		B=B+1	S		B[A]<>0 - got first coefficient
		A=0	W		A[A1] = zeros = 0
		A=A-1	A		A[A] = same = -1
		GOTO	RECURnow	Handle the variables

* Now compute the number of equal leading powers (variable 'same')

+		GOSBVL	=CSLW5		C[A1] = vars2
		CD0EX
		D0=C			D0 = ->slot
		D0=D0+	16		
		C=C-B	A
		C=C-B	A
		C=C-B	A
		C=C-B	A
		C=C-B	A
		CD1EX			D1 = ->previous slot
		P=	16-5
-		CSL	W		Save D1 in C[A1] (vars2 in C[A2])
		DSL	W		Save vars1 in D[A1] (also D[A]=0)
		P=P+1
		GONC	-	
-		A=DAT1	A		A[A] = oldpower
		C=DAT0	A		C[A] = newpower
		D0=D0+	5		->oldpower++
		D1=D1+	5		->newpower++
		D=D+1	A		D[A] = same = same+1
		?A=C	A		Continue until change found
		GOYES	-		Guaranteed to terminate!
		D=D-1	A		D[A] = same (fix back by one)
		A=A-C	A		A[A] = diff
		A=A-1	A		A[A] = diff-1 = zeros
		C=D	A		C[A] = same
		P=	16-5
-		ASL	W		A[A1] = zeros
		DSR	W		D[A] = vars1
		P=P+1
		GONC	-		
		A=C	A		A[A] = same
		GOSBVL	=CSRW5		C[A] = ->output	C[A1] = vars2
		D1=C
		GOSBVL	=CSRW5		C[A] = vars2

		CD0EX			Skip back to start of slot
		C=C-A	A		-5*same nibbles
		C=C-A	A
		C=C-A	A
		C=C-A	A
		C=C-A	A		
		CD0EX
		D0=D0-	16		-16 for the coefficient
		D0=D0-	5		-5 for the extra D0=D0+ 5

* Now handle the variables as instructed in the heading
*
*	B[A]	= vars		A[A]	= same
*	D[A]	= vars1		A[A1]	= zeros (diff-1)
*	C[A]	= vars2

RECURnow

* Terminate vars1-same-1 lists

		CDEX	A		D[A] = vars2 (to be new vars1)
		C=C-A	A
		C=C-1	A		C[A] = vars1-same-1
		?C#0	A
		GOYES	+		If nonzero then do terminate
		GOTO	RECURzeros	Else skip to outputting the zeros

+		B=B-C	A		vars-(vars1-same-1)
		B=B-A	A		vars-(vars1-same-1)-same
		B=B-1	A		vars-vars1
		CD0EX
		RSTK=C			RSTK = ->input
		C=C-B	A		Skip back over the trailing
		C=C-B	A		zero powers in previous slot
		C=C-B	A
		C=C-B	A
		C=C-B	A
		CD0EX
		B=B+1	A		Calculate vars back from
		B=B+A	A		vars-vars1
		B=B+C	A
		P=	16-5
-		ASL	W		A[A1] = same	A[A2] = zeros
		DSL	W		D[A1] = vars2
		P=P+1
		GONC	-
		D=C	A		D[A] =vars1-same-1

		LCZ0			Output [power] zeros
--		D0=D0-	5
		A=DAT0	A		A[A] = power
-		A=A-1	A		power--
		GOC	+
		CR4EX
		C=C-CON	A,11		Allocate Z0
		CR4EX
		GOC	RECURmem1	CS: Insufficient memory
		LC(5)	=DOHSTR		Output Z0
		DAT1=C	11
		D1=D1+	11
		GONC	-
+		CR4EX
		C=C-CON	A,5		Allocate SEMI
		CR4EX
		GOC	RECURmem1	CS: Insufficient memory
		LC(5)	=SEMI		Output SEMI
		DAT1=C	A
		D1=D1+	5
		D=D-1	A		D[A] = (vars1-same-1)--
		?D#0	A
		GOYES	--		Loop until terminated all
		P=	16-5
-		ASR	W		A[A] = same	A[A1] = zeros
		DSR	W		D[A] = vars2
		P=P+1
		GONC	-
		C=RSTK
		D0=C			D0 = ->input

* Output diff-1 zeros

RECURzeros	C=A	A		C[A] = same
		GOSBVL	=ASRW5		A[A] = zeros = diff-1
		RSTK=C			RSTK = same
		LCZ0
-		A=A-1	A		zeros--
		GOC	+
		CR4EX
		C=C-CON	A,11		Allocate Z0
		CR4EX
RECURmem1	GOC	RECURmemerr	CS: Insufficient memory
		DAT1=C	11
		D1=D1+	11
		GONC	-	
+		C=RSTK			C[A] = same
	
* Start vars2-same-1 new lists

		A=C	A		A[A] = same
		A=A+1	A		A[A] = same+1
		C=D	A		C[A] = vars2
		C=C-A	A		C[A] = vars2-same-1
		GOC	++		If overflow then none!
		A=C	A		A[A] = vars2-same-1
		RSTK=C			RSTK = vars2-same-1
		LC(5)	=DOLIST
-		A=A-1	A		(vars2-same-1)--
		GOC	+
		CR4EX
		C=C-CON	A,5		Allocate DOLIST
		CR4EX
		GOC	RECURmemerr	CS: Insufficient memory
		DAT1=C	A
		D1=D1+	5
		GONC	-
RECURmemerr	C=RSTK			Pop saved pointer
		RTNSC			CS: Insufficient memory
RECURdone	RTNCC
+		C=RSTK

* Output coefficient

++		A=DAT0	W		A[W] = coefficient
		D0=D0+	16		Skip it
		?A=0	W
		GOYES	RECURdone	If zero then done!
		RSTK=C			C[A] = vars2-same-1
		CR4EX
		C=C-CON	A,5		Allocate DOHSTR
		CR4EX
		GOC	RECURmemerr	CS: Insufficient memory
		LC(5)	=DOHSTR		Output DOHSTR
		DAT1=C	A
		D1=D1+	5
		P=	15
-		P=P-1			Trim the coefficient
		?A=0	P
		GOYES	-
		C=0	A
		C+P+1
		C=C+CON	A,6		C[A] = Zlen
		AR4EX
		A=A-C	A		Allocate Zlen and body
		AR4EX
RECURmem2	GOC	RECURmemerr	CS: Insufficient memory
		DAT1=C	A		Output Zlen
		D1=D1+	5
		DAT1=A	WP		Output mantissa
		CD1EX
		C+P+1
		CD1EX
		DAT1=A	S		Output sign
		D1=D1+	1
		P=	0
		C=RSTK
* Output -(vars2-same-1) lists
		A=C	A		A[A] = vars2-same-1
		C=C+C	A		CS if negative
		GONC	+		If not then slot done
		LC(5)	=SEMI		Else output SEMI's
-		CR4EX
		C=C-CON	A,5		Allocate SEMI
		CR4EX
		GOC	RECURmem2	CS: Insufficient memory
		DAT1=C	A
		D1=D1+	5
		A=A+1	A		-(vars2-same-1)--
		GONC	-		Loop until done

* Slot done - skip to next one and loop

+		CD0EX			->slot
		C=C+B	A
		C=C+B	A
		C=C+B	A
		C=C+B	A
		C=C+B	A
		CD0EX			->slot++
		GOTO	RECURloop
RPL		
**********************************************************************

