	STITLE	ALG GCD Algorithm
**********************************************************************
*		ALG Library:	Polynomial/Integer GCD routines
**********************************************************************

**********************************************************************
* EXPORT:
*
* QGcd		( Q1 Q2 --> GCD[Q1,Q2] )
* QLcm		( Q1 Q2 --> LCM[Q1,Q2] )
**********************************************************************

**********************************************************************
* Desc:		Compute the polynomial GCD of Q1 and Q2
*		The result is trimmed.
* Stack:	( Q1 Q2 --> GCD[Q1,Q2] )
**********************************************************************
NULLNAME QGcd
::
  QGcd?
  DUP FALSE EQ NOT?SEMI
  ErrUndefRes
;

**********************************************************************
* Desc:		Compute the polynomial LCM of Q1 and Q2.
*		The result is trimmed.
* Stack:	( Q1 Q2 --> LCM[Q1,Q2] )
**********************************************************************
NULLNAME QLcm
::
  2DUP QGcd		( Q1 Q2 G )
  QDiv QMul PTrim	( L )
;

**********************************************************************
* Desc:		Compute the polynomial GCD of Q1 and Q2 using heuristic.
*		Returns FALSE if GCD is not found.
*		The result is trimmed.
* Stack:	( Q1 Q2 --> GCD[Q1,Q2] / FALSE )
**********************************************************************
NULLNAME QGcd?
::
  CODE
		GOSBVL	=SAVPTR

		C=DAT1	A		* Get polynomials
		D1=D1+	5
		A=DAT1	A
		D0=A			->Q1
		D1=C			->Q2

		C=A	W		Load same high nibbles
		LCZ0
		A=DAT0	11
		?A=C	W
		GOYES	gcdswapdrop	GCD[0,any] = any
		A=DAT1	11
		?A=C	W
		GOYES	gcddrop		GCD[any,0] = any

		LCZ1
		A=DAT0	12
		?A=C	W
		GOYES	gcddrop		GCD[1,any] = 1
		A=DAT1	12
		?A=C	W
		GOYES	gcdswapdrop

		?A=C	A		Dispatch based on type
		GOYES	qgcdz
		A=DAT0	A
		?A=C	A
		GOYES	zgcdp
pgcdp		LC(5)	(PgcdP)-(gcdoffset)
		GONC	+
zgcdp		LC(5)	(ZgcdP)-(gcdoffset)
		GOC	+
qgcdz		A=DAT0	A
		?A#C	A
		GOYES	pgcdz
zgcdz		LC(5)	(ZgcdZ)-(gcdoffset)
		GONC	+
pgcdz		LC(5)	(PgcdZ)-(gcdoffset)

+		A=PC
gcdoffset	A=A+C	A
		GOVLNG	=GPPushALp


gcddrop		LC(5)	=DROP
		GOTO	+
gcdswapdrop	LC(5)	=SWAPDROP
+		GOVLNG	=GETPTREVALC

  ENDCODE
  EVAL

* The trick in above is that EVAL ignores lists, but evaluates the possibly
* pushed dispatchee. Programmed this way for benefit of SDB only

;

**********************************************************************
* Calculate GCD[P,Z]
**********************************************************************
LABEL PgcdZ
:: SWAP PZCont ZGcd ;

**********************************************************************
* Calculate GCD[Z,P]
**********************************************************************
LABEL ZgcdP
:: PZCont ZGcd ;

**********************************************************************
* Calculate GCD[A,B] (polynomials)
**********************************************************************
LABEL PgcdP
::
   PPZ ROT PPZ ROT ZGcd UNROT		( Zgcd A' B' )
*  PGcdTriv UNROT			( gcd A' B' )
*  DUPTYPEHSTR? 3PICK TYPEHSTR?
*  ORcase 2DROP				( --> gcd )

   DUP PNMax 3PICK PNMax			( c A B bm am )
   ZNMin ZAbs					( c A B m )
   HXS 2 20 QMul HXS 2 20 QAdd			( c A B 2m+2 )
   FIVE SWAP					( c A B k z )
   BEGIN
	?ATTNQUIT
	4PICKOVER PEval				( c A B k z Az )
	4PICK 3PICK PEval			( c A B k z Az Bz )
	QGcd?					( c A B k z Gz/FALSE )
	DUP FALSE EQ ITE DROP
	::
		OVER PZadic			( c A B k z G'' )
		DUPTYPELIST?
		NOTcase :: 5DROP 2RDROP ;	( return c )
		PPZ DROP			( c A B k z G' )
		5PICK OVER QDiv?		( c A B k z G' [A/G' T/F] )
		SWAPDROP NOT?SEMI		( exit ITE with c A B k z )
		4PICK OVER QDiv?		( c A B k z G' [B/G' T/F] )
		SWAPDROP NOT?SEMI		( exit ITE with c A B k z )
		6UNROLL 4DROP			( G' c )
		QMul				( G )
		2RDROP				( return )
	;
	SWAP#1-					( c A B z k-1 )
	DUP#0<>
   WHILE
	SWAP					( c A B k-1 z )
	HXS 6 240210 QMul			( c A B k-1 z'=z*73794 )
	HXS 5 38960 QDiv			( c A B k-1 z''=z'/27011 )
   REPEAT
   5DROP FALSE					( FALSE = failure )
;

**********************************************************************
* Calculate GCD[P,P] for short univariate polynomials
* Algorithm is the polynomial remainder sequence (p.r.s.)
* Possible improvement would be to use the sub-resultant p.r.s.
* which makes coefficient growth linear.
*
* Note: Coefficient explosion makes this very very bad when the leading
*	coefficients are not one, thus this is not used.	
**********************************************************************
*	::
*	  PPZ ROT PPZ ROT ZGcd UNROT		( content A B )
*	
*	  OVER LENCOMP OVER LENCOMP		( sort by degree )
*	  #< IT SWAP
*	
*	  BEGIN
*		DUP CARCOMP			( c A B B[1] )
*		3PICK LENCOMP			( c A B B[1] deg[A] )
*		3PICK LENCOMP			( c A B B[1] deg[A] deg[B] )
*		#-+1 PPow#			( c A B k )
*		ROT QMul			( c B A*k )
*		OVER QDivRem SWAPDROP		( c B rem )
*		DupQIsZero? casedrop
*		:: RDROP PPZ DROP QMul ;	( --> gcd )
*		PTrim
*		DUPTYPEHSTR? ?SKIP
*	  AGAIN
*	  RDROP 2DROP				( --> c )
*	;
	
**********************************************************************
* Calculate GCD[Z,Z]
**********************************************************************
NULLNAME ZGcd
LABEL ZgcdZ
::
  CODE
	GOTO	FastGcd?
  ENDCODE
  ?SEMI					( Done already )

* Atleast one integer too big for a register. Use slow division
* until both integers fit in

  ZAbs SWAP ZAbs			( N2 N1 )
  BEGIN
	DupQIsZero? NOT
  WHILE					( A B )
	SWAPOVER QDivRem		( B A/B AmodB )
	SWAPDROP			( B AmodB )
	CODE
		GOTO	FastGcd?
	ENDCODE
	case :: 2RDROP ;		( FastGcd finished )
  REPEAT
  DROP
;


ASSEMBLE

* Desc:		FastGcd? tries to calculate GCD[Z1,Z2] in registers
*		If success then pushes GCD & TRUE, else FALSE

FastGcd?	GOSBVL	=SAVPTR
		GOSUBL	SHORTPOPZ?	A[W]C[S]=Z2
		GOC	fastgcdfail	Not short integer - fail
		?A=0	W
		GOYES	fastgcddrop1	GCD[any,0] = any
		B=A	W		B[W] = N2
		GOSUBL	SHORTPOPZ?	A[W]C[S]=Z1
		GOC	fastgcdfail
		?A=0	W
		GOYES	fastgcddrop2	GCD[0,any]=any
		C=B	W

* Suspect following might be better:
*		GOSUBL	SHORTGCD
-		GOSUBL	BDIV		B[W]=remainder	C[W]=divisor
		A=C	W		A[W]=divisor
		C=B	W		C[W]=remainder
		?C#0	W
		GOYES	-
		GOSUBL	PushN		Push GCD
		A=DAT1	A		Overwrite 2nd stack level with Z
		D1=D1+	10
		D=D+1	A
		GONC	fastgcdtrue

fastgcdfail	GOVLNG	=GPPushFLoop	( Z1 Z2 --> Z1 Z2 FALSE )

fastgcddrop1	GOVLNG	=GPOverWrTLp	( Z1 0 --> Z1 TRUE )

fastgcddrop2	GOSBVL	=GETPTR		( 0 Z2 --> Z2 TRUE )
		A=DAT1	A
		D1=D1+	5
fastgcdtrue	DAT1=A	A
		D1=D1-	5
		GOVLNG	=OverWrTLoop

RPL
**********************************************************************
