	STITLE	ALG Z Poly
**********************************************************************
*		ALG Library: Integer Polynomial Routines
**********************************************************************

**********************************************************************
* EXPORT:
*
*   CKMkP	( pn .. p1 n -> P )
*   MkCstP	( Q #n -> Q' )
*   MkSplP	( #n #k -> P )
*
*   DoP   	( P proc --> P' )
*   DoPS   	( P S proc --> P' )
*
*   PZCont	( P --> Z )
*   PPZ		( P --> PP Z )
*
*   PEval	( Pn Z --> Pn-1 )
*   PZadic	( Pn-1 Z --> Pn )
*   PPP		( P --> PP PC )
*
*   PPow#	( P # --> P^# )
*   PDer	( P --> P' )
*
*   PSetSign	( P1 P2 --> sign[P2]*P1 )
*   PLCZ	( P --> Z )
*   PTrim	( P --> P' )
*   Variables	( P --> P #vars )
**********************************************************************


**********************************************************************
* Desc:		Make polynomial from coeffs
* Stack:	( pn .. p1 n --> P )
**********************************************************************
NULLNAME CKMkP
::
  DUP#1= NOTcase {}N
  DROP DUPTYPELIST?
  case ONE{}N
;

**********************************************************************
* Desc:		Make constant polynomial with + n variables
* Stack:	( Q #n -> Q' )
**********************************************************************
NULLNAME MkCstP
::
  DUP#0=csDROP			( leave unchanged if n=0 )
  OVER TYPEHSTR? caseDROP	( leave unchanged if Z )
  ZERO_DO (DO)
    ONE{}N
  LOOP
;

**********************************************************************
* Desc:		Make polynomial of nth var to the power k
* Stack:	( #n #k -> P )
**********************************************************************
NULLNAME MkSplP
CODE
		GOSBVL	=POP2#		A[A]=n	C[A]=k
		?A#0	A		n=0?
		GOYES	makervnow
		GOVLNG	(=SETTYPEERR)+5	Cannot be 0th variable!
makervnow	?C#0	A
		GOYES	+		Nonzero power - continue
		GOSBVL	=SAVPTR
		A=0	W
		A=A+1	A		A[W]=1
		GOLONG	PushNLoop	Push Z1

+		R0=A			n
		R1=C			k
		GOSBVL	=SAVPTR

* Needed size = k*size(Z0)+10n+size(Z1) = 11k+10n+12 =k+10(n+k)+12

		C=R1			k
		A=A+C	A		n+k
		A=A+A	A		2(n+k)
		B=A	A		2(n+k)
		B=B+B	A		4(n+k)
		B=B+B	A		8(n+k)
		A=A+B	A		10(n+k)
		C=C+A	A		10(n+k)+k
		C=C+CON	A,12		10(n+k)+k+12
		GOSBVL	=GETTEMP
		AD0EX
		D0=A			->list
		AR0EX
		A=A-1	A		n-1
		B=A	A		n-1
		LC(5)	=DOLIST
makervlists	DAT0=C	A
		D0=D0+	5
		A=A-1	A
		GONC	makervlists
		LCZ1
		DAT0=C	12
		D0=D0+	12
		A=R1			k
		A=A-1	A
		GOC	makervnoZ0s
		LCZ0
makervZ0s	DAT0=C	11
		D0=D0+	11
		A=A-1	A
		GONC	makervZ0s
makervnoZ0s	LC(5)	=SEMI
makervsemis	DAT0=C	A
		D0=D0+	5
		B=B-1	A
		GONC	makervsemis
		GOVLNG	=GPPushR0Lp
ENDCODE

**********************************************************************
* Desc:		Apply proc to every elem of P
* Stack:	( P proc --> P' )
**********************************************************************
NULLNAME DoP
::
  1LAMBIND
ASSEMBLE
AlgDoPAgain
RPL
  ::
    DUPTYPELIST? NOTcase :: 1GETLAM EVAL ;
    FIVE
    BEGIN
      NEXTCOMPOB
    WHILE
	CODE
		A=PC
		LC(5)	(AlgDoPAgain)-(*)
		A=A+C	A
		PC=(A)
	ENDCODE
	UNROT
    REPEAT
    LENCOMP {}N
  ;
  ABND
;


**********************************************************************
* Desc:		Apply S proc to every (scalar) elem of P
* Stack:	( P S proc --> P' )
**********************************************************************
NULLNAME DoPS
::
  { NULLLAM NULLLAM } BIND
ASSEMBLE
AlgDoPSAgain
RPL
  ::
    DUPTYPELIST? NOTcase :: 2GETLAM 1GETLAM EVAL ;
    FIVE
    BEGIN
      NEXTCOMPOB
    WHILE
      CODE
		A=PC
		LC(5)	(AlgDoPSAgain)-(*)
		A=A+C	A
		PC=(A)
      ENDCODE
      UNROT
    REPEAT
    LENCOMP {}N    
  ;
  ABND
;


**********************************************************************
* Desc:		Compute the # power of P.
*		The result is trimmed (provided P was)
* Stack:	( P # --> P^# )
**********************************************************************
NULLNAME PPow#
ASSEMBLE
MAXPOW	EQU 8				P^8 still has special code

		CON(5)	=DOCOL		Provide leading DOCOL
		CON(5)	=DOCODE
		REL(5)	->PowEnd
		GOSBVL	=POP#
		C=0	A
		LC(1)	(MAXPOW)+1
		?A<C	A
		GOYES	powok
		A=C	A		Max dispatchee
		D1=D1-	5		Unpop #
		D=D-1	A
powok		GOSUB	passpowtable	Dispatch given A[A]
		REL(3)	pow#0
		REL(3)	pow#1
		REL(3)	pow#2
		REL(3)	pow#3
		REL(3)	pow#4
		REL(3)	pow#5
		REL(3)	pow#6
		REL(3)	pow#7
		REL(3)	pow#8
		REL(3)	pow#big		The common code for # > MAXPOW
passpowtable	C=RSTK
		C=C+A	A
		C=C+A	A
		C=C+A	A
		D0=C			Old D0 is trashed!
*		A=0	A		Not needed for small MAXPOW
		A=DAT0	X
		A=A+C	A		dispatch address
		D0=A			Continue RPL there!
		GOVLNG	=Loop
RPL

* Dispatchees (Hopefully Moz enables the LABEL flag..)

LABEL pow#0	DROP HXS 2 10
LABEL pow#1	[;]
LABEL pow#8	DUP QMul
LABEL pow#4	DUP QMul
LABEL pow#2	DUP QMul [;]
LABEL pow#6	DUP QMul
LABEL pow#3	DUPDUP QMul QMul [;]
LABEL pow#5	DUPDUP QMul DUP QMul QMul [;]
LABEL pow#7	DUPDUP QMul DUPDUP QMul QMul QMul [;]

* For big powers a logarithmic algorithm is used

LABEL pow#big
	HXS 2 10 SWAP			( Zn R # )
	BEGIN
	  DUP #2/ 4UNROLL		( #' Zn R # )
	  ONE #AND #0=			( #' Zn R notfactor? )
	  ?SKIP :: OVER QMul ;		( #' Zn R' )
	  3PICK #0=?SKIP		( !! Avoid extra * if end of loop !! )
	  :: SWAPDUP QMul SWAP ;	( #' Zn*Zn R' )
	  ROT				( Zn*Zn R' #' )
	#0=UNTIL (UNTIL)
	DROPSWAPDROP			( Result )
ASSEMBLE
->PowEnd	CON(5)	=SEMI			Embedding SEMI
RPL

**********************************************************************
* Desc:		Evaluates Pn at X1=Z. The result is trimmed
* Stack:	( Pn Z --> Pn-1 )
**********************************************************************
NULLNAME PEval
::
  OVER TYPELIST? NOTcaseDROP
  SWAP				( Z Pn )
  PEvalFast?			( Z Pn FALSE / Pn[Z] TRUE )
  ?SEMI				( --> P[Z] )
  FIVE				( Z L start-offset )
  NEXTCOMPOB DROP		( Z L offset cn )
  BEGIN
	4UNROLL			( C Z L offset )
	NEXTCOMPOB
  WHILE				( C Z L offset' ci )
	5ROLL			( Z L offset' ci C )
	5PICK			( Z L offset' ci C z )
	QMul			( Z L offset' ci zC )
	QAdd			( Z L offset' C' )
  REPEAT
  2DROP PTrim
;

**********************************************************************
* Desc:		Attempts to evaluate Pn at X1=Z using fast
*		register arithmetic.
* Stack:	( Z Pn --> Z Pn FALSE / Pn[Z] TRUE )
* Notes:	Fails if
*		- Pn is not sunivariate
*		- Z is polynomial after all
*		- Z size is too big for register
*		- Any overflow occurs during horner evaluation
**********************************************************************
NULLNAME PEvalFast?
CODE
		GOSBVL	=SAVPTR

		C=DAT1	A		->P
		D0=C
		GOSUBL	SUNIVAR?	CS if sunivariate
		GONC	phornfail	Not sunivariate - fail

		C=DAT1	A		->P
		D1=D1+	5
		A=DAT1	A		->Z
		D1=C			->P
		D0=A			->Z

		A=DAT0	A		Check Z is really Z
		LC(5)	=DOHSTR
		?A#C	A
		GOYES	phornfail
		D0=D0+	5		Check if Z seems small enough
		A=DAT0	A
		LC(5)	5+7+1		Allow 7 nibble mantissa
		?A<=C	A		(reasonable for quadratic)
		GOYES	+		Short enough - try PEval

phornfail	GOVLNG	=GPPushFLoop	Failed to eval - simply continue

+		D0=D0-	5		->Z
		GOSUBL	SHORTPOP	A[W] C[S] = X

		R0=A	W		|Z|
		R1=C	S		Zsign

		B=0	W		result mantissa
		D=0	S		result sign
		D1=D1+	5		->first coefficient

phornloop	GOSUBL	SHORTPOPD1	A[W] C[S] = Z	CS if zero
		GOC	phornnext	No need to add Z0 to result
		CBEX	W		B[S] = Zsign	C[W]=result
		GOSUBL	SHORT+		A[W] C[S] = result
		GOC	phornfail	Overflow - failure
		B=A	W
		D=C	S		Save result
phornnext	A=DAT1	A		We're done if last coefficient
		LC(5)	=SEMI		was just added
		?A=C	A
		GOYES	phorndone

		A=B	W		A[W] = result
		C=R1.F	S
		CDEX	S		D[S] = Zsign
		B=C	S		B[S] = result sign
		C=R0	W		C[W] = Z
		GOSUBL	SHORT*		A[W] C[S] = result'
		GOC	phornfail	Overflow - failure
		B=A	W
		D=C	S		Save result
		GOTO	phornloop

phorndone	A=B	W		A[W] = result
		C=D	S		C[S] = result sign
		GOSUBL	PushZ		Push the result
		A=DAT1	A
		D1=D1+	10		DROP2
		D=D+1	A
		D=D+1	A
		DAT1=A	A		Return result
		GOVLNG	=PushTLoop	And TRUE to indicate result
ENDCODE

**********************************************************************
* Desc:		Interpolate Pn from its Z-adic representation Pn-1
*		The result is trimmed
* Stack:	( Pn-1 Z --> Pn )
**********************************************************************
NULLNAME PZadic
::
  ZERO ROT				( Z 0 A )
  BEGIN					( gi-1 .. g0 Z i Ai )
  	SWAP#1+SWAP			( gi-1 .. g0 Z i+1 Ai )
  	DUP 4PICK			( gi-1 .. g0 Z i+1 Ai Ai Z )
  	QMod				( gi-1 .. g0 Z i+1 Ai gi' )
  	DUP 4PICK			( gi-1 .. g0 Z i+1 Ai gi gi i+1 )
	#4+ UNROLL			( gi gi-1 .. g0 Z i+1 Ai gi )
	QSub				( gi gi-1 .. g0 Z i+1 Ai-gi )
	3PICK QDiv			( gi gi-1 .. g0 Z i+1 Ai+1' )
	DupQIsZero?			( gi gi-1 .. g0 Z i+1 Ai+1 flag )
  UNTIL
  DROPSWAPDROP				( gi .. g0 i+1 )
  CKMkP					( G )
;

**********************************************************************
* Desc:		Compute the integer content of P
* Stack:	( P --> Z )
**********************************************************************
NULLNAME PZCont
::
  DUPTYPELIST? NOTcase ZAbs		( Return |Z| )

* Try to do it all in registers first. Always succeeds if there
* is atleast one non-zero short coefficient.

  CODE
sGCD1st	EQU 9	* Flag to indicate first coefficient found
sGCDbig	EQU 8	* Flag to indicate a big coefficient was skipped
sGCDpas	EQU 7	* Flag to indicate 2nd pass

		GOSBVL	=SAVPTR
		CLRST			Clear flags
pzcontrerun	A=DAT1	A
		D0=A			D0 = ->P
		A=DAT0	A
		LC(5)	=DOLIST
		?A=C	A
		GOYES	pzcontlist

* Integer argument - just do ZAbs

		GOSBVL	=GETPTR
		GOLONG	dozabs

* List argument - calculate the GCD

pzcontlist	D0=D0+	5		Skip DOLIST
		C=0	A
pzcontloop	RSTK=C			depth=0
pzcontmain	A=DAT0	A
		D0=D0+	5
		LC(5)	=DOHSTR		#02A4E
		?A=C	A
		GOYES	pzcontz
		LC(2)	=DOLIST		#02A74
		?A#C	A
		GOYES	pzcontsemi
		C=RSTK
		C=C+1	A		depth++
		GONC	pzcontloop
pzcontsemi	C=RSTK
		C=C-1	A		depth--
		GONC	pzcontloop
		GOTO	pzcontexit

pzcontz		C=DAT0	A		C[A] = Zlen
		C=C-CON	A,7
		GONC	+
		D0=D0+	6		Skip remainder of Z0
		GONC	pzcontmain
+		CPEX	0
		?C#0	A
		GOYES	pzcontbig	Handle big coefficient
		D0=D0+	5
		A=0	W
		A=DAT0	WP		A[W]=Z
		CD0EX
		C+P+1
		CD0EX
		D0=D0+	1		Skip sign too
		P=	0
		?ST=1	sGCDpas
		GOYES	pzcontmain	Already included if 2nd pass
		?ST=0	sGCD1st
		GOYES	+		First coefficient - save as is
		C=R0	W		C[W] = gcd
pzcontgcd	GOSUBL	SHORTGCD	A[W] = gcd'
+		ST=1	sGCD1st		Got first one for sure now
		R0=A	W		R0[W] = gcd'
		A=A-1	A
		?A#0	W
		GOYES	pzcontmain	Continue unless gcd=1
		A=A+1	A
		GOLONG	PushN1Loop

pzcontbig	P=	0
		?ST=1	sGCDpas		If 2nd pass then handle it
		GOYES	+
		A=DAT0	A		Skip the Z
		CD0EX
		C=C+A	A
		D0=C
		ST=1	sGCDbig		Flag 2nd pass needed
		GOTO	pzcontmain	And postpone until 2nd pass
+		D0=D0-	5		->Z
		C=R0	W		C[W]=gcd
		GOSUBL	ZREMSHORT	A[W]=Zrem
		GOTO	pzcontgcd	Remaining code as for shorts

* Handle exit from a single pass

pzcontexit	?ST=1	sGCDpas
		GOYES	pzcontok	Did 2 passes fine - done
		?ST=0	sGCDbig
		GOYES	pzcontok	Didn't need 2 passes - done
		?ST=0	sGCD1st		Didn't get a single short! *SRPL*
		GOYES	pzcontfail
		ST=1	sGCDpas		Setup second pass
		GOSBVL	=D1=DSKTOP
		GOTO	pzcontrerun
pzcontok	A=R0	W
		?ST=1	sGCD1st		Make sure we actually got something
		GOYES	+
		A=0	W		Else the polynomial was zero!
+		GOLONG	PushN1Loop	Return the gcd

* Do it all in SRPL - all the coefficients are long

pzcontfail	GOVLNG	=GPPushFLoop
  ENDCODE

  DUPTYPEHSTR? ?SEMI		( --> Zcont )
  DROP PZContRPL
;

* This subroutine is intended for polynomials containing only
* long coefficients, but it would work for any input.

NULLNAME PZContRPL
::
  DUPTYPEHSTR? case ZAbs	( Z --> |Z|	;			)
  >R 'R PZContRPL		( Z'		; ->P[2]		)
  BEGIN				( Z'		; ->P[i] ->BEGIN	)
     RSWAP ticR NOTcase
     :: RDROP ;			( --> Z'	;			)
     RSWAP			( Z' ob		; ->P[i+1] ->BEGIN	)
     PZContRPL ZGcd		( Z''		; ->P[i+1] ->BEGIN	)
     DupZIsOne? ?SKIP
  AGAIN				( Z''		; ->P[i+1] ->BEGIN	)
  2RDROP			( Z''		;			)
;

**********************************************************************
* Desc:		Compute the primitive polynome PP of P and its content Z
*		The result is trimmed (provided P was)
* Stack:	( P --> PP Z )
**********************************************************************
NULLNAME PPZ
::
  DUP PZCont
  OVER PSetSign
  SWAPOVER QDiv
  SWAP
;

**********************************************************************
* Desc:		Compute primitive polynomial and content of non-const
*		P wrt X1. The results are trimmed (provided P was)
* Stack:	( P --> PP PC )
**********************************************************************
NULLNAME PPP
::
   DUPTYPELIST? NOTcase
   :: HXS 2 10 SWAP ;			( --> 1 Z )

   FIVE NEXTCOMPOB DROPDUP PLCZ		( P offset p1 z )
   ZIsNeg? IT QNeg			( P offset |p1| )
   BEGIN
   	UNROT				( C P offset )
   	NEXTCOMPOB
   WHILE				( C P offset pi )
   	4ROLL				( P offset pi C )
   	QGcd				( P offset C' )
   REPEAT				( C P )
   SWAPOVER PSetSign			( P PC )
   SWAPOVER				( PC P PC )
   DUPTYPELIST? IT ONE{}N		( PC P {PC} )
   QDiv SWAP				( PP PC )
;

**********************************************************************
* Desc:		Compute the derivative of P wrt X1
*		The result is trimmed (provided P was)
* Stack:	( P --> P' )
**********************************************************************
NULLNAME PDer
::
  DUPTYPELIST?			( P flag )
  NOTcasedrop HXS 1 0		( return 0 )
  DUPLENCOMP			( P n )
  #1= casedrop HXS 1 0 		( return 0 )

  INNERCOMP			( pn-1 .. p1 p0 n )
  SWAPDROPDUP			( pn-1 .. p1 n n )
  ONE_DO (DO)			( pn-1 .. p1 n )
    SWAP			( pn-1 .. p2 n p1 )
    INDEX@ #>Z			( pn-1 .. p2 n p1 i )
    QMul			( pn-1 .. p2 n p1' )
    OVER UNROLL			( p1' pn-1 .. p2 n )
  LOOP				( pn-1' .. p1' n )
  #1- 				( pn-1' .. p1' n-1 )
  CKMkP				( P' )
;

**********************************************************************
* Desc:		Set sign of P2 according to leading coeff of P1
* Stack:	( P1 P2 --> sign[P2]*P1 )
**********************************************************************
NULLNAME PSetSign
::

*---------------------------------------------------------------------
* Desc:		Get the leading integer coefficient of P
* Stack:	( P --> Z )
*---------------------------------------------------------------------
   NULLNAME PLCZ
   CODE
		C=DAT1	A
		CD0EX
		RSTK=C
		LC(5)	=DOLIST
PLCZloop	A=DAT0	A
		D0=D0+	5
		?A=C	A
		GOYES	PLCZloop
		D0=D0-	5
		C=RSTK
		CD0EX
		DAT1=C	A
		GOVLNG	=Loop
   ENDCODE

   ZIsNeg?		( P1 flag )
   NOT?SEMI QNeg	( P1' )
;

**********************************************************************
* Desc:		Put P in canonical form
* Stack:	( P --> P' )
* Algorithm:
*	Traverse seeking for "{ Z0" or "{ Z1 }"
*	If not found then done
*	Else duplicate object, flag a new pass is wanted, and strip:
*		"{ Z0 .. Z0 {"		--> "{ {"
*		"{ Z0 .. Zn .. Zn ?"	--> "{ Zn .. Zn ?"
*		"{ Z0 .. Z0 }"		--> "Z0"
*		"{ Z1 }"		--> "Z1"
**********************************************************************
NULLNAME PTrim
CODE

* Note: GC and TOTEMPOB uses ST10
*	SKIPOB uses ST1

sTRIMAGAIN	EQU 5			Flag to indicate we have restarted
sTRIMMED	EQU 6

* Provide also special entry from external subroutines

DoPTrim		GOSBVL	=SAVPTR
		ST=0	sTRIMMED
ptrimagain	ST=0	sTRIMAGAIN		
		A=DAT1	A
		D0=A
		A=DAT0	A
		LC(5)	=DOLIST
		?A=C	A
		GOYES	doptrim
ptrimnone	GOTO	ptrimdone?

doptrim		B=0	A		complevel=1
		B=B-1	A		complevel=0
		GOSUB	NeedsTrim?	CS if list needs trimming
		GONC	ptrimnone	No need - exit

* If first pass then duplicate, else just start the modification looop

		AD0EX
		?ST=1	sTRIMMED
		GOYES	ptrimA[A]
		C=DAT1	A
		A=A-C	A
		R4=A			Offset to ->Zsize
		A=B	A
		R3=A			complevel
		GOSUB	PassPTrimRPL
		CON(5)	=DOCOL
		CON(5)	=TOTEMPOB
		CON(5)	=COLA
		CON(5)	=DOCODE
		REL(5)	->PTrimEnd
		GOSBVL	=SAVPTR
		GONC	ptrimnow
PassPTrimRPL	C=RSTK
		GOVLNG	=GETPTREVALC

* Now start trimming

ptrimnow	ST=1	sTRIMMED
		C=R3
		B=C	A		complevel
		A=DAT1	A
		R0=A			->POLY for =recover call
		C=R4			Offset to ->Z0size in "{ Z0"
		A=A+C	A
ptrimA[A]	ST=1	sTRIMAGAIN
		D0=A			->input
		D1=A			->output

* Now traverse to see what to do with the found Z
ptrimdellp	A=DAT0	A
		A=A-CON	A,7
		GOC	ptrimdelZ0	Handle "{ Z0"
* Handle "{ Z1 }"
		D1=D1-	10		->dolist
		D0=D0-	5		->Z
		C=A	A
		C=C+CON	A,12		Zsize (total)
		GOSBVL	=MOVEDOWN
		D0=D0+	5		Skip SEMI
		GONC	ptrimsemi-
* Handle "{ Z0 ..."
ptrimdelZ0	D1=D1-	5		->Z0
		LC(5)	=DOHSTR
ptrimdelZ0lp	D0=D0+	6		->Z0tail
		A=DAT0	A
		D0=D0+	5
		?A#C	A
		GOYES	ptrimnonhstr
		A=DAT0	A
		A=A-CON	A,7
		GOC	ptrimdelZ0lp
		GONC	ptrimlist
* Handle "{ Z0 ... ?"
ptrimnonhstr	LC(2)	=DOLIST
		?A=C	A
		GOYES	ptrimlist
* Handle "{ Z0 .. }"
		D1=D1-	5		Back to DOLIST
		LCZ0
		DAT1=C	11
		D1=D1+	11
ptrimsemi-	B=B-1	A
		GONC	ptrimnext
		GOC	ptrimdonehere?

* Start traversing again, looking for "{ Z0" or "{ Zn }"

ptrimlist	D0=D0-	5		->Zn or ->DOLIST
ptrimnext	CD0EX
		D=C	A		Save address for later copying
		D0=C
		GOSUB	NeedsTrim?
		GONC	ptrimlast	No more matches, copy, shrink, exit
		GOSUB	ptrimcopy
		GOTO	ptrimdellp

* No more stripping, copy last segment & finish up


ptrimlast	GOSUB	ptrimcopy
ptrimdonehere?	CD1EX
		R1=C			->listend (last known)
ptrimdone?	?ST=0	sTRIMMED
		GOYES	ptrimexit
		?ST=0	sTRIMAGAIN
		GOYES	ptrimdone
		GOSBVL	=D1=DSKTOP
		GOTO	ptrimagain

ptrimdone	C=R1
		D0=C			->listend
		A=R0			->list
		C=C-A	A		list size
		C=C+CON	A,6		+link +GCnib
		DAT0=C	A
		D0=D0+	5
		GOSBVL	=recover
ptrimexit	GOVLNG	=GETPTRLOOP

		
* Copy D[A] - D0 to D1
ptrimcopy	CD0EX			C[A]=->tail
		C=C-D	A		nibbles		
		CDEX	A
		D0=C			->src
		C=D	A		nibbles
		GOVLNG	=MOVEDOWN

**********************************************************************
* Input:	D0 = ->stream
* Output:	CS: Found "{ Z0"	D0 = ->Z0size	B[A]=complevel
*		CS: Found "{ Zn }"	D0 = ->Zsize	B[A]=complevel
*		CC: No match		D0 = ->listend
* Note:		"{{ Zn }}" is a no match if Zn <> 0
**********************************************************************

* First skip until DOLIST so that trailing zeros are ok

NeedsTrim?	A=DAT0	A
		D0=D0+	5
		LC(5)	=DOLIST
		?A=C	A
		GOYES	needstrimnow?	Found DOLIST - start checking
		LC(2)	=DOHSTR
		?A=C	A
		GOYES	sneedstrimZ
		B=B-1	A
		GONC	NeedsTrim?
		RTNCC			CC: No match
sneedstrimZ	C=DAT0	A
		AD0EX
		A=A+C	A
		AD0EX
		GONC	NeedsTrim?
needstrimnow?	B=B+1	A		complevel++

needstrimlp	A=DAT0	A		Second entry point inside composite
		D0=D0+	5
		LC(5)	=DOHSTR
		?A=C	A
		GOYES	needstrimZ
needstrimback	LC(2)	=DOLIST
		?A#C	A
		GOYES	needstrimsemi
needstrimlist	B=B+1	A		complevel++
		GONC	needstrimlp
needstrimsemi	B=B-1	A
		GONC	NeedsTrim?	Seek DOLIST again
		RTNCC			CC: No need to trim
needstrimZ	A=DAT0	A		Zsize
		A=A-CON	A,7		Zmantissa nibbles
		RTNC			CS: Found "{ Z0"
		A=A+CON	A,7
		CD0EX
		R4=C.F	A		->Zsize
		C=C+A	A
		CD0EX
		A=DAT0	A
		?A#C	A
		GOYES	needstrimZ1?	Found "{ Zn }" or "{ Zn {"
		D0=D0+	5
needstrimZLP	A=DAT0	A
		CD0EX			Skip following Z's, they are ok
		C=C+A	A
		CD0EX
		A=DAT0	A
		D0=D0+	5
		?A=C	A
		GOYES	needstrimZLP
		GONC	needstrimback
needstrimZ1?	D0=D0+	5
		LC(2)	=DOLIST
		?A=C	A
		GOYES	needstrimlist
		C=R4.F	A
		D0=C			->Zsize (single element in list)
		RTNSC
->PTrimEnd
ENDCODE

**********************************************************************
* Desc:		Calculates number of variables used in P
* Stack:	( P --> P #vars )
* Notes:	#vars == max composite depth while traversing P
**********************************************************************
NULLNAME Variables
CODE
		GOSBVL	=SAVPTR
		A=DAT1	A
		D0=A			->P
		GOSUBL	QVARS
		GOVLNG	=PUSH#ALOOP
ENDCODE
**********************************************************************
