	STITLE	ALG R2Sym
**********************************************************************
*		ALG Library: Rational => Symbolic routines
**********************************************************************

**********************************************************************
* EXPORT:
*
*   RGetSign	( A B --> sflag |A| B )
*
*   ZZ2SSF	( Zn Zd --> sflag Meta )
*   R2S		( Qn Qd --> Meta )
*   QZ2S	( Q Z --> Meta )
*   PZ2S	( P Z --> Meta )
*   R2SF	( Qn Qd --> Meta )
*
*   FF2SS	( { {F1 #k1} .. { Fm #km} } Zn Zd --> sflag Meta )
*   QZFact2S	( Q Zq --> Meta )
*   RFact2S	( P Q v --> Meta )
*   
*   R2A		( P Q v --> Symb )
*   R2FA	( P Q v --> Symb )
*   P2AMOD	( P Zn v --> Symb )
*
*   FFactor	( Sym --> Sym' )
*   ZFactorSym	( Z --> Sym )
*   RFactor	( Sym --> Sym' )
*
*   ROrder	( Sym LMainVars --> Sym' )
**********************************************************************

**********************************************************************
* COMMAND SUFFIX CONVENTION:
*
*   -A   :  Symbolic 
*   -S	 :  Meta symbolic
*   -SS  :  Sign flag + Meta symbolic
*   -SF  :  Meta symbolic factor (#0 means symbolic is %1)
*   -SSF :  Sign flag + Meta symbolic factor
*
* NOTE:
*
*   The -S.. routines expect
*
*       List of variables:		Vars in 3LAM
*	Number of main variables:	Main in 2LAM
*
*   Zero for Main indicates all variables have same level
*   
**********************************************************************

DEFINE Rvars@ 3GETLAM
DEFINE Rmain@ 2GETLAM

**********************************************************************
* Name:		RGetSign
* Desc:		Extract sign from rational
* Stack:	( A B --> sflag |A| B )
**********************************************************************
NULLNAME RGetSign	
::
  SWAPDUP PLCZ ZIsNeg?		( B A sflag )
  DUP4UNROLL IT QNeg		( sflag B |A| )
  SWAP				( sflag |A| B )
;
 
**********************************************************************
* Desc:		Convert rational number into sign + Meta symb factor
* Stack:	( Zn Zd --> sflag Meta )
**********************************************************************
NULLNAME ZZ2SSF
::
  DupZIsOne? ?SKIP RCanno	( Zn' Zd' )
  SWAPDUP ZIsNeg?		( Zd' Zn' sflag )
  DUP4UNROLL IT QNeg		( sflag Zd' |Zn'| )
  2DUP EQUAL case2drop ZERO	( -> sflag #0 )
  Z>R SWAP			( sflag %n Zd' )
  DupZIsOne? casedrop ONE	( -> sflag %n #1 )
  Z>R ' x/ THREE		( sflag %n %d x/ #3 )
;

**********************************************************************
* Desc:		Convert rational Qn Qd into Meta symb
* Note: 	Lam Vars and Main must be bound
* Stack:	( Qn Qd --> Meta )
**********************************************************************
NULLNAME R2S
::
  DUPTYPELIST? NOTcase QZ2S	( -> Meta )

  SWAPDUP QIsZero? case
   :: 2DROP %0 ONE ;		( -> %0 #1 )

  DUPTYPELIST? NOTcase
   :: QZ2S onlyputopINV ;	( -> Meta&Inv )

  HXS 2 10 QZ2S			( Qd Mn )
  get1 HXS 2 10 QZ2S		( Mn Md )
  SDiv				( Meta )
;

**********************************************************************
* Desc:		Convert Q with deno Z into Meta symb
* Note: 	Lam Vars and Main must be bound
* Stack:	( Q Z --> Meta )
**********************************************************************
NULLNAME QZ2S
::
  OVER TYPELIST? case PZ2S
  ZZ2SSF 			( sflag Metaf )
  DUP#0<> case SSign		( -> Meta )
  DROP ITE %-1 %1		( %+/-1 )
  ONE				( %+/-1 #1 )
;

**********************************************************************
* Desc:		Make meta symbolic from non-const P with deno Z
* Note: 	Lam Vars and Main must be bound
* Stack:	( P Z --> Meta )
* Idea:
*	Keep the current degree of each variable in a LAM
*	Thread the polynomial using the return stack
*	Keep depth counter to indicate which degree is changing
*	  and which variables are currently involved
*	Recurse when depth reaches # main variables
**********************************************************************
NULLNAME PZ2S

DEFINE Pdstart@ 1GETLAM
DEFINE Pdstart! 1PUTLAM
DEFINE Pmain@	2GETLAM
DEFINE Pvars@	3GETLAM
DEFINE Pdeno@   4GETLAM
DEFINE Pdepth@  5GETLAM
DEFINE Pdepth!  5PUTLAM

::
  Rvars@ LENCOMP #1+		( P Z #n+1 )
  ZEROSWAP NDUPN		( P Z #0 .. #0 #n+1 )
  get1 Rvars@ Rmain@		( P #0 .. #0 #n+1 Z vars #main )
  4ROLL #3+ FALSE CACHE		( P )
    Poly2Ssub			( Meta )
  ABND
;

*-----------------------------------------------------------------
NULLNAME Poly2Ssub			( P -> Meta )
::
  ( * set depth-start to current depth 	* )
  ( * and set deg[++depth]	 	* )

  DUPLENCOMP				( P deg-start )
  Pdepth@ DUP Pdstart!			( P deg-start depth )
  #1+DUP Pdepth!			( P deg-start depth' )
  #5+ PUTLAM				( P )

  ( * start new meta * )  
  
  >R ZERO				( #0 )

  ( * main loop * )

  [BEGIN]					( Meta )

    RSWAP ticR NOTcase				( * done with current poly * )
    ::
      Pdepth@ #1-DUP Pdepth!			( Meta ; depth-- )
      RDUP Pdstart@ #> ?SEMI			( Meta ; continue if more )
      ZERO Pdstart!				( Meta ; reset depth-start )
      2RDROP					( Meta ; exit )
    ;

    RSWAP					( Meta Q )

    Pdepth@ DUP #5+ DUP GETLAM 			( Meta Q dp @deg[dp] deg[dp] )
    #1-SWAP PUTLAM				( Meta Q dp ; deg[depth--] )

    Pmain@ #<> OVER TYPELIST? ANDcase		( * Q is poly same level * )
    ::
      DUPLENCOMP				( Meta Pi degree )
      Pdepth@ #1+DUP Pdepth!			( Meta Pi degree ; depth++ )
      #5+ PUTLAM				( Meta Pi ; set degree[depth] )
      >R RSWAP [AGAIN]				( Meta ; loop for new P )
    ;

    RDUP					( Meta Q ; AGAIN on exit )
    DupQIsZero? caseDROP			( Meta ; no add for zero )
    Pdeno@ 					( Meta Q Zd )
    OVER TYPELIST? ITE 			
      :: RGetSign DROP Poly2Ssub ;		( Meta sflag Mq )
      ZZ2SSF					( Meta sflag Mq )

* Remaining section takes half of the time

   Pdepth@ #1+ Pdstart@ #1+ DO			( Meta sflag meta )
      ::
       INDEX@ #5+ GETLAM 			( Meta sflag meta deg[i] )
       DUP#0=csDROP				( continue if deg=0 )
       Pvars@ INDEX@ NTHCOMPDROP		( Meta sflag meta deg[i] Xi )
       syminner	get1 SPow# SMul			( Meta sflag meta' )
      ;
    LOOP
    DUP#0=IT :: %1 SWAP#1+ ;			( Meta sflag meta )
    get1 case SSub SAdd				( Meta' )
;

**********************************************************************
* Desc:		Convert rational Qn Qd into Meta symb factor
* Note: 	Lam Vars and Main must be bound
* Stack:	( Qn Qd --> Meta )
**********************************************************************
NULLNAME R2SF
::
  2DUP EQUALNOTcase R2S
  2DROP ZERO			( Meta is zero if Qd/Qd=1 )
;

**********************************************************************
* Desc:		Make a symbolic from rational P/Q in v
*		Q must be non-zero
* Stack:	( P Q v --> Symb )
* Note:		The CACHE-ABND takes 0.9s for 100 calls.
**********************************************************************
NULLNAME R2A
::
  3PICK TYPEHSTR? 3PICK TYPEHSTR?
  AND casedrop
  :: R2S CKSymbN ;

  ZERO				( all variables have same level )
  #TWO#TWO CACHE
    R2S CKSymbN
  ABND
;

**********************************************************************
* Desc:		Create sign + meta-sym from list of factors and terms
* Note: 	Lam Vars and Main must be bound
* Stack:	( { {F1 #k1} .. { Fm #km} } Zn Zd --> sflag Meta )
**********************************************************************
NULLNAME FF2SS
::					( Lf Zn Zd )
  ZZ2SSF				( Lf sflag Mz )
  DUP #3+ ROLL				( sflag Meta Lf )
  DUPLENCOMP				( sflag Meta Lf m )
  DUP#0=case 2DROP			( return sflag Meta )
  SWAP >R				( push Lf to stack )
  ZERO_DO (DO)				( sflag Meta )
      RSWAP 'R RSWAP			( sflag Meta Lfi )
      INCOMPDROP SWAP			( sflag Meta #ki Fi )
      HXS 2 10 QZ2S			( sflag Meta #ki MFi )
      get1 SPow#			( sflag Meta MFi^ki )
      SMul				( sflag Meta' )
  LOOP
;

**********************************************************************
* Desc:		Construct meta symbolic of factorization of Q with
*		denominator Zq (<>0)
* Note: 	Lam Vars and Main must be bound
* Stack:	( Q Zq --> Meta )
**********************************************************************
NULLNAME QZFact2S
::
  OVER TYPELIST? NOTcase QZ2S	( -> Meta )
  SWAP PFactor			( Zq Lf Z )
  ROT FF2SS			( sflag Meta )
  SSign				( Meta' )
;
   
**********************************************************************
* Desc:		Construct meta symbolic of factorization of P/Q (Q<>0)
* Note: 	Lam Vars and Main must be bound
* Stack:	( P Q v --> Meta )
**********************************************************************
NULLNAME RFact2S
::
  DUPTYPELIST? NOTcase QZFact2S		( -> Meta )

  SWAPDUP QIsZero? case
   :: 2DROP %0 ONE ;			( -> %0 #1 )

  DUPTYPELIST? NOTcase
   :: QZFact2S onlyputopINV ;		( -> Meta&Inv )

  PFactor HXS 2 10 FF2SS		( Qden sflag Mnum )
  DUP #3+ ROLL				( sflag Mnum Qden )
  HXS 2 10 QZFact2S			( sflag Mnum Mden )
  SDiv SSign				( Meta )
;    

**********************************************************************
* Desc:		Construct symbolic of factorization of rational P/Q 
*		with variables v
*		Q must be non-zero
* Stack:	( P Q v --> Symb )
**********************************************************************
NULLNAME R2FA
::
  ZERO				( all variables have same level )
  #TWO#TWO CACHE
    RFact2S CKSymbN
  ABND
;

**********************************************************************
* Desc:		Compute factorization of Sym or real
* Stack:	( Sym --> Sym' )
**********************************************************************
NULLNAME FFactor
::
  DUPTYPEREAL? NOTcase RFactor
  R>Z ZFactorSym
;

**********************************************************************
* Desc:		Compute symbolic factorization for integer
* Stack:	( Z --> sym )
**********************************************************************
NULLNAME ZFactorSym
::
  DupQIsZero? casedrop %0	( -> %0 )
  DUPDUP PSetSign		( Z |Z| )
  DUP PNFctr			( Z |Z| Lf )
  INNERCOMP reversym {}N	( * Reverse so output has increasing order * )
  UNROT				( Lf Z |Z| )
  QDiv HXS 2 10 FF2SS		( sflag Meta-factor )
  DUP#0=IT :: %1 SWAP#1+ ;      ( sflag Meta )
  SSign CKSymbN			( symb )
;

**********************************************************************
* Desc:		Compute factorization of Sym
* Stack:	( Sym --> Sym' )
**********************************************************************
NULLNAME RFactor
::
   CK&DISPATCH0
	real	NOP
	idnt	NOP
	lam	NOP
	symb	:: RSim0 R2FA ;
;

**********************************************************************
* Desc:		Make a symbolic from modular polynomial in v
* Stack:	( P Zn v --> Symb )
**********************************************************************
NULLNAME P2AMOD
::
  ROT PTrim SWAP		( Zn P v )
  HXS 2 10 SWAP			( Zn P 1 v )
  R2A syminner			( Zn Meta )
  get1 Z>R SWAP#1+		( Meta&n )
  putop: xMOD			( Meta&n&MOD )
  SYMBN				( symb )
;

**********************************************************************
* Desc:		Re-Order rational w.r.t given main variables 
* Stack:	( Sym LMainVars --> Sym' )
**********************************************************************
NULLNAME ROrder
::
  DUPLENCOMP UNROT		( #main sym LMainVars )
  RSimVars 4ROLL		( P Q Lvars #main ) 
  R2Aord			( sym' )
;

**********************************************************************
* Desc:		Convert rational to symbolic with a given number
*		of main variables.
* Stack:	( P Q Lvars #main --> Sym )
**********************************************************************
NULLNAME R2Aord
::
  #TWO#TWO CACHE		( P Q )
    R2S CKSymbN			( sym )
  ABND
;
**********************************************************************
