	STITLE	ALG Z Comparison
**********************************************************************
*		ALG library: Integer comparison routines
**********************************************************************

**********************************************************************
* Infinite precision integers are HXS binaries whose body contains
* the absolute value of the number first, then separate sign nibble.
* Integer is positive if sign is 0, negative if F
* Thus:
*	 1	= HXS 2 10
*	-1	= HXS 2 1F
*	 0	= HXS 1 0
*
* Naming conventions:
*	Z	Infinite precision integer
*	N	Positive infinite precision integer
*	P	Polynomial ( list of Q, where Q is P or Z )
*	Q	Z or P
**********************************************************************

**********************************************************************
* EXPORT:
*
*  DupQIsZero?	( Q --> Q flag )
*  QIsZero?	( Q --> flag )
*  DupZIsOne?	( Z --> Z flag )
*  ZIsOne?	( Z --> flag )
*  DupZIsTwo?	( Z --> Z flag )
*  DupZIsEven?	( Z --> Z flag )
*  DupZIsNeg?	( Z --> Z flag )
*  ZIsNeg?	( Z --> flag )
*
*  Univar?	( P --> P flag )
*  SUnivar?	( P --> P flag )
*
*  ListPos	( ob {} --> #i / #0 )
*  AppendList	( {} ob --> {}' )
*  Contains?	( ob1 ob2 --> ob1 ob2 flag )
*
*  ZBits	( Z --> Z #bits )
*  ZBit?	( Z #bit --> Z flag )
*
*  ZTrim	( Z --> Z' )
*  ZAbs		( Z --> |Z| )
*
*  PNMax	( P --> Z )
*  ZNMax	( Z1 Z2 --> NormMax[Z1,Z2] )
*  ZNMin	( Z1 Z2 --> NormMin[Z1,Z2] )
*  ZNLT?	( Z --> flag )
*
**********************************************************************

**********************************************************************
* Desc:		Test if Q is zero. Assumes list contains only lists
*		or hexes!
* Stack:	( Q --> Q flag )
**********************************************************************
NULLNAME DupQIsZero?
::
  DUP
  NULLNAME QIsZero?		( Q --> flag )
  CODE
		GOSBVL	=SAVPTR
		A=DAT1	A
		D0=A			->Q
		GOSUB	QISZERO?	CS if is zero
		GOVLNG	=GPOverWrT/FL

QISZERO?	LA(5)	=DOLIST
		LCZ0
		D=C	W		D = Z0
		B=0	A		CompLevel
* Traverse until end of object or non-zero Z is found
QIsZLp		C=DAT0	11
		?C#D	A		Not Z?
		GOYES	QIsZnotZ
		?C#D	W		Not Z0?
		GOYES	QIsZNotZero
		D0=D0+	11		Skip Z0
QIsZNext	?B#0	A		Continue if inside composite
		GOYES	QIsZLp
		RTNSC			CS: All zeros
QIsZNotZero	RTNCC			CC: Nonzero
QIsZnotZ	D0=D0+	5		Skip DOLIST or SEMI
		?C=A	A
		GOYES	QIsZlist	Handle list
		B=B-1	A		CompLevel--
		GONC	QIsZNext
QIsZlist	B=B+1	A		CompLevel++
		GONC	QIsZLp		End test known to fail, loop directly
  ENDCODE
;

**********************************************************************
* Desc:		Test if Q is Z1
* Stack:	( Z --> Z flag )
**********************************************************************
NULLNAME DupZIsOne?
CODE
		A=DAT1	A
zisone??	AD1EX
		C=DAT1	12
		AD1EX
		A=C	W
		LCZ1
Zis?exit	?A=C	W
		GOYES	+
+		GOVLNG	=PushT/FLoop
ENDCODE
*---------------------------------------------------------------------
NULLNAME ZIsOne?
CODE
		A=DAT1	A
		D1=D1+	5
		D=D+1	A
		GONC	zisone??
ENDCODE
**********************************************************************
* Desc:		Test if Q is Z2
* Stack:	( Z --> Z flag )
**********************************************************************
NULLNAME DupZIsTwo?
CODE
		A=DAT1	A
		AD1EX
		C=DAT1	12
		AD1EX
		A=C	W
		LC(N)	12
		CON(5)	=DOHSTR
		CON(5)	7
		CON(2)	2
		GOTO	Zis?exit
ENDCODE

**********************************************************************
* Desc:		Test if Z is even
* Stack:	( Z --> Z flag )
**********************************************************************
NULLNAME DupZIsEven?
CODE
		A=DAT1	A
		AD1EX
		D1=D1+	10
		C=DAT1	B		low nibbles
		D1=A
		?CBIT=0	0
		GOYES	+
+		GOVLNG	=PushT/FLoop
ENDCODE
**********************************************************************
* Desc:		Test if Z is negative
* Stack:	( Z --> Z flag )
**********************************************************************
NULLNAME DupZIsNeg?
CODE
		C=DAT1	A
zisneg??	CD1EX
		D1=D1+	5
		A=DAT1	A		Zlen
		CD1EX
		C=C+A	A		->Ztail
		CD1EX
		D1=D1-	1		->Zsign
		C=DAT1	S		Zsign
		D1=C
		?C#0	S		Test if sign <> 0 and push flag
		GOYES	+
+		GOVLNG	=PushT/FLoop
ENDCODE

NULLNAME ZIsNeg?
CODE
		C=DAT1	A
		D1=D1+	5
		D=D+1	A
		GONC	zisneg??
ENDCODE

**********************************************************************
* Desc:		Test if polynomial is univariate
* Stack:	( P --> P flag )
**********************************************************************
NULLNAME Univar?
CODE
		GOSBVL	=SAVPTR
		A=DAT1	A
		D0=A
		GOSUBL	UNIVAR?
		GOVLNG	=GPPushT/FLp
ENDCODE

**********************************************************************
* Desc:		Test if polynomial is univariate
*		and the coefficients are bounded by register size.
* Stack:	( P --> P flag )
**********************************************************************
NULLNAME SUnivar?
CODE
		GOSBVL	=SAVPTR
		A=DAT1	A
		D0=A
		GOSUBL	SUNIVAR?
		GOVLNG	=GPPushT/FLp
ENDCODE

**********************************************************************
* Name:		ListPos
* Desc:		Replacement for slow NTHOF
* Stack:	( ob {} --> #i / #0 )
**********************************************************************
NULLNAME ListPos
CODE
		C=DAT1	A		C[A] = ->list
		D1=D1+	5
		D=D+1	A
		RSTK=C
		GOSBVL	=PopASavptr	A[A] = ->ob
		C=RSTK
		D0=C			D0 = ->list
		D1=A			D1 = ->ob
		GOSUBL	CONTAINS?
		GOVLNG	=PUSH#LOOP	Push R0[A] = #pos
ENDCODE

**********************************************************************
* Name:		AppendList
* Desc:		Replacement for slow apndvarlst
* Stack:	( {} ob --> {}' )
**********************************************************************
NULLNAME AppendList
::

*---------------------------------------------------------------------
* Name:		Contains?
* Desc:		Test if ob1 contains ob2
* Stack:	( ob1 ob2 --> ob1 ob2 flag )
* Notes:	If ob1 is a symbolic then
*		   ob1 is searched for embedded ob2
*		If ob1 is a list then
*		   ob1 is traversed for a direct match
*		Else
*		   test if ob1 and ob2 are equal
*---------------------------------------------------------------------
  NULLNAME Contains?
  CODE
		GOSBVL	=SAVPTR
		C=DAT1	A
		D1=D1+	5
		A=DAT1	A
		D0=A			D0 = ->ob1
		D1=C			D1 = ->ob2
		GOSUBL	CONTAINS?
		GONC	+
		GOVLNG	=GPPushFLoop
+		GOVLNG	=GPPushTLoop
  ENDCODE

  caseDROP >TCOMP
;

**********************************************************************
* Desc:		Calculate number of bits used in Z
* Stack:	( Z --> Z #bits )
**********************************************************************
NULLNAME ZBits
CODE
		GOSBVL	=SAVPTR
		A=DAT1	A
		D0=A			->Z
		GOSUB	BitCount
		GOVLNG	=PUSH#ALOOP

* Input:	D0   ->Z
* Output:	A[A] #bits

BitCount	D0=D0+	5
		A=DAT0	A		Zlen
		CD0EX
		C=C+A	A
		D0=C
		A=A-CON	A,6		nibbles in mantissa
		?A=0	A
		RTNYES			If zero then zero bits too
		A=A+A	A
		A=A+A	A
		D0=D0-	2
		C=DAT0	B
-		C=C+C	P		Decrease unused high bits
		RTNC			Done when high bit is found
		A=A-1	A
		GONC	-
ENDCODE

**********************************************************************
* Desc:		Test if a bit in Z is set
* Stack:	( Z #bit --> Z flag )
* Note:		Count starts from zero, as opposed to ZBits
**********************************************************************
NULLNAME ZBit?
CODE
		GOSBVL	=POP#		A[A] = #bit
		GOSBVL	=SAVPTR
		C=DAT1	A		C[A] = ->Z
		GOSUB	BitTest
		GOVLNG	=GPPushT/FLp	TRUE if bit is set


* Provide subroutine for other programs
* Output:	A[A] = bit
*		C[A] = ->Z
* Output:	CS if bit is set

BitTest		D1=C			->Z
		LC(1)	3
		C=C&A	A		C[0] = lowbits
		P=C	0
		CD1EX			->Z
		ASRB.F	A
		ASRB.F	A
		C=C+A	A
		D1=C
		D1=D1+	10		->nibble to test
		A=DAT1	B
		LCHEX	1248
		P=C	3
		C=P	0
		P=	0
		A=A&C	A
		?A#0	P
		RTNYES
		RTN
ENDCODE
**********************************************************************
* Desc:		Strip Z from unnecessary leading nibbles
* Stack:	( Z --> Z' )
* Algorithm:	Count nibbles required for representation
*		If equals used nibbles then quick exit
*		Else allocate new object, copy significant mantissa
*		nibbles and apend original sign.
**********************************************************************
NULLNAME ZTrim
CODE
		GOSBVL	=SAVPTR
		A=DAT1	A		->Z
		A=A+CON	A,5
		D0=A
		C=DAT0	A		Zlen
		A=A+C	A
		D0=A			->Ztail
		D0=D0-	1		->Zsign
		A=DAT0	S		Zsign
		C=C-CON	A,7		Zlen-7
		GONC	ZTrimDo		No jump if NULLHXS or Z0
ZTrimOK		GOVLNG	=GETPTRLOOP
* First check how many nibbles should be used so we don't necessarily
* have to allocate a new object if the Z is already trimmed
ZTrimDo		D=C	A		Save for later comparison
ZTrimLoop	D0=D0-	1
		C=DAT0	S
		?C#0	S
		GOYES	ZTrimGotHi
		C=C-1	A
		GONC	ZTrimLoop
		A=0	W		Stripped down to Z0. Easy way out
		GOSBVL	=PUSHhxs
		A=DAT1	A
		D1=D1+	5
		D=D+1	A
		GOVLNG	=OverWrALoop
ZTrimGotHi	?C=D	A		Now see if Z was already trimmed
		GOYES	ZTrimOK
* Have to trim, A[S] = sign, C[A] = nibbles-1 to use
		R4=A
		RSTK=C
		C=C+CON	A,5+5+2
		GOSBVL	=GETTEMP	D0 = ->Z'
		GOSBVL	=D1=DSKTOP
		A=DAT1	A		->Z
		AD0EX
		R0=A			->Z'
		D0=D0+	10		->Zmantissa
		D1=A			->Z'
		LC(5)	=DOHSTR
		DAT1=C	A
		D1=D1+	5
		C=RSTK			Znibbles-1
		C=C+CON	A,7		Zlen = Znibbles + 1
		DAT1=C	A
		D1=D1+	5
		C=C-CON	A,6		Znibbles
		GOSBVL	=MOVEDOWN	Copy mantissa
		A=R4
		DAT1=A	S		Zsign
		GOVLNG	=GPOverWrR0Lp
ENDCODE

**********************************************************************
* Desc:		Take the absolute value of Z
* Stack:	( Z --> |Z| )
* Algorithm:	If Z is already positive then do nothing
*		Else duplicate object & change sign
**********************************************************************
NULLNAME ZAbs
CODE
dozabs		A=DAT1	A		->Z
		AD1EX
		D1=D1+	5
		C=DAT1	A		Zlen
		AD1EX
		A=A+C	A		->Ztail
		AD1EX
		D1=D1-	1		->Zsign
		A=DAT1	S		This works for Z0 & NULLHXS too
		AD1EX
		?A#0	S
		GOYES	ZAbsYes
		GOVLNG	=Loop		Already positive

ZAbsYes		GOSBVL	=SAVPTR
		A=DAT1	A		->Z
		D0=A
		D0=D0+	5
		C=DAT0	A		Zlen
		C=C+CON	A,5
		GOSBVL	=GETTEMP	D0 = ->Z'
		RSTK=C			Zsize
		GOSBVL	=D1=DSKTOP
		A=DAT1	A
		AD0EX
		DAT1=A	A		OverWrite Z'
		D1=A
		C=RSTK
		C=C-1	A		Don't copy the sign nibble
		GOSBVL	=MOVEDOWN
		A=0	A		Clear sign
		DAT1=A	1
		GOVLNG	=GETPTRLOOP
ENDCODE

**********************************************************************
* Desc:		Get the coefficient of P with max norm
* Stack:	( P --> Z )
**********************************************************************
NULLNAME PNMax

CODE

* First check that stk1 is a list, if not then leave it unchanged

		C=DAT1	A
		CD0EX
		A=DAT0	A		prolog
		D0=C
		LC(5)	=DOLIST
		?A=C	A
		GOYES	dopnmax
		GOVLNG	=Loop

dopnmax		GOSBVL	=SAVPTR
		A=DAT1	A
		D0=A			->list
		D0=D0+	5		->ob1
		B=0	A		complevel

* Now scan for the first Z

		LC(5)	=DOHSTR
pnmaxZ1lp	A=DAT0	A
		?A=C	A
		GOYES	pnmaxgotZ1
		D0=D0+	5		Must be DOLIST
		B=B+1	A
		GONC	pnmaxZ1lp

pnmaxgotZ1	AD0EX
		R0=A			->Zmax
		D0=A

* Now scan the remaining list, comparing the Z's to the max in R0

pnmaxLOOP	D0=D0+	5
		C=DAT0	A		Zmaxlen
		AD0EX
		A=A+C	A		->Zmaxtail
		D0=A
pnmaxloop	A=DAT0	A
		LC(5)	=DOHSTR
		?A=C	A
		GOYES	pnmaxcmp
		LC(4)	=SEMI
		?A=C	A
		GOYES	pnmaxsemi
pnmaxlist	B=B+1	A
pnmax05		D0=D0+	5
		GONC	pnmaxloop
pnmaxsemi	B=B-1	A
		GONC	pnmax05
		GOVLNG	=GPOverWrR0Lp	Return ->Zmax

pnmaxcmp	C=R0
		CD0EX			D0 ->Zmax
		R1=C			R1 ->Z
		D1=C			D1 ->Z
		GOSUB	ZNCMP		A[0]={1:Zmax>Z 0:Zmax=Z F:Zmax<Z
		C=R1
		D0=C			->Z
		?ABIT=0	3
		GOYES	pnmaxLOOP	Zmax>=Z, keep looking
		R0=C			->Zmax (new)
		GONC	pnmaxLOOP
ENDCODE

**********************************************************************
* Desc:		Return Z with max absolute value in Z1,Z2
*		( return Z1 if |Z1|>=|Z2| ; return Z2 if |Z1|<|Z2| )
* Stack:	( Z1 Z2 --> NormMax[Z1,Z2] )
**********************************************************************
NULLNAME ZNMax
CODE
		GOSBVL	=SAVPTR
		C=DAT1	A		
		D1=D1+	5
		A=DAT1	A		
		D0=A			D0 ->Z1
		D1=C			D1 ->Z2

		GOSUB	ZNCMP

		GOSBVL	=GETPTR
		C=DAT1	A		->Z2
		D1=D1+	5		DROP one from stack
		D=D+1	A
		A=A+1	P		is |Z1|<|Z2| ? (carry set)
		GONC	ZNMaxOk		no  -> return Z1
		DAT1=C	A		yes -> return Z2
ZNMaxOk		GOVLNG	=Loop
ENDCODE

**********************************************************************
* Desc:		Return Z with min absolute value in Z1,Z2
*		( return Z1 if |Z1|<=|Z2| ; return Z2 if |Z1|>|Z2| )
* Stack:	( Z1 Z2 --> NormMin[Z1,Z2] )
**********************************************************************
NULLNAME ZNMin
CODE
		GOSBVL	=SAVPTR
		C=DAT1	A		
		D1=D1+	5
		A=DAT1	A		
		D0=A			D0 ->Z1
		D1=C			D1 ->Z2

		GOSUB	ZNCMP

		GOSBVL	=GETPTR
		C=DAT1	A		->Z2
		D1=D1+	5		DROP one from stack
		D=D+1	A
		A=-A	P
		A=A+1	P		is |Z1|>|Z2| ? (carry set)
		GONC	ZNMinOk		no  -> return Z1
		DAT1=C	A		yes -> return Z2
ZNMinOk		GOVLNG	=Loop
ENDCODE

**********************************************************************
* Desc:		TRUE if |Z1|<|Z2|
* Stack:	( Z1 Z2 --> flag )
**********************************************************************
NULLNAME ZNLT?
CODE
		GOSBVL	=SAVPTR
		C=DAT1	A		
		D1=D1+	5
		A=DAT1	A		
		D0=A			D0 ->Z1
		D1=C			D1 ->Z2

		GOSUB	ZNCMP

		GOSBVL	=GETPTR
		D1=D1+	5		DROP Z2
		D=D+1	A
		A=A+1	P		Set carry if |Z1|<|Z2|
		GOVLNG	=OverWrT/FLp	return TRUE/FALSE

**********************************************************************
* Name:		ZNCMP
*
* Desc:		Compare Z1 to Z2 in absolute value (norm)
*
* In:		D0	->Z1
*		D1	->Z2
*
* Out:		P=0
*			/ +1	if |Z1|>|Z2|
*		A[P]   =   0 	if |Z1|=|Z2|
*			\ -1	if |Z1|<|Z2|
*		
*
* Requires:	Z1, Z2 must be trimmed 
*
* Alters:	A[W], C[W], D[A], D0, D1, P
*
* Stack Level:	0
*
**********************************************************************

ZNCMP		D0=D0+	5
		D1=D1+	5
		A=DAT0	A		Z1size
		C=DAT1	A		Z2size
		?A=C	A
		GOYES	ZNCMPeqlen	same length -> compare mantissa
		?A>C	A
		GOYES	ZNCMPgt		

* |Z1|<|Z2| - return -1

ZNCMPlt		P=	0
		LAHEX	F
		RTN

* |Z1|>|Z2| - return +1

ZNCMPgt		P=	0
		LAHEX	1
		RTN

* Z1size=Z2size - compare mantissae, starting with MS nibs

ZNCMPeqlen	A=A-1	A		A[A]=len mantissa + 5
		CD0EX
		C=C+A	A
		CD1EX			D1 -> Z1sign
		C=C+A	A
		CD0EX			D0 -> Z2sign
		D=C	A
		D=D-CON	A,7		D[A]=len mantissa - 1 
		GOC	ZNCMPeq		Z1=Z2=0 -> return equal
		P=	15
ZNCMPlp1	C=0	A		
		C+P+1			Try using 16 nibs
		?D>=C	A
		GOYES	ZNCMPlp2
		C=D	B		No more than 16 nibs lefts
		P=C	0		Use what's left
		C=C+1	B		
ZNCMPlp2				C[A]=P+1=#nibs for next block
		AD0EX
		A=A-C	A
		AD0EX			D0 ->next block of Z2
		AD1EX
		A=A-C	A
		AD1EX			D1 ->next block of Z1
		
		A=DAT1	WP		A[WP] = block of Z1
		C=DAT0	WP		C[WP] = block of Z2
		
		?A<C	WP
		GOYES	ZNCMPlt		|Z1|<|Z2|
		?A>C	WP
		GOYES	ZNCMPgt		|Z1|>|Z2|

		D=D-CON	A,16		blocks are equal 
		GONC	ZNCMPlp1	repeat if more blocks

* |Z1|=|Z2| - return 0

ZNCMPeq		P=	0
		A=0	P
		RTN		
ENDCODE
**********************************************************************

