	STITLE	ALG Groebner Bases
**********************************************************************
*	ALG library: Groebner Bases
**********************************************************************

**********************************************************************
* EXPORT:
*
*    GBASIS	( Lp --> G )
*    GSOLVE	( Lp --> Lg )
*    REDUCE	( p G --> q )
*
*    HTERM	( P --> hterm )
*    SPLIT	( P #i --> P1 P2 )
*
*    EQS2R	( {eqs} {vars} --> Lp Lvars )
*    R2EQS	( Lp Lvars --> {polys} )
**********************************************************************

**********************************************************************
* Name:		EQS2R
* Desc:		Convert equations to polynomials
* Stack:	( {eqs} {vars} --> Lp Lvars )
**********************************************************************
NULLNAME EQS2R
::

* Replace possible topmost '=' functions with '-' functions

  SWAP INNERCOMP			( {vars} eq1 .. eqn #n )
  DUP ZERO_DO (DO)
     DUP #1+ROLL
     DUPTYPEIDNT? IT ONESYMBN		( * Allow identifiers e.g. id=0 * )
     DUPTYPESYMB? NcaseTYPEERR		( * Error if not symbolic * )
     DUP DUPLENCOMP NTHCOMPDROP
     EQ: x= IT
     :: INNERCOMP swapop: x- SYMBN ;
     SWAP
  LOOP
  {}N					( {vars} {eqs}' )

* Collect used variables from the symbolics

  SWAPDUP LENCOMP UNROT			( #vars {eqs} {vars} )
  1LAMBIND				( * Bind PVarsSub variable * )
  ZERO BEGIN				( #vars {eqs} #0 )
      NEXTCOMPOB
  WHILE
      syminner PVarSub
  REPEAT
  SWAP 1GETLAM				( {eqs} #vars Lvars )

* Sort the additional variables

  INNERCOMP get1			( {eqs} var1 .. varn #n #vars )
  SWAPOVER#- DUPUNROT #2+UNROLL		( {eqs} var1 .. vark #vars .. )
  {}N RESOROMP IsV>V? SortList
  INNERCOMP top& {}N			( {eqs} Lvars' )
  1PUTLAM

* Transform symbolics to polynomials (rational parts ignored!)

  ZERO BEGIN
      NEXTCOMPOB
  WHILE
      syminner A2Rsub
      QIsZero? case ErrInfRes		( * Error for division by zero * )
      PTrim UNROT
  REPEAT
  LENCOMP {}N 1GETABND			( Lp Lvars )
;

**********************************************************************
* Name:		R2EQS
* Desc:		Convert list of polynomials to algebraics
* Stack:	( Lp Lvars #main --> {pols} )
**********************************************************************
NULLNAME R2EQS
::
   3PICK NULLCOMP? case2DROP		( --> {}  ; for {} )
   #TWO#TWO CACHE			( Lp )
      toLEN_DO (DO)			( .. Lp )
         DUPINDEX@ NTHCOMPDROP		( .. Lp Lp[i] )
         HXS 2 10 R2S CKSymbN		( .. Lp 'Lp[i]' )
         SWAP				( .. 'Lp[i]' Lp )
      LOOP
      LENCOMP {}N			( 'Lp' )
   ABND
;
**********************************************************************
* Name:		GBASIS
* Desc:		Calculate Grobner basis
* Stack:	( Lp --> G )
* Notes:	G = { 1 }	if no solutions
*		G = { 0 }	if identically true
**********************************************************************
NULLNAME GBASIS
::
   FALSE GFACTOR		( G )
;

**********************************************************************
* Name:		GSOLVE
* Desc:		Calculate factorized Grobner basis
* Stack:	( Lp --> Lg )
* Notes:	Lg = { Lg1 Lg2 .. Lgn }
*		Lgi = independant solution (probably)
*		Lg = {}		if no solutions
*		Lg = { { 0 } }	if identically true
**********************************************************************
NULLNAME GSOLVE
::
   TRUE GFACTOR			( Lg )
   REDUCESOL			( Lg' )
;

*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
* Name:		GFACTOR
* Desc:		Calculate Grobner basis or factorized Grobner basis.
*		Redundant bases are not removed.
* Stack:	( Lp fctr? --> Lg )
* Local vars:	fctr?	TRUE if factorized basis wanted
*		G	Current basis
*		k	Size of current basis
*		tmp	Help variable
*		todo	Number of subproblems to do (factorized basis)
*		Lg	Current factorized basis
*		Irred	List of irreducible polynomials
*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
NULLNAME GFACTOR
::

* Bind variables

  ZEROZEROZERO NULL{} NULL{}
  {{ G fctr? k tmp todo Lg Irred }}

* Trim set of polynomials to remove 0=0 equations

  G PTrim
  DUPTYPEHSTR? ITE ZERO DUPLENCOMP
  k! G!

* Handle trivial solutions

  G QIsZero? case				( * Identically true * )
  :: HXS 1 0 ONE{}N fctr? ABND case ONE{}N ;

  G TYPEHSTR? casedrop				( * No solutions * )
  :: HXS 2 10 ONE{}N fctr? ABND case ONE{}N ;

* Handle the case of a single basis polynomial

  k #1=case
  :: G PPZ DROP					( G' )
     fctr? ABND					( G' fctr? )
     NOT?SEMI					( --> G' )
     CARCOMP PFactor DROP			( Lf )
     toLEN_DO (DO)				( .. Lf )
        DUPINDEX@ NTHCOMPDROP CARCOMP		( .. Lf Lfi )
	ONE{}N SWAP				( .. Lgi Lf )
     LOOP
     LENCOMP {}N				( --> Lg )
  ;

* Handle the general case. Create initial list of pairs

  ZERO						( * pairs=null * )
  k ONE_DO (DO)					( * For i=1 to k-1 * )
    k #1+ INDEX@ #1+ DO				( * For j=i+1 to k * )
      JINDEX@ INDEX@ NEWPAIR			( * Add pair [i,j] * )
    LOOP
  LOOP

* Handle plain Grobner basis calculation

  fctr? NOTcase
  :: GFACTOR2 G					( G )
     "GBASIS " k #>$ &$ Verbose1
     NULL$ Verbose2
     ABND
     DUPNULLCOMP? casedrop { HXS 1 0 }		( * Fix trivial basis * )
  ;

* Grobner factorizer loop

  BEGIN
	GFACTOR2				( * Current main problem * )
	Lg G					( * Save solution * )
	DUP CARCOMP TYPEHSTR?
	ITE_DROP AppendList			( * Unless trivial one * )
	Lg!
	todo #0<>				( * Next subproblem? * )
  WHILE
	todo #1- todo!
	"TODO: " todo #>$ &$ Verbose3
	DUPLENCOMP k! G!			( * Set next problem * )
  REPEAT

* Return found solutions

  Lg ABND					
  "SOLUTIONS " OVER LENCOMP #>$ &$ Verbose1
  NULL$ Verbose2
;

*---------------------------------------------------------------------
* Name:		GFACTOR2
* Desc:		Grobner factorizer - returns when one subproblem
*		has been finished.
* Notes:	Lambda variables {{ fctr? G k tmp todo Lg Irred }}
*---------------------------------------------------------------------
NULLNAME GFACTOR2
::

* Prereduction and attempt first split

  GREDUCE GSPLIT

* Buchberger algorithm with full interreduction and extra split checks

  BEGIN					( pairs )
	G NEXTPAIR			( pairs' #i #j TRUE / #0 FALSE )
  WHILE
	"GBASIS " k #>$ &$
	" PAIRS " &$ 4PICK #>$ &$
	Verbose1

	G ROT NTHCOMPDROP		( pairs #j Gi )
	G ROT NTHCOMPDROP		( pairs Gi Gj )
	"S-POLY" Verbose2 SPOLY		( pairs spoly )
	"REDUCE" Verbose2 G REDUCE	( pairs h )

	DupQIsZero? caseDROP		( * Zero - try next pair * )
	DUPTYPEHSTR? casedrop		( * Constant - no solutions! * )
	:: NDROP ZERO			( #0 )
	   HXS 2 10 ONE{}N G! ONE k!	( * G={1} * )
	;
	"ADDING" Verbose2		( * Add new polynomial to basis * )
	PPZ DROP			( * Remove content * )
	G SWAP >TCOMP G!		( * G=G&h * )
	k #1+ k!			( * k=k+1 * )
	k ONE_DO (DO)			( * For i=1 to k-1, [i,k], Next * )
	   INDEX@ k NEWPAIR
	LOOP
	GREDUCE				( * Interreduce * )
	GSPLIT				( * Try to split * )
  REPEAT
  DROP					( #0 --> )
;

*---------------------------------------------------------------------
* Name:		GSPLIT
* Desc:		Try to split current basis
* Notes:	Lambda variables {{ fctr? G k tmp todo Lg Irred }}
*---------------------------------------------------------------------
NULLNAME GSPLIT
::

* Abort if factorized basis not wanted

  fctr? NOT?SEMI

* Try to factor any polynomial in the basis. We search in reverse
* order since last polynomials should have less variables and thus
* should be simpler to factor. A better strategy for us might be
* to test in order of size since PFactor is not that fast.
* The loop has protection against constants just in case.

  ZERO						( #0 )
  k #1+_ONE_DO (DO)				( * Test all Gi * )
  :: "FACTORING " ISTOP-INDEX #>$ &$ Verbose2
     Irred					( #0 Irred )
     G ISTOP-INDEX NTHCOMPDROP			( #0 Irred Gi )
     DUPTYPEHSTR? case2DROP			( --> #0 ; Gi is constant )
     Contains? case2DROP			( --> #0 ; Gi is irreducible )
     SWAPDROPDUP				( #0 Gi Gi )
     PFactor DROP				( #0 Gi Lfi )
     DUPLENCOMP #1<>				( #0 Gi Lfi flag1 )
     OVER CARCOMP TWO NTHCOMPDROP #1<>		( #0 Gi Lfi flag1 flag2 )
     ORcase					( * Factorized! * )
     :: INDEX@ 2SWAP 2DROP ExitAtLOOP ;		( --> Gi #i )
     Irred SWAPDROPSWAP AppendList Irred!	( * Irred=Irred&Gi * )
  ;
  LOOP						( Gi #i / #0 )

  "TODO: " todo #>$ &$ Verbose3
  DUP#0=csDROP					( -->  ; No split )

* Found a factorizable polynomial - split problem to subsets

  k! tmp!					( * k=i ; tmp=Gi * )
  tmp LENCOMP					( #factors )
  todo OVER#+ #1- todo!				( * todo=todo+factors-1 * )
  #1+_ONE_DO (DO)
     metaDUP {}N INNERCOMP			( pairs pairs )
     G INNERCOMP				( pairs pairs G1..Gi..Gn #n )
     k #1+ROLL DROP				( pairs pairs G1..Gn #n )
     tmp INDEX@ NTHCOMPDROP CARCOMP		( pairs pairs G1..Gn #n Lfij )
     DUP PSetSign
     Irred OVER AppendList Irred!		( * Irred=Irred&Lfij * )
     k #1+UNROLL				( pairs pairs G1..Lfij..Gn #n )
     {}N OVER#2+UNROLL				( pairs G' pairs )
  LOOP

* And continue Grobner factorizer with the last subproblem.
* Since the basis changed we interreduce and try to split again.

  NDROP						( .. pairs G' )
  DUPLENCOMP k! G!				( * Re-establish k and G * )
  "TODO: " todo #>$ &$ Verbose3
  GREDUCE COLA GSPLIT				( * Interreduce and split * )
;

*---------------------------------------------------------------------
* Name:		GREDUCE
* Desc:		Interreduce basis
* Notes:	Lambda variables {{ fctr? G k tmp todo Lg Irred }}
*---------------------------------------------------------------------
NULLNAME GREDUCE
::

* No interreduction if basis does not have atleast two polynomials
* In such cases pairs is already zero so with the current algorithm
* this test is unnecessary. We keep the test anyway just in case
* we try moving GREDUCE calls, although I have already tested many
* places to perform worse..

  k TWO #<case NOP

* Sort the basis

  G INNERCOMP GSORT {}N G!

* Find reduction, do it, sort basis and loop until no reductions found

  BEGIN
  ::
    "GBASIS " k #>$ &$ " PAIRS " &$ OVER #>$ &$ Verbose1

    G REDFIND				( G #i )
    tmp! DROP				( )

    tmp #0=?SEMI			( * No reduction found -- done * )

    "REDUCE " tmp #>$ &$ Verbose2
    G tmp NTHCOMPDROP			( Gi )
    G tmp #1+ MINUSONE SUBCOMP		( Gi G' ; G' with all Gj, j>i )
    REDUCE				( Gi' )

    DupQIsZero? casedrop		( * Remove Gi from basis * )
    :: G INNERDUP			( G1 .. Gi .. Gn #n #n )
       tmp #- #2+ROLL			( G1 .. Gn #n Gi )
       DROP#1- {}N			( G' )
       G! k #1- k!
       tmp REMPAIRS			( * Update pairs * )
    ;

    DUPTYPEHSTR? casedrop		( * Constant - unsolvable * )
    :: NDROP ZEROZERO tmp!		( * Drop pairs and set tmp=0 * )
       HXS 2 10 ONE{}N G! ONE k!	( * G={1} * )
    ;

    PPZ DROP				( * Update Gi in basis * )
    G INNERCOMP get1			( G1 .. Gi .. Gn #n Gi' )
    OVER tmp #-DUP #4+ ROLLDROP		( G1 .. Gn #n Gi' # )
    #2+UNROLL				( G1 .. Gi' .. Gn #n )
    GSORT				( * Re-sort basis * )
    {}N G!				( )
  ;

  tmp #0= UNTIL				( * Until no reduction found * )
;

*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
* Name:		REDUCESOL
* Desc:		Simplify a set of solutions by removing solutions
*		which can be seen redundant by calculating the
*		Grobner basis for the union of two solutions.
* Stack:	( Lg --> Lg' )
*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
NULLNAME REDUCESOL
::

* Quit if no solutions (NULL{}) or identically true ( {{0}} )
* We manage with a single test for both.

  DupQIsZero? ?SEMI			
  
* Sort solutions so that smallest sets come first. Smallest sets
* should be most parametric and thus should be most likely to
* cause reductions.

  ' :: LENCOMP SWAP LENCOMP #< ;
  SortList

* Now test each possible pair of solutions, if the basis for the
* union of two solutions equals either one of the individual solutions,
* that solution is redundant. This will not detect all redundant
* solutions though!

  {{ Lg }}

  Lg LENCOMP #1+_ONE_DO (DO)
  ::
     Lg INDEX@ NTHELCOMP		( Lgi TRUE / FALSE )
     NOT?SEMI				( * No more solutions * )
     ISTOP-INDEX #1=case DROP		( * No pair for Lgi * )
     ISTOP@ INDEX@ #1+ DO		( With all Lgj, j>i )
     ::
	"SOLUTIONS: " Lg LENCOMP #>$ &$ Verbose3
	Lg INDEX@ NTHELCOMP		( Lgi Lgj TRUE / Lgi FALSE )
	NOT?SEMI			( * No pair for Lgi * )
	2DUP &COMP GBASIS		( Lgi Lgj Lg )
	3PICKOVER EQUALcase		( * If Lgi==Lg * )
	::
	   2DROP			( Lgi )
	   Lg INNERCOMP			( Lgi Lg1 .. Lgi .. Lgn #n )
	   DUP JINDEX@ #- #2+ROLL	( Lgi Lg1 .. Lgn #n Lgi )
	   DROP#1- {}N			( Lgi Lg' )
	   Lg!				( Lgi )
	   JINDEX@ #1- JINDEXSTO	( * Test the new Lgi * )
	   ExitAtLOOP			( * Restart Lgj loop * )
	;
	EQUAL				( * If Lgj==Lg * )
	NOT?SEMI			( * Else try next Lgj * )
	Lg INNERCOMP			( Lg1 .. Lgj .. Lgn #n )
	DUPINDEX@ #- #2+ROLL		( Lg1 .. Lgn #n Lgj )
	DROP#1- {}N			( Lg' )
	Lg!				( )
	INDEX@ #1- INDEXSTO		( * Test the new Lgj * )
     ;
     LOOP				( Lgi )
     DROP				( )
  ;  
  LOOP					( )
  Lg ABND				( Lg' )
;

*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
* Name:		SPOLY
* Desc:		Calculate s-polynomial for two *polynomials*
* Stack:	( p q --> spoly )
* Notes:	spolynomial calculation involves division by
*		the gcd of the leading terms. Since we intend
*		to remove the content later on anyway we divide
*		by the gcd of the powers only, the integer gcd
*		will be automatically removed later on.
*
*		Mp = hterm(p)	lcm = lcm(Mp,Mq)
*		Mq = hterm(q)	gcd = gcd(Mp,Mq)
*
*		spoly(p,q) = lcm*(p/Mp-q/Mq)
*		           = Mq/gcd*p-Mp/gcd*q
*
*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
NULLNAME SPOLY
::
  OVER HTERM			( p q Mp )
  OVER HTERM			( p q Mp Mq )

  PFactPowCnt DROPROT		( p q Mq Pq Mp )
  PFactPowCnt DROPROT		( p q Mq Mp Pp Pq )

* Note that PDivLk assumes a power exists for each variable, thus
* we must pad with zeros at the tail according to which polynomial
* has more variables. For the gcd calculation we must check only
* the first powers though according to which polynomial has less
* variables.
* Note also that this method is usually much faster than calculating
* QGcd and dividing by that. Even the following method of calculating
* the gcd itself is usually clearly faster than QGcd!

  DUPLENCOMP 3PICK LENCOMP	( p q Mq Mp Pp Pq #q #p )
  2DUP #MIN UNROT #MAX		( p q Mq Mp Pp Pq #min #max )
  OVER#- 2SWAP			( p q Mq Mp #min #max-min Pp Pq )

  4PICK #1+_ONE_DO (DO)		( p q Mq Mp .. #min #max-min Pp Pq )
    OVER INDEX@ NTHCOMPDROP	( p q Mq Mp .. #min #max-min Pp Pq #p^ )
    OVER INDEX@ NTHCOMPDROP	( p q Mq Mp .. #min #max-min Pp Pq #p^ #q^ )
    #MIN 5UNROLL		( p q Mq Mp .. #gcd #min #max-min Pp Pq )
  LOOP
  2DROP				( p q Mq Mp .. #gcd #min #max-min )
  ZEROSWAP NDUPN top& {}N	( p q Mq Mp Lgcd )
  
  DUPUNROT PDivLk		( p q Mq Gp Mp' )
  UNROT PDivLk			( p q Mp' Mq' )
  4ROLL QMul UNROT QMul QSub	( p*Mq'-q*Mp' )
  PTrim				( spoly*const )
  DupQIsZero? ?SEMI		( --> 0 )
  PPZ DROP			( --> spoly )
;

*---------------------------------------------------------------------
* Name:		GSORT
* Desc:		Sort basis to decending order
* Stack:	( pairs Meta --> pairs' Meta' )
* Algorithm:	For i=1 to k-1
*		  For j=i+1 to k
*		    If Gi<Gj then swap Gi,Gj on stack
*		             swap indexes i,j in all pairs
*		  Next
*		Next
*		We actually do the test in reverse since we index
*		from bottom of stack instead of from start of G.
*---------------------------------------------------------------------
NULLNAME GSORT
CODE

		GOSBVL	=SAVPTR
		GOSBVL	=POP#
		R4=A	A		R4[A] = k

		A=0	A
		A=A+1	A		i=1
GSORTLOOP	R3=A	A		R3[A] = i
		A=A+1	A		j=i+1
gsortloop	R2=A	A		R2[A] = j
		GOSUB	CmpPair		CS if Gi<Gj and thus no change
		GOC	gresort		Change order
gsortnext	A=R2	A
		A=A+1	A		A[A] = j+1
		C=R4	A		C[A] = k
		?A<=C	A
		GOYES	gsortloop	Until j>k
		A=R3	A
		A=A+1	A		A[A] = i+1
		?A<C	A
		GOYES	GSORTLOOP	Until i>=k
		GOVLNG	=GETPTRLOOP

* Swap Gi and Gj on stack and swap indexes in pairs

gresort		GOSUB	GetStkPair	Get pair #i #j
		A=DAT0	A		->Gi
		C=DAT1	A		->Gj
		DAT0=C	A
		DAT1=A	A		Swapped

		GOSBVL	=D1=DSKTOP	C[A] = ->stack
		A=R4	A		A[A] = k
		A=A+1	A
		C=C+A	A
		A=A+A	A
		A=A+A	A
		C=C+A	A
		D1=C			D1 = ->pairs
		GOSBVL	=POP#
		B=A	A		B[A] = #pairs
--		?B=0	A
		GOYES	gsortnext	Reindexing done - coninue sort
		C=DAT1	A		C[A] = ->$pair
		D1=D1+	5
		D0=C			D0 = ->$pair
		D0=D0+	10		D0 = ->pair_i
		P=	16-2		Reindex two indexes
-		A=DAT0	A		A[A] = index
		C=R2	A		C[A] = j
		?A=C	A
		GOYES	gstst1		If index=j set index=i
		C=R3	A		C[A] = i
		?A#C	A
		GOYES	gstst3		If neither then just continue
		C=R2	A		C[A] = j
		GONC	gstst2		Had index=i set index=j
gstst1		C=R3	A		C[A] = i
gstst2		DAT0=C	A		Change index
gstst3		D0=D0+	5
		P=P+1
		GONC	-		Check both indexes in pair
		B=B-1	A
		GONC	--		Check all pairs

* Comparison subroutine for Gi and Gj
* We test the degrees of the leading coefficient only.

CmpPair		GOSUB	GetStkPair
		A=DAT0	A
		C=DAT1	A
		D0=A			D0 = ->Gi
		D1=C			D1 = ->Gj
--		A=DAT0	A		A[A] = @Gi
		C=DAT1	A		C[A] = @Gj
		?A#C	A
		GOYES	gcmptst		Different prologs - compare them
		LC(2)	=DOLIST
		?A#C	A
		GOYES	gcmpcc		CC: "EQUAL" to avoid extra work
		CD0EX
		RSTK=C			RSTK = ->Gi
		D0=C
		GOSUBL	QLEN
		R0=A	A		R0[A] = len(i)
		CD1EX
		D1=C
		D0=C			D0 = ->Gj
		GOSUBL	QLEN		B[A] = len(j)
		C=RSTK
		D0=C			D0 = ->Gi
		D0=D0+	5		->Gi++
		D1=D1+	5		->Gj++
		A=R0	A		A[A] = len(i)
		?A=B	A
		GOYES	--		Test next variable if equal
		C=B	A		C[A] = len(j)
gcmptst		?A<=C	A
		RTNYES
gcmpcc		RTNCC

* Fetch stack positions for Gi and Gj given indexes i,j in R3[A],R2[A]

GetStkPair	GOSBVL	=D1=DSKTOP	D1 = ->stack
		C=R4	A		C[A] = k
		A=R3	A		A[A] = i
		C=C-A	A
		C=C+1	A		C[A] = k-i+1
		AD1EX			A[A] = ->stack
		D1=A			D1 = ->stack
		A=A+C	A
		C=C+C	A
		C=C+C	A
		A=A+C	A
		D0=A			D0 = ->stack Gi
		C=R4	A		C[A] = k
		A=R2	A		A[A] = j
		C=C-A	A
		C=C+1	A		C[A] = k-j+1
		AD1EX
		A=A+C	A
		C=C+C	A
		C=C+C	A
		A=A+C	A
		D1=A			D1 = ->stack Gj
		RTN
ENDCODE

*---------------------------------------------------------------------
* Name:		REDFIND
* Desc:		Given a sorted basis find a reducible polynomial.
* Stack:	( G --> G #Gi )
* Notes:	#Gi = 0 if no polynomial reduces
* Algorithm:
*	Test all polynomials in basis in reverse order for possible
*	reductions. The test is done in two phases, the first one tests
*	only the leading coefficients for reductions. This should give
*	us final leading coefficients as fast as possible.
*	Should consider testing the leading coefficients in reverse order
*	to see if it can given a faster path to the final basis.
*---------------------------------------------------------------------
NULLNAME REDFIND
CODE

* Init flags

		ST=1	sDUMMYCOEFF	8	Ignore coefficients	
*		ST=0	sMODULAR	9	Use modular Distribute
		ST=0	sGARB		10	No GC yet
*		ST=1	sTRASH		11	No overflow will happen!

redfindretry	GOSBVL	=SAVPTR

* Check basis has atleast two polynomials

		A=DAT1	A
		D0=A			D0 = ->G
		GOSUBL	QLEN
		R2=A	A		R2[A] = k
		A=A-CON	A,2
		GONC	+		Continue if k>=2
		A=0	A
		GOVLNG	=PUSH#ALOOP	Else ( --> G #0 )

* Allocate all free memory to be work memory

+		GOSBVL	=ROOM
		A=C	A
		LC(5)	2000		We want atleast 1K
		?A>=C	A
		GOYES	+
redfindgc	GOSBVL	=DOGARBAGE
		GONC	redfindretry
+		CSRB.F	A
		C=A-C	A
		R4=C	A		R4[A] = free
		GOSBVL	=MAKE$N
		AD0EX
		R0=A	A		R0[A] = ->poly1
	
* Count variables used in basis

		A=DAT1	A
		D0=A			D0 = ->G
		GOSUBL	QVARS
		A=A-1	A		Substract embedding list
		R3=A	A		R3[A] = vars

* Convert base polynomials to distributed form

		C=DAT1	A
		D0=C			D0 = ->G
		D0=D0+	5		D0 = ->G1
		C=R0	A
		D1=C	A		D1 = ->output
redfindpop	C=R3	A
		B=C	A		B[A] = vars
		GOSUBL	Distribute1
		GOC	redfindgc
		A=DAT0	A
		LC(5)	=SEMI
		?A#C	A
		GOYES	redfindpop	Repeat until end of G

* Output stopped

		A=R4	A
		A=A-CON	A,16		free = free - 16
		GOC	redfindgc
		R4=A	A
		A=0	W
		DAT1=A	W

* Now pass through the polynomials in reverse order, the terms
* in currently selected polynomial left to right in order to
* find a reduction. During the first pass we actually test only
* the leading terms for reductions.

* R0[A] = ->poly1
* R1[A] = k
* R2[A] = i (current index)
* R3[A] = vars
* R4[A] = free memory (not needed anymore)

sPASS1	EQU 0		Flag to indicate first pass

		ST=1	sPASS1		Set first pass
		A=R2	A
		R1=A	A		Set i=k

redfindloop
		A=R2	A
-		A=A-1	A		i=i-1
		R2=A	A
		A=A-1	A
		GONC	++		Continue if i>0

		?ST=1	sPASS1		If pass 1 then set pass 2
		GOYES	+
		GOTO	redfinddone	Else exit with no reductions found
+		ST=0	sPASS1		Set pass 2
		A=R1	A		Set i=k again
		GOC	-		And restart loop

* Find the polynomial in the basis with the given index i

++		B=A	A		B[A] = i-1
		A=R0	A
		D0=A			D0 = ->poly1
-		B=B-1	A
		GOC	+
		GOSUBL	DISTSKIP
		GONC	-		Skip until ->polyi reached
+		CD0EX
		B=C	A		B[A] = ->polyi

* Test if hterm of some Gj with j>i divides a term in Gi. Restrict
* test to hterm of Gi in pass 1.
	
redfind1loop
		C=B	A
		D0=C			D0 = ->termi
		GOSUBL	DISTSKIP	D0 = ->Gj
--		C=B	A
		D1=C			D1 = ->termi
		D0=D0+	16		Skip coefficients
		D1=D1+	16
		C=R3	A
		D=C	A		D[A] = vars
-		A=DAT1	A		A[A] = termi^
		C=DAT0	A		C[A] = htermj^
		D1=D1+	5
		D0=D0+	5
		A=A-C	A
		GOC	redfind#	not divisible - try next Gj
		D=D-1	A
		?D#0	A
		GOYES	-		Test all powers in term
		GOTO	redfinddone	Was divisible - return #i

* hterm(Gj) did not divide termi - try next Gj
redfind#	D=D-1	A		Skip to end of htermj powers
		?D=0	A
		GOYES	+
		D0=D0+	5
		GONC	redfind#
+		GOSUBL	DISTSKIP	Then skip to end of Gj
		A=DAT0	W
		?A#0	W
		GOYES	--		And loop if there is a Gj+1

* Current term in Gi is not divisible by any hterm(Gj).
* If pass 2 skip termi and retry with the new term - else set new Gi

		?ST=1	sPASS1		If pass 1
		GOYES	+		Then set new Gi

		GOSUBL	CSLOT		Else skip termi
		B=B+C	A		B[A] = ->term++
		A=B	A
		D0=A
		A=DAT0	W
		?A#0	W
		GOYES	redfind1loop	And test new termi unless end of Gi

* Gi is not reducible (or hterm if pass1) - get next Gi

+		GOTO	redfindloop	Index handled at the start..

* Return index in R2[A] - possible zero

redfinddone	GOSUBL	TrashTop	Trash the created $ first
		A=R2	A
		GOVLNG	=PUSH#ALOOP	( G --> G #Gi )
ENDCODE	

*---------------------------------------------------------------------
* Name:		NEWPAIR
* Desc:		Add new pair to pairs list.
* Stack:	( pairs #i #j --> pairs' )
* Notes:
*	A pair is of the format
*
*		CON(5)	=DOHSTR		type
*		CON(5)	10+5		size
*		CON(5)	i		index 1
*		CON(5)	j		index 2
*
*	The choice allows easy changes in the indexes via assembly
*	language. The type itself was chosen to be HSTR instead of CSTR
*	so that the indexes would be easy to see while debugging.
*---------------------------------------------------------------------
NULLNAME NEWPAIR
CODE
		GOSBVL	=POP2#
		R1=A	A		R1[A] = I
		R2=C	A		R2[A] = J
		GOSBVL	=SAVPTR
		GOSBVL	=POP#
		R3=A	A		R3[A] = pairs
		LC(5)	10
		GOSBVL	=MAKE$N		Allocate $pair
		A=R1	A
		DAT0=A	A		Write I
		D0=D0+	5
		A=R2	A
		DAT0=A	A		Write J
		D0=D0-	15
		LC(5)	=DOHSTR
		DAT0=C	A		Change type to HSTR
		A=R0	A
		DAT1=A	A		Overwrite #pairs with ->$pair
		A=R3	A
		A=A+1	A		
		GOVLNG	=PUSH#ALOOP	And push new increased #pairs
ENDCODE

*---------------------------------------------------------------------
* Name:		REMPAIRS
* Desc:		Update pairs when a polynomial in the basis is removed.
*		  o Delete any pair with the given index
*		  o Decrease any index above the given index
* Stack:	( pairs #i --> pairs' )
*---------------------------------------------------------------------
NULLNAME REMPAIRS
CODE
		GOSBVL	=POP#
		R0=A	A		R0[A] = index
		GOSBVL	=SAVPTR
		GOSBVL	=POP#
		R1=A	A		R1[A] = pairs (counter)
		R2=A	A		R2[A] = pairs (original)
		CD1EX
		C=C+A	A
		A=A+A	A
		A=A+A	A
		C=C+A	A
		D1=C			D1 = ->lastpair (target)
		D0=C			D0 = ->lastpair (source)
		A=R0	A
		B=A	A		B[A] = index
		D=0	A		D[A] = removed pairs
--		A=R1	A
		A=A-1	A
		R1=A	A		R1[A] = pairs--
		GOC	rempairok	Done when all pairs checked
		D0=D0-	5		D0 = ->pair (source)
		A=DAT0	A
		AD0EX			A[A] = ->pair (source)
		D0=D0+	5
		P=	16-2		Check 2 indexes
-		D0=D0+	5
		C=DAT0	A		C[A] = index from pair
		?C=B	A
		GOYES	++		If equal then remove pair
		?C<B	A
		GOYES	+		Else if greated then
		C=C-1	A		decrement pair index
		DAT0=C	A		
+		P=P+1
		GONC	-		Loop for two indexes in the pair
		D0=A			D0 = ->pair (source)
		A=DAT0	A
		D1=D1-	5
		DAT1=A	A		Output the pair which is kept
		GONC	--
++		P=	0		Reset index left from above loop
		D0=A			Remove pair by not outputting it
		D=D+1	A		D[A] = removed++
		GONC	--		Loop until all pairs checked

* All pairs now checked - update #pairs etc variables

rempairok	C=D	A
		A=C	A		A[A] = removed pairs
		GOSBVL	=GETPTR
		C=R2	A		C[A] = pairs (original)
		C=C-A	A		C[A] = pairs left
		R0=C	A
		C=A	A		Skip removed levels of stack
		C=C+1	A
		D=D+C	A
		AD1EX
		A=A+C	A
		C=C+C	A
		C=C+C	A
		A=A+C	A
		D1=A
		GOSBVL	=SAVPTR		And save the new stack pointer
		GOVLNG	=PUSH#LOOP	And push the new #pairs count
ENDCODE

*---------------------------------------------------------------------
* Name:		NEXTPAIR
* Desc:		Choose next pair from pairs list.
* Stack:	( pairs G --> pairs' #i #j TRUE )
*		( pairs G --> #0 FALSE )
* Notes:
*	o Any pair whose leading terms are coprime is removed
*	  from pairs since the respective spolynomial is know
*	  to reduce to zero. This is easily seen directly from
*	  the formula for spolynomials.
*	o Any pair i,j whose LCM of the leading terms is divisible
*	  by the hterm of some Gk not included in the pair and
*	  if the respective pairs i,k j,k have been tested can
*	  be removed since it is known to reduce to zero.
*	o From the remaining pairs we choose the best pair to be
*	  the one whose LCM of the lexicographic leading terms
*	  has minimum *total* degree. Experimentally this is a
*	  better choice than comparing the LCM lexicographically.
*	  Note that the general idea is to get as simple spolynomials
*	  as possible, or to be more precise spolynomials which
*	  lead to final Grobner basis as fast as possible.	
*	o Many selection strategies have been tested, each more or
*	  less logical. Ignoring the lucky speedups none appeared
*	  to improve the "totally" normal selection strategy above.
*	  What remains is the sugar strategy, but my guess is that
*	  it is not worth the extra trouble in coding.
*
*	Since the divisibility critetion is not succesful too often
*	we delay testing it until a minimum has been selected.
*	If the criterion then removes the pair we simply restart
*	the search from scratch. Note that this way we have also
*	removed all gcd=1 pairs too so that the criterion has a
*	higher chance for success.
*---------------------------------------------------------------------
NULLNAME NEXTPAIR
CODE

* Quick exit if there are no pairs left

		D1=D1+	5
		A=DAT1	A
		LC(5)	=ZERO
		?A#C	A
		GOYES	+
		D=D+1	A
		GOVLNG	=PushFLoop
+		D1=D1-	5

* Allocate space and calculate powers for the leading terms of
* the polynomials in the basis.

		GOSUB	PopHTerms

* Calculate LCM for each pair.
* If GCD=1 then remove pair
* Else test if new minimum total degree


		A=R4	A
		GOSBVL	=ASLW5		A[A1] = k
		GOSBVL	=D1=DSKTOP
		D1=D1+	5		D1 = ->pairs
		GOSBVL	=POP#
		R4=A	A		R4[A] = pairs	R4[A1]=k

* Restart point if minimum is discarded

FindNextPair	A=0	A
		A=A+1	A
		R1=A	A		R1[A] = pair

sGOTMIN	EQU 9		Flag to indicate minimum exists
sGCD1	EQU 10		Flag to indicate GCD = 1

		ST=0	sGOTMIN		No minimum yet

* During the loop R2[A]=mindegree R2[A1]=minpair

NextPairLoop	A=R1	A		A[A] = pair
		C=R4	A		C[A] = pairs
		?A>C	A
		GOYES	NextPairTest	No more pairs - test minimum
		GOSUB	PairLCM		LCM for pair	B[A]=totdegree

		?ST=0	sGCD1		If gcd<>1 then
		GOYES	pairgcd#1	test if new minimum

		A=R1	A		Else remove pair from stack
		GOSUB	PairRemove
		GOTO	NextPairLoop	And continue with the same index!

pairgcd#1	?ST=0	sGOTMIN		If no minimum yet
		GOYES	pairnewmin	Then this is one
		A=R2	A		A[A] = mindegree
		?B>A	A
		GOYES	nextpair
pairnewmin	A=R1	A		A[A] = pair
		GOSBVL	=ASLW5		A[A1] = pair
		A=B	A		A[A] = totdegree
		R2=A	W		R2[W] = minimum pair info
		ST=1	sGOTMIN		Got a minimum now
nextpair	A=R1	A
		A=A+1	A		pair++
		R1=A	A
		GONC	NextPairLoop	Continue search

* All pairs tested. Do a criterion 2 test on the found minimum.

NextPairTest	?ST=1	sGOTMIN		If we actually did get a minimum
		GOYES	nextpairtest	then test it

		GOSBVL	=GETPTR		Else return #0
		D1=D1+	10		Drop G and #pairs
		D=D+1	A
		D=D+1	A
		GOSBVL	=SAVPTR
		A=0	A
		R0=A	A
		GOVLNG	=Push#FLoop	( --> #0 FALSE )

* Do a criterion 2 test on the minimum

nextpairtest	A=R2	W		A[A1]=minpair	A[A]=mindegree
		GOSBVL	=ASRW5
		R2=A	A		R2[A]=minpair
		GOSUB	PairLCM		Calculate the LCM of the pair

* Test if the LCM is divisible by some Gk with k<>i k<>j

		A=0	A
		R1=A	A		R1[A] = k = 0

CritLoop	A=R1	A
		A=A+1	A		k=k+1
		R1=A	A
		C=R4	W
		GOSBVL	=CSRW5		C[A] = size of G
		?A<=C	A
		GOYES	+		Not tested all Gk yet
		GOTO	CritOK		Tested all Gk - pair is valid

+		A=R2	A		A[A] = minpair
		GOSUB	PairGet		D1 = ->pairdata
		A=R1	A		A[A] = k
		C=DAT1	A		C[A] = pair i
		?A=C	A
		GOYES	CritLoop	i=k - don't test Gk
		D1=D1+	5
		C=DAT1	A		C[A] = pair j
		?A=C	A
		GOYES	CritLoop	j=k - don't test Gk

* k<>i and k<>j - does Gk divide the LCM?

		C=R0	A		C[A] = ->$
		C=C+CON	A,10
		D0=C			D0 = ->slot[LCM]
		GOSUB	PairSlot	A[A] = ->slot[k]
		D1=A			D1 = ->slot[k]
		C=R3	A
		D=C	A		D[A] = vars
-		A=DAT0	A		A[A] = pow[lcm]
		C=DAT1	A		C[A] = pow[k]
		D0=D0+	5
		D1=D1+	5
		A=A-C	A		A[A] = pow[lcm]-pow[k]
		GOC	CritLoop	If overflow does not divide!
		D=D-1	A		D[A] = vars--
		?D#0	A
		GOYES	-		Test all variables

* hterm(Gk) divides the LCM - test if all pairs [i,k] [j,k] are tested
	
		A=R2	A		A[A] = minpair
		GOSUB	PairGet		D1 = ->pairdata[minpair]
		A=DAT1	A		A[A] = i
		D1=D1+	5
		C=DAT1	A
		B=C	A		B[A] = j
		C=R4	A
		D=C	A		D[A] = pairs
		GOSBVL	=D1=DSKTOP	D1 = ->stack
		D1=D1+	10		D1 = ->pair1 (on stack)
--		C=DAT1	A		C[A] = ->$pair
		D1=D1+	5
		D0=C			D0 = ->$pair
		D0=D0+	10		D0 = ->pair_i
		C=DAT0	A		C[A] = pair_i
		D0=D0+	5		D0 = ->pair_j
		?A=C	A		If pair_i == i or j then test pair_j
		GOYES	++
		?B=C	A
		GOYES	++
		AR1EX	A		Else if pair_i == k then test pair_j
		?A=C	A
		GOYES	+
critnxt1	AR1EX
critnxt2	D=D-1	A		Not this pair - try next one
		?D#0	A
		GOYES	--		Until all pairs checked
		GOTO	CritFAIL	Thus all pairs tested - discard pair!
+		AR1EX			pair_i == k -- test pair_j
		C=DAT0	A		C[A] = pair_j
		?A=C	A
		GOYES	goCritLoop	[i,k] not tested - try another k!
		?B=C	A
		GOYES	goCritLoop	[j,k] not tested - try another k!
		GONC	critnxt2	Else try next pair
++		C=DAT0	A		pair_i == i or j -- test pair_j
		AR1EX	A
		?A#C	A
		GOYES	critnxt1	pair_j <> k - try next pair
		AR1EX	A
goCritLoop	GOTO	CritLoop	[i,k] or [j,k] not tested - next k!

* The LCM was divisible and pairs [i,k] [j,k] were tested - discard pair!

CritFAIL	A=R2	A		A[A] = minpair
		GOSUB	PairRemove	Remove it from stack
*		LC(2)	#C2		Beep to indicate criterion worked
*		GOSBVL	=RCKBp
		GOTO	FindNextPair	Restart search for minimum

* Found a minimum pair - remove it from stack and return Gi,Gj

CritOK		GOSBVL	=GETPTR		( .. #pairs G )
		D1=D1+	10		Drop #pairs and G
		D=D+1	A
		D=D+1	A
		GOSBVL	=SAVPTR
		A=R4	A		A[A] = pairs (after removals)
		A=A-1	A		Remove the minimum pair too
		R0=A	A
		GOSBVL	=PUSH#		( .. #pairs' )
		GOSBVL	=SAVPTR		Save the current stack
		A=R2	A		A[A] = minpair
		CD1EX			C[A] = ->stack
		D1=C
		C=C+A	A
		A=A+A	A
		A=A+A	A
		C=C+A	A		C[A] = ->pair[minpair] on stack
		D0=C			D0 = ->srcend
		CD1EX			C[A] = ->stack	D1 = ->srcend
		A=DAT1	A		A[A] = ->$pair[minpair]
		R0=A	A		R0[A] = ->$pair[minpair]
		D1=D1+	5		D1 = ->targetend
		AD0EX
		D0=A
		C=A-C	A		C[A] = size
		GOSBVL	=MOVEUP		Fill the slot left by minimum pair
		GOSBVL	=GETPTR
		D1=D1+	5		And drop one object from stack
		D=D+1	A
		GOSBVL	=SAVPTR
		A=R0	A		A[A] = ->$pair[minpair]
		D0=A
		D0=D0+	10
		A=DAT0	A		A[A] = i[minpair]
		D0=D0+	5
		C=DAT0	A		C[A] = j[minpair]
		R0=A	A
		R1=C	A
		GOSBVL	=PUSH2#
		GOVLNG	=PushTLoop	( --> .. #pairs' #i #j TRUE )
* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
* Remove pair with index A[A] from stack
* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
PairRemove	A=A+1	A		A[A] = pair+1
		GOSBVL	=D1=DSKTOP	C[A] = ->stack
		C=C+A	A
		A=A+A	A
		A=A+A	A
		C=C+A	A		C[A] = ->pair on stack
		D0=C			D0 = ->srctail
		AD1EX			A[A] = ->stack
		D1=C
		D1=D1+	5		D1 = ->targettail
		C=C-A	A		C[A] = size
		GOSBVL	=MOVEUP
		GOSBVL	=GETPTR
		D1=D1+	5		DROP one stack object
		D=D+1	A
		GOSBVL	=SAVPTR
		A=R4	A
		A=A-1	A		R4[A] = pairs--
		R4=A	A
		RTN
* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
* Calculate LCM of powers for given pair A[A] into slot0
* Exit:		B[A] = total degree of LCM
*		sGCD1 set if GCD=1
* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
PairLCM		GOSUB	PairGet		D1 = ->pairdata
		A=DAT1	A		A[A] = i
		GOSUB	PairSlot	A[A] = ->slot[i]
		C=A	A
		D=C	A		D[A] = ->slot[i]
		D1=D1+	5
		A=DAT1	A		A[A] = j
		GOSUB	PairSlot	A[A] = ->slot[j]
		C=R0	A
		D1=C
		D1=D1+	10		D1 = ->slot[0] = ->slot[LCM]
		C=R3	A
		CDEX	A		D[A] = vars	C[A] = ->slot[i]
		B=0	A		B[A] = totdeg = 0

		ST=1	sGCD1		Assume GCD will be one
--		D0=A			D0 = ->power[j]
		A=DAT0	A		A[A] = pow[j]
		D0=D0+	5
		CD0EX
		RSTK=C			RSTK = ->power[j]++
		C=DAT0	A		C[A] = pow[i]
		D0=D0+	5
		?A>=C	A
		GOYES	+
		ACEX	A		A[A] = max(pow[i],pow[j]) = lcm
+		DAT1=A	A		Save lcmpow
		D1=D1+	5
		B=B+A	A		B[A] = totdeg = totdeg + lcm
		?C=0	A		If gcdpow <> 1 then..
		GOYES	+
		ST=0	sGCD1		...flag gcd<>1
+		AD0EX			A[A] = ->power[i]++
		C=RSTK			C[A] = ->power[j]++
*		ACEX	A		Order is actually irrelevant!
		D=D-1	A		D[A] = vars--
		?D#0	A
		GOYES	--		Loop for all variables
		RTN

* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
* Fetch ->pair from data stack, index A[A]
* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
PairGet		GOSBVL	=D1=DSKTOP	C[A] = ->stack
		C=C+A	A
		A=A+A	A
		A=A+A	A
		C=C+A	A
		D1=C
		D1=D1+	5		D1 = ->pair on stack
		C=DAT1	A
		D1=C			D1 = ->$pair
		D1=D1+	10		D1 = ->pairdata
		RTN
* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
* Calculate address of the slot for the powers of polynomial A[A]
* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
PairSlot	C=R3	A		C[A] = vars
		GOSBVL	=MUL#		B[A] = vars*index
		A=R0	A		A[A] = ->$
		A=A+B	A
		B=B+B	A
		B=B+B	A
		A=A+B	A		A[A] = ->$+5*index*vars
		A=A+CON	A,10		A[A] = ->slot[index]
		RTN
* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
* Allocate space and pop powers of leading terms into the memory.
* The first slot is left empty to be used by whatever subroutine.
* Input:	RPLCPU		List of polynomials on stack level 1
* Output:	R0[A]	->$
*		R3[A]	vars
*		R4[A]	k	(size of G)
* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
PopHTerms	GOSBVL	=SAVPTR

* Count variables

		A=DAT1	A
		D0=A			D0 = ->G
		GOSUBL	QVARS
		A=A-1	A		embedding list
		R3=A	A		R3[A] = vars

* Count polynomials in list

PopHTerms.1	A=DAT1	A
		D0=A			D0 = ->G
		GOSUBL	QLEN
		R4=A	A		R4[A] = k

* Allocate memory for leading monomials

		A=A+1	A		A[A] = k+1
		C=R3	A		C[A] = vars
		GOSBVL	=MUL#		B[A] = (k+1)*vars
		C=B	A
		C=C+C	A
		C=C+C	A
		C=C+B	A		C[A] = 5*(k+1)*vars
		RSTK=C
		GOSBVL	=MAKE$N
		A=DAT1	A		A[A] = ->G
		AD0EX			D0 = ->G
		D1=A			D1 = ->memory
		C=RSTK
		GOSBVL	=WIPEOUT	Clear data

* Now pop the powers for all leading terms

		A=R0	A		A[A] = ->$work
		D1=A
		D1=D1+	10		D1 = ->slot0
		D0=D0+	5		D0 = ->G1

PopHTermLoop
		AD1EX			A[A] = ->slot		* Next slot *
		C=R3	A		C[A] = vars
		A=A+C	A
		C=C+C	A
		C=C+C	A
		A=A+C	A
		AD1EX			D1 = ->slot++
		GOSUB	PopHTerm	Pop powers for Gi	* Powers *
		GOSUBL	QSKIP		D0 = ->Gi++		* Next Gi *
		A=DAT0	A
		LC(5)	=SEMI
		?A#C	A
		GOYES	PopHTermLoop
		RTN						* Done *

* Get powers of leading term of polynomial in D0 to D1
* Assumes the memory in D1 has been cleared.

PopHTerm	CD0EX
		RSTK=C			RSTK = ->Gi
		CD0EX
		CD1EX
		RSTK=C			RSTK = ->slot ->Gi
		CD1EX
-		A=DAT0	A
		LC(5)	=DOLIST
		?A#C	A
		GOYES	+		If not list then done
		GOSUBL	QLEN
		A=A-1	A		degree = len-1
		DAT1=A	A
		D0=D0+	5
		D1=D1+	5
		GONC	-
+		C=RSTK
		D1=C			D1 = ->slot
		C=RSTK
		D0=C			D0 = ->Gi
		RTN
* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
ENDCODE

*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
* Name:		HTERM
* Desc:		Return leading term of polynomial (trimmed!)
* Stack:	( p --> hterm )
*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
NULLNAME HTERM
CODE
		GOSBVL	=SAVPTR
		A=DAT1	A		For integers hterm = p itself
		D0=A			D0 = ->P
		A=DAT0	A
		LC(5)	=DOLIST
		?A=C	A
		GOYES	+
		GOVLNG	=GETPTRLOOP	Quick exit for integers

+		GOSUBL	QSIZE		C[A] = Psize
		GOSBVL	=GETTEMP	hterm always smaller than P!
		A=C	A		A[A] = Psize
		GOSBVL	=D1=DSKTOP
		C=DAT1	A		C[A] = ->P
		RSTK=C			RSTK = ->P
		A=A+C	A		A[A] = ->Ptail
		CD0EX			D0 = ->P
		D1=C			D1 = ->output (hterm)
		R0=C	A		R0[A] = ->hterm
		R1=A	A		R1[A] = ->Ptail
		GOSUB	FindHTerm	D0 = ->term (exists for trimmed!)

* Copy until hterm and hterm itself

		D0=D0+	5
		C=DAT0	A
		AD0EX
		A=A+C	A		A[A] = ->Ztail
		C=RSTK			C[A] = ->P
		D0=C
		C=A-C	A		C[A] = size to copy
		GOSBVL	=MOVEDOWN

* Then zero coefficients and SEMI's until the end of P

		C=R1	A
		B=C	A		B[A] = ->Ptail
		GOSUB	TailZeros

* Shrink memory and return hterm

		CD1EX			C[A] = ->hterm tail	R0[A]=->hterm
		GOSUBL	ShrinkQ		Shrink the allocated memory
		GOVLNG	=OverWrALoop	And overwrite P with hterm

ENDCODE

*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
* Name:		SPLIT
* Desc:		Split polynomial from the given term into two parts
* Stack:	( P #i --> P1 P2 )
* Notes:	If i=0	P1=0  P2=P
*		If i>n	P1=P  P2=0
*		If i=1  P1=hterm(P)
*		P1 and P2 are not trimmed!
*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
NULLNAME SPLIT
CODE
		GOSBVL	=POP#
		R1=A	A		R1[A] = i
		GOSBVL	=SAVPTR
		?A#0	A
		GOYES	++		If nonzero then continue

* i=0 --> P1=0 P2=P

		A=0	A
		GOSBVL	=PUSHhxs	( P Z0 )
		GOVLNG	(=SWAP)+5	( Z0 P )

* i>0 - check if P=const

++		A=DAT1	A
		D0=A			D0 = ->P
		A=DAT0	A
		LC(5)	=DOHSTR
		?A#C	A
		GOYES	+		Not constant - continue

* For constants there is no second term, thus P2=0

SPLITOVF	A=0	A
		GOVLNG	=PUSHhxsLoop	( P Z0 )

* find (i+1)'th term

+		A=R1	A
		A=A+1	A		A[A] = i+1
		GOSUB	FindTerm
		GOC	SPLITOVF	No such term - P2=0

		A=DAT1	A		A[A] = ->P
		CD0EX
		C=C-A	A
		R4=C	A		R4[A] = offset to term[i+1]

* Find i'th term

		D0=A			D0 = ->P
		A=R1	A		A[A] = i
		GOSUB	FindTerm
		A=DAT1	A		A[A] = ->P
		CD0EX
		C=C-A	A
		R3=C	A		R3[A] = offset to term[i]
	
* Allocate memory for P1 & P2 - twice size(P) is sufficient

		D0=A			D0 = ->P
		GOSUBL	QSIZE		C[A] = Psize
		C=C+C	A		C[A] = 2*Psize
		GOSBVL	=GETTEMP
		AD0EX
		R0=A	A		R0[A] = ->output = ->P1

* Output P1:
*	o Copy until term[i]
*	o Copy term[i]
*	o Transform remainder to zeros

		GOSBVL	=D1=DSKTOP	D1 = ->stack
		A=DAT1	A		A[A] = ->P
		C=R3	A		C[A] = offset to term[i]
		C=C+A	A
		CR4EX	A		R4[A] = ->term[i]
		C=C+A	A
		R3=C	A		R3[A] = ->term[i+1]

		D0=A			D0 = ->P
		GOSUBL	QSKIP
		C=DAT1	A		C[A] = ->P
		CD0EX			D0 = ->P
		D=C	A		D[A] = ->Ptail

		A=R0	A
		D1=A			D1 = ->output = ->P1
	
* Copy start until term[i] and term[i] itself

		C=R4	A		C[A] = ->term[i]
		CD0EX
		D0=D0+	5
		A=DAT0	A		A[A] = Zlen
		CD0EX
		C=C+A	A		C[A] = ->Ztail (term[i])
		AD0EX
		D0=A			D0 = ->P
		C=C-A	A
		GOSBVL	=MOVEDOWN

* Transform remainder to zeros

		C=D	A
		B=C	A		B[A] = ->Ptail
		GOSUB	TailZeros

* Save address of P2

		CD1EX
		R1=C	A		R1[A] = ->output = ->P2
		CD1EX

* Output P2:
*   o dolists until term[i+1] with backtracking on SEMI's
*   o copy term
*   o copy remainder

		GOSBVL	=D0=DSKTOP
		A=DAT0	A
		D0=A			D0 = ->P
		A=R3	A
		B=A	A		B[A] = ->term[i+1]
		GOSUB	LeadZeros

* Copy remaining coefficients as is

		C=D	A		C[A] = ->Ptail
		AD0EX
		D0=A			D0 = ->term[i+1]
		C=C-A	A
		GOSBVL	=MOVEDOWN

* Shrink memory

		CD1EX			C[A] = ->P2tail	R0[A]=->P1
		GOSUBL	ShrinkQ

* Return P1 & P2

		DAT1=A	A		Overwrite P with P1
		A=R1	A		A[A] = ->P2
		GOVLNG	=GPPushALp	( P --> P1 P2 )

* Should consider jumping to some PTrim2 subroutine

* Subroutine which transforms the remaining polynomial at D0 into
* zero coefficients and outputs to D1. The tail of the polynomial
* is given in B[A]

TailZeros	A=DAT0	A
		LC(5)	=SEMI
		?A=C	A
		GOYES	+
		GOSUBL	QSKIP		DOLIST and DOHSTR --> Z0
		LCZ0
		DAT1=C	11
		D1=D1+	11
		GONC	TailZeros
+		DAT1=A	A		Output SEMI
		D0=D0+	5
		D1=D1+	5
		AD0EX			Continue until end of P reached
		D0=A
		?A<B	A
		GOYES	TailZeros
		RTN

* Subroutine which transforms the start of the polynomial into zero
* coefficients and outputs to D1. The termination address is in B[A].

LeadZeros	A=DAT0	A
		D0=D0+	5
		LC(5)	=DOLIST
		?A=C	A
		GOYES	+		DOLIST --> copy it
		LC(2)	=DOHSTR
		?A=C	A
		GOYES	++		DOHSTR --> ignore or terminate
		D1=D1-	5		SEMI --> backtrack in output
		GONC	LeadZeros
+		DAT1=A	A
		D1=D1+	5
		GONC	LeadZeros
++		AD0EX			DOHSTR - check if at the end
		D0=A
		D0=D0-	5		D0 = ->Z
		?A>=B	A
		RTNYES			Yes - done
		D0=D0+	5
		C=DAT0	A
		A=A+C	A
		D0=A			D0 = ->Ztail
		GONC	LeadZeros	Continue until termination point

* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
* Find i'th term from polynomial
* Input:	D0	->poly
*		A[A]	index
* Exit:		D0	->term[i]
*		CS if no such term in polynomial
* Alters:	A[A] B[A] C[A] D[A]
* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
FindHTerm	B=0	A		*Entry for finding leading term*
		GOTO	+

FindTerm	A=A-1	A		*Entr for finding i'th term*
		RTNC			There is no 0'th term
		B=A	A		B[A] = i
+		D=0	A		D[A] = depth
--		A=DAT0	A
		D0=D0+	5
		LC(5)	=DOLIST
		?A=C	A
		GOYES	ftermlist	Handle list prolog
		LC(2)	=DOHSTR
		?A=C	A
		GOYES	ftermhstr	Handle coefficient
		D=D-1	A		D[A] = depth-- (for SEMI)
ftermnext	?D#0	A
		GOYES	--		Loop if depth <> 0
		RTNSC			CS: no such term
ftermlist	D=D+1	A		D[A] = depth++ (for DOLIST)
		GONC	--		And continue
ftermhstr	C=DAT0	A		C[A] = Zlen
		D0=D0+	6		Assume Z0 and skip it
		C=C-CON	A,7		C[A] = mantissalen-1
		GOC	ftermnext	Is zero - continue search
		B=B-1	A		B[A] = term--
		GOC	++		Overflow - fount the term
		AD0EX
		A=A+C	A		Skip to end of Zmantissa
		AD0EX
		D0=D0+	1		Skip Zsign
		GONC	ftermnext	And continue loop
++		D0=D0-	11		Skip back to Zprolog
ftermok		RTNCC			CC: found the term
* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
ENDCODE

*---------------------------------------------------------------------
* Name:		FINDSPLIT
* Desc:		Find a reducible term in r given a nontrivial basis G
* Stack:	( P G --> P G #i #Gi
* Notes:	#i indicates the index of the reducible term (0 if none)
*		#Gi is the reducing basis polynomial (0 if none)
*---------------------------------------------------------------------
NULLNAME FINDSPLIT
CODE
		GOSBVL	=SAVPTR

* We assume the basis does not contain constants and thus a constant
* P is irreducible

		D1=D1+	5
		A=DAT1	A
		D1=D1-	5
		D0=A			D0 = ->P
		A=DAT0	A
		LC(5)	=DOLIST
		?A=C	A
		GOYES	+
		GOTO	FoundNoSplit

* Count variables

+		GOSUBL	QVARS2		B[A]=vars(P)  C[A]=vars(G)
		C=C-1	A		embedding list
		?C>=B	A		choose maximum
		GOYES	+
		C=B	A
+		R3=C	A		R3[A] = vars

* Pop the powers of hterms of G

		GOSUB	PopHTerms.1	slot0 is free for powers of P

* Now:	R0[A] = ->$
*	R3[A] = vars
*	R4[A] = k	(size of G)
	
* Get powers for r

		GOSBVL	=D1=DSKTOP
		D1=D1+	5
		A=DAT1	A
		D0=A			D0 = ->P
		GOSUBL	QSKIP
		A=DAT1	A
		AD0EX			D0 = ->P
		R2=A	A		R2[A] = ->Ptail
		C=R0	A
		D1=C
		D1=D1+	10		D1 = ->slot0 = ->slot[P]
-		A=DAT0	A
		LC(5)	=DOLIST
		?A#C	A
		GOYES	+
		GOSUBL	QLEN
		A=A-1	A		degree = len-1
		DAT1=A	A
		D0=D0+	5
		D1=D1+	5
		GONC	-
+

* Now we test if the term from P (initially hterm) is divisible by
* the leading term of some Gi (coefficients ignored). If not we
* advance to the next term in P and try again until P is exhausted.
* Variables:
*	R0[A]  = ->$	(slot0 slot1 .. slotk )
*	R1[A]  = ->Pptr	(pointer to current term in P)
*	R1[A1] = ->Ppow (powers of P)
*	R2[A]  = ->Ptail
*	R3[A]  = vars
*	R4[A]  = k (basis size)
*	R4[A1] = i (P term index)

+		A=0	W
		A=R4.F	A
		R4=A	W		R4[A1] = index (Pterm) = 0

FindSplitLoop

		AD1EX			Save current ->Ppow
		GOSBVL	=ASLW5
		AD0EX			Save current ->Pptr
		R1=A	W

		A=R4	W		Increase term index
		GOSBVL	=ASRW5
		A=A+1	A		i=i+1
		GOSBVL	=ASLW5
		A=R4.F	A
		R4=A	W
	
* Divisibility test by all Gj if necessary

		A=R0	A
		D1=A
		D1=D1+	10		D1 = ->slot[P]
		D=0	A		D[A] = #Gi = 0
--		C=R3	A		Skip to next slot
		AD1EX
		A=A+C	A
		C=C+C	A
		C=C+C	A
		A=A+C	A
		D1=A			D1 = ->slot[Gi]
		B=A	A		B[A] = ->slot[Gi]
		D=D+1	A		D[A] = #Gi++
		A=R0	A
		D0=A
		D0=D0+	10		D0 = ->slot[P]
		C=R3	A		C[A] = vars
-		RSTK=C			RSTK = vars
		A=DAT0	A		A[A] = pow[P]
		C=DAT1	A		C[A] = pow[Gi]
		D0=D0+	5
		D1=D1+	5
		A=A-C	A		A[A] = pow[P]-pow[Gi]
		C=RSTK
		GOC	+		CS: not divisible -- try next Gi
		C=C-1	A
		?C#0	A
		GOYES	-		Test all variables
		GOTO	FoundSplit	Found split with indexes i and #Gi
+		A=B	A
		D1=A			D1 = ->slot[Gi]
		C=R4	A		C[A] = k
		?D<C	A
		GOYES	--		Try division with all Gi

* Current term in P is not divisible by any Gi - advance to next term

		A=R1	W
		D0=A			D0 = ->Pptr
		GOSBVL	=ASRW5
		D1=A			D1 = ->Ppow
		D1=D1-	5		Back to currently changing power
		D0=D0+	5		Skip the coefficient of the term
		C=DAT0	A
		AD0EX
		A=A+C	A
		AD0EX
--		C=DAT1	A		Decrease current power
		C=C-1	A
		GOC	+		If underflow we have a SEMI
		DAT1=C	A		Else store the decreased power
		A=DAT0	A
		LC(5)	=DOLIST
		?A=C	A
		GOYES	fndspllist	Handle DOLIST
		GONC	fndsplhstr	Handle DOHSTR
+		D1=D1-	5		Handle SEMI - back in ->Ppow
		D0=D0+	5		Skip SEMI
		AD0EX
		D0=A
		C=R2	A		C[A] = ->Ptail
		?A<C	A
		GOYES	--		Repeat unless reached Ptail
		GOTO	FoundNoSplit	In which case P is irreducible
	
fndspllist	D1=D1+	5		Create new power
		GOSUBL	QLEN
		DAT1=A	A		Set new power = len(list)
		D0=D0+	5		It will be decreased immediately
		GONC	--		at the start of the loop to "fix" it

fndsplhstr	D0=D0+	5
		C=DAT0	A		C[A] = Zlen
		D0=D0+	6		Assume Z0 and skip it
		C=C-CON	A,7
		GOC	--		Is Z0 - continue loop
		D0=D0-	11		D0 = ->Z

* Found new term from P - continue main loop to see if it is reducible

		D1=D1+	5		D1 = ->Ppow
		GOTO	FindSplitLoop	And loop

* Exit with found split (won't bother with a TrashTop call)

FoundSplit	C=D	A
		R1=C	A		R1[A] = #Gi
		A=R4	W
		GOSBVL	=ASRW5
		R0=A	A		R0[A] = #i
		GOVLNG	=Push2#Loop	( P G --> P G #i #Gi )

* Exit with no found split

FoundNoSplit	LC(5)	=ZEROZERO
		GOVLNG	=GETPTREVALC	( P G --> P G #0 #0 )

ENDCODE

*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
* Name:		REDUCE
* Desc:		Reduce polynomial with respect to given basis
* Stack:	( p G --> q )
*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
NULLNAME REDUCE
::

* Check trivial cases

  OVER TYPEHSTR? caseDROP		( * No reduction for constant * )
  DUPNULLCOMP? caseDROP			( * Empty basis * )

* Try fast assembly language reduction. This code is the main bottleneck
* in the Grobner basis algorithm. Basically what happens is that when
* a single reduction is done the polynomial being reduced gets new terms
* which often reduce too - all too commonly by the original polynomial
* which reduced the polynomial. Imagine adding a X^10*G1 polynomial to the
* polynomial when G1 is univariate - obviously we will be able to reduce
* that polynomial by G1 itself 10 (?) times.

  FASTREDUCE ?SEMI			( --> q )

* Slow reduction loop.

  HXS 1 0 ZERO
  {{ r G q i }}

  BEGIN
	i #1+ i!
	"Slow reduce " i #>$ &$ Verbose3
	r G FINDSPLIT			( r G #i #Gi )
	DUP#0<>
  WHILE
	ROTSWAP NTHCOMPDROP UNROT	( Gi r #i )
	#1- SPLIT			( Gi r1 r2 )
	q ROT QAdd q!			( Gi r2  ; q=q+r1 )
	DUP HTERM 3PICK HTERM		( Gi r2 Mr Mi )
	PPZ 4UNROLL			( Gi lc r2 Mr Mi' )
	QDiv				( Gi lc r2 Mr/Mi' )
	ROT RCanno			( Gi r2 Mr/Mi'' lc' )
	q OVER QMul q!			( Gi r2 Mr/Mi'' lc' ; q=q*lc' )
	ROT QMul			( Gi Mr/Mi'' r2*lc' )
	UNROT QMul			( r2*lc' Gi*Mr/Mi'' )
	QSub PTrim			( r2*lc'-Gi*Mr/Mi'' )
	r!
  REPEAT
					( r G #0 #0 )
  3DROP					( r )
  q QAdd				( q+r ; q=q+r )
  ABND
  NULL$ Verbose3
;

*---------------------------------------------------------------------
* Name:		FASTREDUCE
* Desc:		Assembly version of REDUCE for polynomials with short
*		coefficients. Returns FALSE if an overflow occurs
*		during the reduction.
* Stack:	( r P --> q TRUE / r P FALSE )
* Notes:	Assumes r is a genuine polynomial (not constant)
*		Assumes G is not empty
*		Assumes G does not contain zeros (is trimmed)
*---------------------------------------------------------------------
NULLNAME FASTREDUCE
CODE

* Flag equates

sREDUCED	EQU 8	* Indicates reduction has occurred
*sDUMMYCOEFF	EQU 8	* Dummy coefficients in Distribute
*sMODULAR	EQU 9	* Modular Distribute
*sGARB		EQU 10	* Garbagecollection flag
*sTRASH		EQU 11	* Trash allocated memory if overflow occurs

* Initialize flags

		ST=1	sTRASH		* Avoids almost certain GC..
		ST=0	sMODULAR	* Not modular
		ST=0	sGARB		* No GC yet

ReduceRetry	GOSBVL	=SAVPTR

* Check polynomials are short

		GOSUBL	QSHORT2?	CS if both are short
		GONC	ReduceFail

* Allocate all free memory to be work memory:
*
*		poly1		basis polynomial 1
*		poly2
*		...
*		polyk		basis polynomial k
*		BSS 16		16 nibble stopper
*		q		current result  (see REDUCE)
*		r		r		(see REDUCE)
*		[free]

		GOSBVL	=ROOM
		A=C	A		A[A] = room
		LC(5)	2000		Require 2000 nibbles free
		?A>=C	A
		GOYES	+
ReduceGC	GOSBVL	=DOGARBAGE
		GONC	ReduceRetry
ReduceFail	GOVLNG	=GPPushFLoop

+		CSRB.F	A
		C=A-C	A
		R4=C	A		R4[A] = free memory
		GOSBVL	=MAKE$N
		AD0EX
		R0=A	A		R0[A] = ->poly1
	
* Count variables

		GOSUBL	QVARS2		B[A] = vars(r)	C[A] = vars(G)
		C=C-1	A		embedding list
		?C>=B	A
		GOYES	+
		C=B	A
+		R3=C	A		R3[A] = vars

* Convert base polynomials to distributed form

		C=DAT1	A
		D0=C			D0 = ->G
		D0=D0+	5		D0 = ->G1
		C=R0	A
		D1=C	A		D1 = ->output = ->poly1
ReducePopLoop	C=R3	A
		B=C	A		B[A] = vars
		GOSUBL	Distribute
		GOC	ReduceGC1
		A=DAT0	A
		LC(5)	=SEMI
		?A#C	A
		GOYES	ReducePopLoop	Repeat until end of G

* 16-nibble stopper and q

		GOSUB	CSLOT		C[A] = slotsize
		C=C+CON	A,16		C[A] = slotsize + 16
		A=R4	A		Allocate it
		A=A-C	A
		GONC	+
ReduceGC1	GOTO	ReduceGC
+		R4=A	A		R4[A] = free
		GOSBVL	=WIPEOUT	Set q=0

* Convert r to distributed form

		AD1EX
		D1=A	
		R1=A	A		R1[A] = ->r (output)
		GOSBVL	=D0=DSKTOP
		D0=D0+	5
		A=DAT0	A
		D0=A			D0 = ->r (stack)
		C=R3	A
		B=C	A		B[A] = vars
		GOSUBL	Distribute
		GOC	ReduceGC1
		CD1EX
		GOSBVL	=CSLW5
		C=R4.F	A
		R4=C	W		R4[A1] = ->free

* R0[A]  = ->poly1
* R1[A]  = ->r
* R3[A]  = vars
* R4[A]  = free
* R4[A1] = ->free

* Reduce r

		GOSUB	REDUCE1		D0 = ->q

* Move q up to end of work memory

		A=R0	A
		A=A-CON	A,10
		R0=A	A		R0[A] = ->$
		A=A+CON	A,5
		D1=A
		C=DAT1	A
		A=A+C	A
		D1=A			D1 = ->$tail
		AD0EX
		B=A	A		B[A] = ->q
		AD0EX
		GOSUBL	DISTSKIP
		CD0EX			C[A] = ->qtail
		D0=C
		C=C-B	A		C[A] = qsize
		GOSBVL	=MOVEUP		D1 = ->q (at the end of $)

* All memory below q is now free memory

		A=R0	A
		A=A+CON	A,10		A[A] = ->output (q)
		CD1EX
		D0=C			D0 = ->q (at the end of $)
		D1=A			D1 = ->output
		C=C-A	A
		R4=C	A		R4[A] = free memory

* Convert to recursive form

		C=R3	A
		B=C	A		B[A] = vars
		GOSUBL	Recurse
		GONC	+
		GOTO	ReduceGC	Insufficient memory

* Shrink memory and exit

+		AD1EX
		D0=A			D0 = ->qtail (recursive)
		GOSBVL	=Shrink$
		A=R0	A
		A=A+CON	A,10		A[A] = ->q
		GOSBVL	=GETPTR
		D1=D1+	5
		DAT1=A	A		( q G )
		D1=D1-	5
		GOVLNG	=OverWrTLoop	( q TRUE )

* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
* Reduce r with respect to basis
* Input:	R0[A]	->poly1		(sequence of basis polynomials)
*			->q		followed by q
*		R1[A]	->r		polynomial to reduce
*		R3[A]	vars		variables
*		R4[A]	free		free memory
*		R4[A1]	->free		free memory address (after r)
* Exit:		R0[A]	->poly1
*		D0	->q
*		sREDUCED set if reductions occurred
* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

REDUCE1		ST=0	sREDUCED	Set no reductions yet

		A=R0	A		Save basis pointer
		GOSBVL	=ASLW5
		R0=A	W		R0[A1] = ->poly1

		GOSUB	CSLOT		Make sure q=0
		A=R1	A		A[A] = ->r
		A=A-C	A		A[A] = ->q
		D1=A
		GOSBVL	=WIPEOUT

* Inner loop:			 R0 = (->Gi)   ->G1  
*  If Mr divides Mf then	 R1 = ->r            
*    r=r-Mr/Mf*f : f=G1		 R2 =                
*  Else				 R3 = vars           
*    f=f++			 R4 = free     ->free
*  End
*  Until no more f or r=0

reduceloop	A=R0	W		Set f = poly1
		GOSBVL	=ASRW5
		R0=A.F	A

reduce1

* Calculate Mr/Mf = hterm(r)/hterm(f) = hterm(r)/hterm(Gi)

		A=R1	A
		D0=A			D0 = ->r (hterm)
		A=R0	A
		D1=A			D1 = ->f (hterm of basis polynomial)

		GOSUB	MKSLOT		Allocate 1 slot for Mr/Mf

		A=R4	W
		GOSBVL	=ASRW5
		B=A	A		B[A] = ->free (where Mr/Mf is put)

		A=DAT1	W		If G is empty then do nothing!
		?A=0	W
		GOYES	reducenxtf	(remains from old full assembly)

		D0=D0+	16		Skip coefficient fields
		D1=D1+	16
		B=B+CON	A,16
		C=R3	A
		D=C	A		D[A] = vars
-		A=DAT0	A		A[A] = pow[r]
		C=DAT1	A		C[A] = pow[f]
		D0=D0+	5
		D1=D1+	5
		A=A-C	A		A[A] = pow[r]-pow[f]
		GOC	reducenxtf	Not divisible - try next f
		C=B	A
		CD1EX
		DAT1=A	A		Output pow[Mr/Mf]
		D1=D1+	5
		CD1EX
		B=C	A
		D=D-1	A
		?D#0	A		Loop for all variables
		GOYES	-
		GONC	reducedof	Mf divides Mr - do reduction

* f does not reduce r - try next f

reducenxtf	GOSUB	DESLOT		deallocate Mr/Mf slot
		A=R0	A
		D0=A			D0 = ->f
		A=DAT0	W		Safety against empty G
		?A=0	W		(remains from old full assembly)
		GOYES	+
		GOSUB	DISTSKIP	D0 = ->f++
+		C=DAT0	W
		AD0EX
		R0=A.F	A		R0[A] = ->f++
		?C=0	W
		GOYES	+		
		GOTO	reduce1		Got next f - try if it reduces r
+		GOTO	reducer+	Final f checked - next term from r

* f reduces r - now calculate coefficients for reduction

reducedof	ST=1	sREDUCED	flag reduction occurred
		A=R1	A
		D0=A			D0 = ->r (->Mr)
		A=R0	A
		D1=A			D1 = ->f (->Mf)
		A=DAT0	W		A[W] = rcoeff
		C=DAT1	W		C[W] = fcoeff
		GOSUBL	ZSCANNO		Simplify
		R2=C	W		R2[W] = denominator (lc' in RPL code)
		C=R4	W
		GOSBVL	=CSRW5
		D1=C			D1 = ->free
		A=-A-1	S
		DAT1=A	W		->free has now -Mr/Mf with coefficient

* Multiply q with the denominator (lc' in RPL code)

		C=R2	W
		C=C-1	A
		?C=0	W		Save time of denominator==1
		GOYES	++
		A=R0	A
		D0=A			D0 = ->f
-		GOSUB	DISTSKIP	D0 = ->f++
		A=DAT0	W
		?A#0	W
		GOYES	-		Loop until basis skipped
		D0=D0+	16		D0 = ->q
-		A=DAT0	W		A[W] = qcoeff
		?A=0	W
		GOYES	++		done if q==0 or end of q
		C=R2	W		C[W] = denominator
		GOSUBL	ZSMUL
		DAT0=A	W		Set qcoeff *= denominator
		GOSUB	CSLOT		Skip to next slot in q
		AD0EX
		A=A+C	A
		AD0EX
		GONC	-		Repeat until end of q
++

* Now move r and -Mr/Mf (right after r) to end of work memory

		GOSUB	CSLOT
		AD1EX
		A=A+C	A
		D0=A			D0 = -Mr/Mf tail
		B=A	A		B[A] = -Mr/Mf tail
		A=R4	A		A[A] = free
		A=A+B	A		A[A] = ->end of free memory
		D1=A			D1 = ->freetail
		C=A-C	A
		CBEX	A		B[A] = ->-Mr/Mf' (new address)
		A=R1	A		A[A] = ->r
		C=C-A	A		C[A] = size of r and -Mr/Mf
		GOSBVL	=MOVEUP
		CD1EX
		CD0EX			D0 = ->r (new = input)
		CD1EX			D1 = ->r (old = output)
		C=B	A
		D=C	A		D[A] = ->-Mr/Mf

* Now calculate new r = denominator*r - Mr/Mf*f

ReduceAddLp1	GOSUB	MKSLOT		Allocate 1 slot for next term

* Calculate -Mr/Mf*f into the slot

		A=R0	A
		AD0EX			D0 = ->fterm
		R0=A	A		R0[A] = ->rterm
		A=DAT0	W
		?A#0	W
		GOYES	++		End of f not reached yet - sum

* f traversed - output remaining denominator*r
		A=R0	A
		D0=A			D0 = ->rterm
--		A=DAT0	W		A[W] = rcoeff
		C=R2	W		C[W] = denominator
		GOSUBL	ZSMUL
		DAT1=A	W		Output rcoeff*denominator
		D0=D0+	16
		D1=D1+	16
		?A=0	W		If was at end of r then we're done
		GOYES	+
		C=R3	A		Else output powes = powers of r
-		A=DAT0	A		A[A] = pow[r]
		DAT1=A	A
		D0=D0+	5
		D1=D1+	5
		C=C-1	A
		?C#0	A
		GOYES	-		Loop for all variables
		GONC	--		Loop until end of r
+		GOTO	ReduceAddOK	Recalculate variables after addition

* f not traversed yet output -Mr/Mf*fterm

++		CD1EX
		CDEX	A		D[A] = ->r (output)
		CD1EX			D1 = ->-Mr/Mf
		C=DAT1	W		C[W] = -Mr/Mf coefficient
		CD1EX
		CDEX	A		D[A] = ->-Mr/Mf
		CD1EX			D1 = ->r (output
		GOSUBL	ZSMUL		A[W] = -rcoeff*Mr/Mfcoeff
		DAT1=A	W		Output the coefficient
		D0=D0+	16		Skip coefficient fields
		D1=D1+	16		and calculate powers
		C=R3	A
		B=C	A		B[A] = vars
		C=D	A
		C=C+CON	A,16		C[A] = ->Mr/Mf power 1
-		A=DAT0	A		A[A] = fcoeff
		D0=D0+	5
		CD1EX			D1 = ->Mr/Mf power
		RSTK=C			RSTK = ->r (output)
		C=DAT1	A		C[A] = pow[Mr/Mf]
		D1=D1+	5
		A=A+C	A		A[A] = product
		C=RSTK
		CD1EX			D1 = ->r (output)  C[A]=->Mr/Mf power
		DAT1=A	A		Output power
		D1=D1+	5
		B=B-1	A
		?B#0	A
		GOYES	-		Loop for all variables

* Now see which is greater in lexicographic ordering,
* denominator*r or -fterm*Mr/Mf

		AR0EX	A
		AD0EX			D0 = ->rterm
		AR0EX	A		R0[A] = ->fterm++

* This inner loop of ReduceAddLp1 assumes -fterm*Mr/Mf has already been
* calculated. Namely, if r term is greater we simply move the term upwards
* in memory so we don't have to recalculate it. Thus all code above
* can be avoided. We jump to ReduceAddLp1 only when needed.

ReduceAddLp2	A=DAT0	W		Check if r done
		?A#0	W
		GOYES	+
		GOTO	ReduceAddLp1	If so then restart with new terms

+		GOSUB	CSLOT		Else compare powers
		AD1EX
		A=A-C	A		A[A] = ->-Mr/Mf*f
		AD1EX
		C=R3	A
		B=C	A		B[A] = vars
		D0=D0+	16		Skip coefficient fields
		D1=D1+	16
-		A=DAT0	A		A[A] = pow[r]
		C=DAT1	A		C[A] = pow[-Mr/Mf*fterm]
		D0=D0+	5
		D1=D1+	5
		?A#C	A
		GOYES	ReduceAdd#	If <> then one must be greater
		B=B-1	A
		?B#0	A
		GOYES	-		Loop for all variables

* denominator*rterm and -fterm*Mr/Mf have equal powers - add coefficients

ReduceAdd=	GOSUB	CSLOT
		RSTK=C			RSTK = slotsize
		AD0EX
		A=A-C	A
		AD0EX			D0 = ->rterm
		AD1EX
		A=A-C	A
		AD1EX			D1 = ->-fterm*Mr/Mf
		A=DAT0	W		A[W] = rcoeff
		C=R2	W		C[W] = denominator
		GOSUBL	ZSMUL		A[W] = rcoeff*denominator
		C=DAT1	W		C[W] = -fterm*Mr/Mf coefficient
		GOSUBL	ZSADD		A[W] = result coefficient
		C=RSTK
		B=C	A		B[A] = slotsize
		AD0EX
		A=A+B	A		A[A] = ->rterm++
		AD0EX
		C=R4	A
		C=C+B	A		deallocate rterm slot
		?A#0	W
		GOYES	+
		C=C+B	A		cancellation - deallocate output slot
+		R4=C	A
		?A=0	W
		GOYES	+		Ignore output if cancellation
		DAT1=A	W		Output coefficient
		AD1EX
		A=A+B	A
		AD1EX			D1 = ->rterm++ (output)
+		GOTO	ReduceAddLp1	fterm was used - use loop 1

* Not equal powers - first skip to end of powers

ReduceAdd#	B=B-1	A
		?B=0	A
		GOYES	+
		D0=D0+	5
		D1=D1+	5
		GONC	ReduceAdd#

* Now test which term is greater in lexicographic ordering

+		?A>C	A
		GOYES	ReduceAdd>	rterm is greater

* rterm is smaller - output term from fterm

ReduceAdd<	GOSUB	CSLOT		skip back to old rterm
		AD0EX
		A=A-C	A
		AD0EX
		GOTO	ReduceAddLp1	fterm was used - use loop 1
	
* rterm is greater - move -Mr/Mf*fterm up 1 slot and output rterm

ReduceAdd>	GOSUB	MKSLOT		Allocate new slot for -Mr/Mf*fterm
		B=C	A		B[A] = slotsize
		CD0EX
		RSTK=C			RSTK = ->rtermtail
		CD0EX
		AD1EX
		D0=A
		A=A+C	A
		D1=A
		GOSBVL	=MOVEUP		Move -Mr/Mf*fterm up 1 slot
		C=RSTK
		D0=C	
		C=B	A
		GOSBVL	=MOVEUP		Copy ->rterm to output

* The actual term to output is rterm*denominator - calculate the coefficient

		C=B	A
		RSTK=C			RSTK = slotsize
		A=DAT1	W		A[W] = rcoeff
		C=R2	W		C[W] = denominator
		GOSUBL	ZSMUL
		DAT1=A	W

		C=RSTK			C[A] = slotsize
		AD1EX
		A=A+C	A		A[A] = ->out++
		A=A+C	A		A[A] = -Mr/Mf*ftail
		AD1EX
		AD0EX
		A=A+C	A		A[A] = ->rterm++
		AD0EX
		A=R4	A		Deallocate rterm
		A=A+C	A
		R4=A	A
		GOTO	ReduceAddLp2		fterm was not used - loop 2

* Addition done - fix variables and re-enter reduction loop
* To be absolutely sure the following deallocation is correct we
* could just find ->$tail and substract ->rtail from it.
* As it stands, since the algorithm *calculates* ->$tail in the
* algorithm and moves data there and no crashes have occurred we know for
* sure the deallocation is not too big.
* Deallocation:	output slot		+slotsize
*		terminator		-16
*		Mr/Mf slot		+slotsize
*		last rterm slot		+slotsize-16
*					============
*					+3*slotsize
*
* Should insert code here to recalculate free memory and check with
* a debugger above is correct. For now 2*slotsize has worked fine
* The reason I don't want to recalculate free memory now is that I
* may want to use some data at the tail of $ later on, like the full
* assembly version had pairs at the tail.

ReduceAddOK	AD1EX			A[A] = ->free (new) = ->rtail
		GOSBVL	=ASLW5
		GOSUB	CSLOT
		A=R4.F	A
		A=A+C	A		deallocate 2 slots
		A=A+C	A
		R4=A	W

* Restart reduction for new r unless r=0

		A=R1	A
		D0=A			D0 = ->r
		A=DAT0	W
		?A=0	W
		GOYES	ReduceDone	r=0 - done
		GOTO	reduceloop	Set f=poly1 and restart reduction

* Mr e.g. hterm(r) does not reduce - update q=q+Mr ; r=r-Mr
* In lexicographic ordering this is simple - just move Mr
* over the terminator to tail of q.

reducer+	A=R1	A
		D0=A			D0 = ->rterm
		GOSUB	CSLOT		C[A] = slotsize
		A=A-C	A
		D1=A			D1 = ->qtail
		GOSBVL	=MOVEDOWN	Copy the rterm to qtail
		A=0	W
		DAT1=A	W		terminate the new q

* Try with the new Mr unless r=0

		C=DAT0	W
		AD0EX
		R1=A.F	A		R1[A] = ->r (new)
		?C=0	W
		GOYES	ReduceDone	r=0 - done
		GOTO	reduceloop	

* Done - return q

ReduceDone	A=R4	A		deallocate r=0
		A=A+CON	A,16		(+old r terminator)
		R4=A	A
		A=R0	W
		GOSBVL	=ASRW5
		R0=A	A		R0[A] = ->poly1
		D0=A
-		A=DAT0	W		Skip to end of basis
		?A=0	W
		GOYES	+
		GOSUB	DISTSKIP
		GONC	-
+		D0=D0+	16		D0 = ->q
		RTN
* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
* Skip polynomial in distributed form
* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
DISTSKIP	GOSUB	CSLOT		C[A] = slotsize
-		A=DAT0	W		Skip slots until terminator found
		?A=0	W
		GOYES	+
		AD0EX
		A=A+C	A
		AD0EX
		GONC	-
+		D0=D0+	16		Skip terminator
		RTNCC
* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
DESLOT		GOSUB	CSLOT		C[A] = slotsize
		AR4EX	A
		A=A+C	A		deallocate 1 slot
		AR4EX	A
		RTN
* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
MKSLOT		GOSUB	CSLOT		C[A] = slotsize
		AR4EX	A
		A=A-C	A		allocate 1 slot
		AR4EX	A
		RTNNC
		GOLONG	ReduceGC	Insufficient memory
* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
CSLOT		AR3EX	A		A[A] = vars
		C=A	A
		C=C+C	A
		C=C+C	A
		C=C+A	A		C[A] = 5*vars
		AR3EX	A
		C=C+CON	A,16		C[A] = slotsize
		RTN
ENDCODE
**********************************************************************







