/****************************************************************************
 ****				   MODEM.C				 ****
 *--------------------------------------------------------------------------*
 * Modul	: Utility to strip preceding (offset / 2) bytes from a file *
 * Language	: ANSI C						    *
 * Destination	: DOS etc.						    *
 * Author	: Detlef Mueller, 1992/04/14				    *
 * Description	: Q&d hack - Usage: STRIP #[hex-offset] <[input] >[output]  *
 *--------------------------------------------------------------------------*
 * Edition History :							    *
 *    0.000 1992/04/14 DM alpha version					    *
 ****************************************************************************/

/****************************************************************************
 * RCS history
 *
 * $Id: strip.c 1.2 1998/11/05 21:59:11 Detlef Rel $
 *
 * $Log: strip.c $
 * Revision 1.2  1998/11/05 21:59:11  Detlef
 * Overhauled
 *
 ****************************************************************************/


/****************************************************************************
 *			       I N C L U D E S 				    *
 ****************************************************************************/


#include	<stdio.h>
#include	<stdlib.h>

#include	<io.h>				/* MesS-DOS */
#include	<fcntl.h>			/* MesS-DOS */


/****************************************************************************
 *			        D E F I N E S 				    *
 ****************************************************************************/


#define	BUFLEN		1024


/****************************************************************************
 *			      F U N C T I O N S 			    *
 ****************************************************************************/


/** Error *******************************************************************
 * Interface :								    *
 *	Error( msg ) ;							    *
 *--------------------------------------------------------------------------*
 * Parameter :								    *
 *	char *msg	--> to string to output before quiting		    *
 *--------------------------------------------------------------------------*
 * Description :							    *
 ****************************************************************************/

void Error ( char *msg )
{
    fprintf( stderr, "\n%s\n\n", msg ) ;
    exit( -1 ) ;
}

/** main ********************************************************************
 * Interface :								    *
 *	main								    *
 *--------------------------------------------------------------------------*
 * Description :							    *
 *	Put all the stuff together ...					    *
 ****************************************************************************/

int main ( int argc, char *argv[] )
{
    static char
	buffer[BUFLEN],
	work[] = "/-\\|" ;
    long
	adr ;
    int
	i, wi = 0 ;

    fprintf( stderr,
	     "\n$Id: strip.c 1.2 1998/11/05 21:59:11 Detlef Rel $"
	     " (c) 1992-98 DM.\n" ) ;

    if ( argc != 2 ||
	 *argv[1] != '#' ||
	 ! (i = sscanf( argv[1] + 1, "%lX", &adr )) ||
	 i == EOF )
	Error( "Invallid command line.\n"
		"Usage: STRIP #[hex-offset] <[input] >[output]" ) ;

    setmode( fileno( stdout ), O_BINARY ) ;	/* MesS-DOS */
    setmode( fileno( stdin ), O_BINARY ) ;	/* MesS-DOS */

    if ( fseek( stdin, adr >> 1L, SEEK_SET ) )
	Error( "Input out of Range" ) ;

    fprintf( stderr, "\nStripping input ...  " ) ;

    if ( fputs( "HPHP48-X", stdout ) == EOF )
	Error( "Can't write BDL header" ) ;

    do
    {
	i = fread( buffer, 1, BUFLEN, stdin ) ;
	fwrite( buffer, 1, i, stdout ) ;

	if ( ferror( stdin ) || ferror( stdout ) )
	    Error( "Can't copy stdin to stdout" ) ;

	fputc( '\b', stderr ) ;
	fputc( work[wi = (wi + 1) & 3], stderr ) ;
    }
    while ( i == BUFLEN && ! feof( stdin ) ) ;

    fprintf( stderr, "\b \n\n" ) ;

    exit( 0 ) ;
}

/****************************************************************************
 ****			       End of STRIP.C				 ****
 ****************************************************************************/